package HFE;

import java.util.*;
//-----------------------------------------------------------------------------
// Project:     Implementing HFE at UCC as Project CS4095
// Supervisors: Simon Foley (Computer Science) and Patrick Fitzpatrick (Maths)
//-----------------------------------------------------------------------------

/**
<p>
This class encapsulates the operations for a finite field
with 2 elements. <br>
Operations are performed using and (mul) and xor (add).
<p>
@author Christopher Wolf (hfe@Christopher-Wolf.de)
@version 0.1
*/

//Copyright (C) 2002    Christopher Wolf
//
//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License (file LICENCE) for more details.


final class Field2 {

  /** number of elements in this field
   @serial constant for users of that class
  */
  public final static int elements = 2;


  /** Checks if a given integer is within the range of
      a Field2 element.
    @return is elem in the range of an Field2 element
  */
  public static boolean isElement(int elem) {
    return (elem < elements) && (elem >= 0);
  }

  /** Checks if a given vector consists only
      of Field2 elements.
    @return true iff all elements of the vector
            are Field2 elements.
  */
  public static boolean isElement(int[] elems) {
    for (int i = 0; i < elems.length; i++)
      if (!isElement(elems[i])) return false;
    return true;
  }

  /** Checks if a given matrix consists only
      of Field2 elements.
    @return true iff all elements of the matrix
            are Field2 elements.
  */
  public static boolean isElement(int[][] elems) {
    // uses the vector version of isElement
    for (int i = 0; i < elems.length; i++)
      if (!isElement(elems[i])) return false;
    return true;
  }

  /** Adds two numbers using the xor function for intergers.
      <i>Note: This works for any extionsion field over p=2.</i>
   @param ter1 element of F_4
   @param ter2 element of F_4
   @return     (ter1 + ter2) over F_4
  */
  public final static int add(int ter1, int ter2) {
    Check.assert(isElement(ter1) && isElement(ter2));
    return ter1 ^ ter2;
  }


  /** Subtracts two numbers using the xor function for intergers.
      <i>Note: This works for any extionsion field over p=2</i>
   @param ter1 element of F_4
   @param ter2 element of F_4
   @return     (ter1 - ter2) over F_4
  */
  public final static int sub(int ter1, int ter2) {
    Check.assert(isElement(ter1) && isElement(ter2));
    return ter1 ^ ter2;
  }

  /** Multiplies two numbers using finite field arithmetic in F_4.
      For the sake of speed I store the results in an lookup table.
   @param fac1 element of F_4
   @param fac2 element of F_4
   @return     (fac1 * fac2) over F_4
  */
  public final static int mul(int fac1, int fac2) {
    Check.assert(isElement(fac1) && isElement(fac2));
    return fac1 & fac2;
  }

  /** Divides two numbers using finite field arithmetic in F_2.
      Should not be used as the only by possible is 1...
   @param whom which number should be divised
   @param by   which number divides
   @return     (whom/by) over F_2
  */
  public final static int div(int whom, int by) {
    Check.assert(isElement(whom) && isElement(by));
    Check.assert(by != 0);
    return whom;  // our only by possible is 1...
  }

  /** returns a random element from Field2.
  */
  public final static int rnd(Random rnd) {
    return rnd.nextInt(elements);
  }


  /** Tests the functionality of the whole class.
      If there occurs any error it will be printed on stdout.
  */
  public final static void testIt () {
    System.out.println("start testing Field2 ...");

    // test add
    if (add(1,1) != 0) System.out.println("Error 1.1");
    if (add(0,1) != 1) System.out.println("Error 1.2");
    if (add(0,0) != 0) System.out.println("Error 1.3");
    if (add(1,0) != 1) System.out.println("Error 1.4");

    // test mul
    if (mul(0,0) != 0) System.out.println("Error 2.1");
    if (mul(0,1) != 0) System.out.println("Error 2.2");
    if (mul(1,0) != 0) System.out.println("Error 2.3");
    if (mul(1,1) != 1) System.out.println("Error 2.4");

    // test div
    if (div(0,1) != 0) System.out.println("Error 3.1");
    if (div(1,1) != 1) System.out.println("Error 3.2");

    // test the symmetry of addition an multiplication
    for (int a = 0; a < elements; a++) {
      for (int b = 0; b < elements; b++) {
        if (add(a,b) != add(b,a)) System.out.println("Error 4.1 with " + a + " " + b);
        if (mul(a,b) != mul(b,a)) System.out.println("Error 4.2 with " + a + " " + b);
      }
    }

    // test add and mul using the distributive law
    // a(b+c) = ab + ac
    for (int a = 0; a < elements; a++) {
      for (int b = 0; b < elements; b++) {
        for (int c = 0; c < elements; c++) {
          if (mul(a,add(b,c)) != add(mul(a,b),mul(a,c))) System.out.println("Error 4.3 with " + a + " " + b + " " + c);
        }
      }
    }

    // test mul and div using the property
    // (a*b)/b = a  with b != 0
    for (int a = 0; a < elements; a++) {
      for (int b = 1; b < elements; b++) {
        if (a != div(mul(a,b),b)) System.out.println("Error 4.4 with " + a + " " + b);
      }
    }

    // test isElement
    if (isElement(5)) System.out.println("Error 5.1");
    if (!isElement(1)) System.out.println("Error 5.2");

    System.out.println("...finished testing.");
  }

  /** Tests everything in this class
      @see #testIt
    */
  public static void main(String[] args) {
    testIt();
  }
}
