package HFE;

import java.util.*;
import HFE.*;
//-----------------------------------------------------------------------------
// Project:     Implementing HFE at UCC as Project CS4095
// Supervisors: Simon Foley (Computer Science) and Patrick Fitzpatrick (Maths)
//-----------------------------------------------------------------------------

/**
<p>
This class encapsulates the operations for a finite field
with 2^n elements. <br>
The generating polynomials are
<ul>
  <li> f(x) = x^67 + x^5 + x^2 + x + 1 (i.e. 39=32+4+2+1 as reduction factor) </li>
  <li> f(x) = x^129 + x^5 + 1 (i.e. 33=32 + 1 as reduction factor) </li>
  <li> f(x) = x^257 + x^12 + 1 (i.e. 4097=4096+1 as reduction factor) </li>
</ul>
<p>
@author Christopher Wolf (hfe@Christopher-Wolf.de)
@version 0.1
*/

//Copyright (C) 2002    Christopher Wolf
//
//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License (file LICENCE) for more details.


final class Field2_n {

  /** Degree of the underlying irreducible polynomial.
      Value comes from class HFE2_n
  */
  public final static int degree = HFE2_n.extensionDegree;

  /** determinates the size of elements of this class. The assumption is that one integer has 32 bits.
  */
  public final static int size = (degree+31) / 32;

  /** Stores the reduction term as bit vector.
      The assumption is that - apart from x^degree -
      the other factors have degree less than 32
  */
  public final static int reductionTerm = HFE2_n.reductionTerm;

  /** Allocats memory for elements of this class
  */
  private int[] number = new int[size];

  /** Standart Constructor to obtain an element with all bits = 0
  */
  Field2_n () {
  }

  /** Constructor to produce a copy of another Field2_n element
  */
  Field2_n (final Field2_n father) {
    this.copy(father);
  }

  /** Constructor to obtain an element with least significant integer set and the rest being 0
  */
  Field2_n (int init) {
    number[0] = init;
  }

  /** Constructor to obtain a random element
  */
  Field2_n (Random rnd) {
    rnd(rnd);
  }

  /** Checks if bit[0] is _not_ set, i.e. if the number is even
   @return     true iff bit[0] is 0
  */
  public boolean even() {
    return (number[0] & 1) == 0;
  }

  /** Checks if the element is the zero element
   @return     true iff the element is zero
  */
  public boolean isZero() {
    for (int i = 0; i < size; i++)
      if (number[i] != 0) return false;
    return true;
  }

  /** Checks if the bit at position where is set.
   @param where position in this which should be checked
   @return     true iff bit is 1
  */
  private boolean checkBit(int where) {
    Check.assert(where <= degree, "Where too big " + where);
    Check.assert(where >= 0, "Where too small " + where);
    return (number[where / 32] & (1 << (where % 32))) != 0;
  }

  /** Sets bit at position where.
   @param where position in this which should be set
  */
  private void setBit(int where) {
    Check.assert(where <= degree, "Where too big " + where);
    Check.assert(where >= 0, "Where too small " + where);
    number[where / 32] |= 1 << (where % 32);
  }

  /** Clears bit at position where.
   @param where position in this which should be cleared
  */
  private void clearBit(int where) {
    Check.assert(where <= degree, "Where too big " + where);
    Check.assert(where >= 0, "Where too small " + where);
    number[where / 32] &= ~(1 << (where % 32));
  }

  /** xors the bit at position where.
   @param where position in this which should be xored
  */
  private void xorBit(int where) {
    Check.assert(where <= degree, "Where too big " + where);
    Check.assert(where >= 0, "Where too small " + where);
    number[where / 32] ^= (1 << (where % 32));
  }

  /** this = ter
      - copies father integerwise into this
   @param father element of Field2_n
  */
  public final void copy(Field2_n father) {
    Check.assert(father.isElement());
    for (int i = 0; i < size; i++)
      number[i] = father.number[i];
  }

  /** this <<= 1
       - therefore computes a left shift of this by one bit.
       Does not assume that the highest significant integer
       is valid (i.e. only the bits less the degree
       of the generating polynomial can seen to be valid).
  */
  private void shiftL() {
    number[size-1] <<= 1;  // left shift of our highest integer
    for (int i = size-2; i >= 0; i--) {
      // check if we will loose our highest bit, i.e. is this bit=1?
      if (number[i] < 0)
        number[i+1] |= 1;
      number[i] <<= 1;  // left shift this integer
    }
  }

  /** this <<= 1 % i(t)
       - therefore computes a left shift of this by one bit and reduces
       the result by the generating polynomials i(t).
       This left-shift is equivalent to multiplication with t.
       Assumes that this is valid before and after this computation.
  */
  private void shiftLMod() {
    Check.assert(isElement());
    // first perform left shift
    number[size-1] <<= 1;  // left shift of our highest integer
    for (int i = size-2; i >= 0; i--) {
      // check if we will loose our highest bit, i.e. is this bit=1?
      if (number[i] < 0)
        number[i+1] |= 1;
      number[i] <<= 1;  // left shift this integer
    }
    // now reduce if necessary
    if (checkBit(degree)) { // do we have to reduce?
      // we assume that only number[0] and bit_{degree} are affected by reduction
      clearBit(degree);
      number[0] ^= reductionTerm;
    }
    Check.assert(isElement());
  }

  /** this >>= 1
       - therefore computes a right shift of this by one bit.
       Assumes that bit[0] is 0, i.e. this is even.
       For the implementation platform (x86, Linux, Java 1.3)
       it turned out that >>= 1 keeps the sign bit
       untouched. The workaround it to delete
       the sign flag manually (required a second step...)
  */
  private void shiftR() {
    Check.assert(even(), "this must be even " + stringIt());
    boolean carry = (number[size-1] & 1) == 1;
    number[size-1] >>= 1;  // right shift number
    number[size-1] &= ~(1 << 31); // delete sign flag
    for (int i = size-2; i > 0; i--) {
      boolean carry2 = (number[i] & 1) == 1;
      number[i] >>= 1;  // right shift number
      number[i] &= ~(1 << 31); // delete sign flag
      if (carry) number[i] |= (1 << 31);
      carry = carry2;
    }
    number[0] >>= 1;
    number[0] &= ~(1 << 31); // delete sign flag
    if (carry) number[0] |= (1 << 31);
  }

  /** Checks if the current element is from Field2_n
    @return true iff this is a valid element from Field2_n
  */
  private boolean isElement() {
    for (int i = (degree%32); i < 32; i++)
      if ((number[size-1] & (1 << i)) != 0) return false;
    return true;
  }

  /** Checks if a given integer is within the range of
      a GF(2^n) element.
    @return is elem in the range of an Field2_n element
  */
  public static boolean isElement(Field2_n elem) {
    return elem.isElement();
  }

  /** Checks if a given vector consists only
      of Field2_n elements.
    @return true iff all elements of the vector
            are Field2_n elements.
  */
  public static boolean isElement(Field2_n[] elems) {
    for (int i = 0; i < elems.length; i++)
      if (!isElement(elems[i])) return false;
    return true;
  }

  /** Checks if a given matrix consists only
      of Field2_n elements.
    @return true iff all elements of the matrix
            are Field2_n elements.
  */
  public static boolean isElement(Field2_n[][] elems) {
    // uses the vector version of isElement
    for (int i = 0; i < elems.length; i++)
      if (!isElement(elems[i])) return false;
    return true;
  }

  /** Checks if this and who are bitwise equal
    @return true iff this == who
  */
  public boolean isEqual(Field2_n who) {
    for (int i = 0; i < size; i++)
      if (number[i] != who.number[i]) return false;
    return true;
  }

  /** sets this to zero
  */
  public void setZero() {
    for (int i = 0; i < size; i++)
      number[i] = 0;
  }

  /** sets this to 1
  */
  public void setOne() {
    setZero();
    setBit(0);
  }

  /** this += 1
      - adds one to this
  */
  public void addOne() {
    xorBit(0);
  }

  /** sets field element regarding to a bit vector. this is overwritten.
    @param bitVec elements from GF(2) for field elements
  */
  public void setElement(int [] bitVec) {
    Check.assert(bitVec.length == degree, "wrong length");
    Check.assert(Field2.isElement(bitVec), "only elements from Field2 are allowed");
    setZero();
    for (int i = 0; i < degree; i++)
      number[i >> 5] |= bitVec[i] << (i & 31);
  }

  /** copies this to a bit vector
    @param bitVec element in form of bit vector
  */
  public void getElement(int [] bitVec) {
    Check.assert(bitVec.length == degree, "wrong length");
    for (int i = 0; i < degree; i++)
      bitVec[i] = (number[i >> 5] >> (i & 31)) & 1;
  }

  /** find the degree of this
      - i.e. the position of the highest bit which is not zero.
      If all bits are zero, this function returns -1
    @return position of the first bit which is not zero. Range: -1, .., degree-1
  */
  private int cntDegree() {
    // first find the right integer
    int pos = size-1;
    while ((number[pos] == 0) && (pos > 0)) pos--;
    if (number[pos] == 0) return -1;  // check if bitstring is zero

    // we know that number[pos] != 0
    int value = number[pos];  // for speed - less array invocations
    int bit = 31;
    while ((value & (1 << bit)) == 0) bit--;
    return pos*32 + bit;
  }

  /** this = ter1 + ter2
      Adds two numbers using the xor function for intergers.
      <i>Note: This works for any extionsion field over p=2.</i>
      The result is stored in this.
   @param ter1 element of Field2_n
   @param ter2 element of Field2_n
  */
  public final void add(Field2_n ter1, Field2_n ter2) {
    Check.assert(ter1.isElement() && ter2.isElement());
    for (int i = 0; i < size; i++)
      number[i] = ter1.number[i] ^ ter2.number[i];
  }

  /** this += ter
      Adds two numbers using the xor function for intergers.
      <i>Note: This works for any extionsion field over p=2.</i>
      The result is stored in this.
   @param ter element of Field2_n
  */
  public final void add(Field2_n ter) {
    Check.assert(ter.isElement());
    for (int i = 0; i < size; i++)
      number[i] ^= ter.number[i];
  }

  /** this += generating polynomial
      Adds the modul to this. Exploits the fact that
      generating polynomial only changes two integers in our representation.
      Assumes that this is a valid element of GF(2^n)
  */
  private final void addModul() {
    Check.assert(isElement());
    setBit(degree);  // set the higest bit - we know that it is zero as this is an Element of GF(2^n)
    number[0] ^= reductionTerm; // mimic addition using xor.
  }

  /** this = ter1 - ter2
      Subtracts two numbers using the xor function for intergers.
      <i>Note: This works for any extionsion field over p=2.</i>
      The result is stored in this. Please note that add = sub
      for p=2.
   @param ter1 element of Field2_n
   @param ter2 element of Field2_n
  */
  public final void sub(Field2_n ter1, Field2_n ter2) {
    Check.assert(ter1.isElement() && ter2.isElement());
    for (int i = 0; i < size; i++)
      number[i] = ter1.number[i] ^ ter2.number[i];
  }

  /** this -= ter
      Adds two numbers using the xor function for intergers.
      <i>Note: This works for any extionsion field over p=2.</i>
      The result is stored in this. Please note that add = sub
      for p=2.
   @param ter element of Field2_n
  */
  public final void sub(Field2_n ter) {
    Check.assert(ter.isElement());
    for (int i = 0; i < size; i++)
      number[i] ^= ter.number[i];
  }

  /** this = fac1 * fac2
     - multiplies two numbers using finite field arithmetic in GF(2^n)
     and stores the result in this.
   @param fac1 element of GF(2^n)
   @param fac2 element of GF(2^n)
  */
  public final void mul(Field2_n fac1, Field2_n fac2) {
    Check.assert(fac1.isElement() && fac2.isElement());
    setZero();
    for (int i = degree-1; i >= 0; i--) {
       shiftL();  // compute this *= x
       if (checkBit(degree)) { // do we have to reduce?
         // we assume that only number[0] is affected by reduction
         number[0] ^= reductionTerm;
       }
       if (fac1.checkBit(i))
         add(fac2);
    }
    // make sure that all bits in number[size-1] which are
    // "too high" are cleared
    number[size-1] &= (1 << (degree % 32))-1;
    Check.assert(isElement());
  }

  /** this *= x
     assumes that there is no reduction necessary
  */
  public final void mulX() {
    Check.assert(isElement());
    shiftL();  // multiply by x
    Check.assert(isElement());
  }

  /** static register for multiplication
  */
  private static Field2_n mul_tmp = new Field2_n();

  /** this *= fac
     - multiplies two numbers using finite field arithmetic in GF(2^n)
     and stores the result in this. Uses a static variable to store the result of multiplication
   @param fac element of GF(2^n)
  */
  public final void mul(Field2_n fac) {
    Check.assert(fac.isElement());
    mul_tmp.mul(this,fac);
    this.copy(mul_tmp);
  }

  /** static registers for division
  */
  static Field2_n div_A = new Field2_n(), div_B = new Field2_n(), div_V = new Field2_n();

  /** this = whom / by in GF(2^n)
      - divides two numbers using finite field arithmetic in GF(2^n).
      Implements "Gread Divide" from Sheueling Chang Shantz.
      The result (register U) is stored in this.
      Makes use of three static registers, namely
      div_A, div_B, div_V.
   @param whom which number should be divised
   @param by   which number divides
  */
  public final void div(Field2_n whom, Field2_n by) {
    Check.assert(whom.isElement() && by.isElement());
    // init this with Y(t) (notation from Paper)
    div_A.copy(by);
    int da = div_A.cntDegree();
    Check.assert(da != -1, "Division by zero");

    // U is X(t)
    this.copy(whom);

    // init B with generating polynomial
    int db = degree; // is the degree of the modul and therefore fix
    div_B.setZero();
    div_B.number[0] = reductionTerm;
    div_B.setBit(degree);

    // V is zero
    div_V.setZero();

    while (da > 0 && db > 0) {
      //System.out.println("run da=" + da + "   db=" + db);
      Check.assert(da == div_A.cntDegree(), "da != A.cntDegree with da=" + da + " and " + div_A.cntDegree());
      Check.assert(db == div_B.cntDegree(), "db != B.cntDegree with da=" + db + " and " + div_B.cntDegree());
      if (div_A.even()) {
        div_A.shiftR();
        da--;
        if (this.even()) this.shiftR(); else {
          this.addModul();
          this.shiftR();
        }
      } else {
        if (div_B.even()) {
          div_B.shiftR();
          db--;
          if (div_V.even()) div_V.shiftR(); else {
            div_V.addModul();
            div_V.shiftR();
          }
        } else {
          if (da > db) {
            this.add(div_V);
            if (this.even()) this.shiftR(); else {
              this.addModul();
              this.shiftR();
            }
            div_A.add(div_B);
            div_A.shiftR();
            // adjust da
            while (!div_A.checkBit(da)) da--;
          } else {
            div_V.add(this);
            if (div_V.even()) div_V.shiftR(); else {
              div_V.addModul();
              div_V.shiftR();
            }
            div_B.add(div_A);
            div_B.shiftR();
            // adjust db
            while (!div_B.checkBit(db)) db--;
          }
        }
      }
    }
    // is our result in this or in div_V?
    if (db == 0) this.copy(div_V);
  }

  /** static register for division
  */
  static Field2_n div_tmp = new Field2_n();

  /** this /= fac
     - divides this by fac usind finite field arithmetic in GF(2^n)
     and stores the result in this. Uses a static variable to store the result of multiplication
   @param fac element of GF(2^n)
  */
  public final void div(Field2_n fac) {
    Check.assert(fac.isElement());
    div_tmp.div(this,fac);
    this.copy(div_tmp);
  }

  /** sets this to a random element from Field2_2.
  */
  public final void rnd(Random rnd) {
    Check.assert((degree % 32) != 0);
    for (int i = 0; i < size-1; i++)
      number[i] = rnd.nextInt();
    number[size-1] = rnd.nextInt(1 << (degree % 32));
    Check.assert(isElement());
  }

   /** returns a random position within current Field2_n
  */
  public final static int rndBit(Random rnd) {
    return rnd.nextInt(degree);
  }

  /** converts this to a string
  */
  public String stringIt() {
    String erg = "";
    for (int i = size-1; i >= 0; i--) {
      erg = erg + " " + Integer.toHexString(number[i]);
    }
    return erg;
  }

  /** Tests the functionality of the whole class.
      If there occurs any error it will be printed on stdout.
  */
  public final static void testIt () {
    System.out.println("start testing Field2_n ...");

    final int repeats = 100;
    Random rnd = new Random();

    Field2_n cur = new Field2_n();
    Field2_n cur1 = new Field2_n();
    Field2_n cur2 = new Field2_n();
    Field2_n cur3 = new Field2_n();
    Field2_n cur4 = new Field2_n();
    Field2_n cur5 = new Field2_n();
    Field2_n cur6 = new Field2_n();

    // first figure out if there are really 32 bits in a byte
    int testSize = 1;
    for (int i = 0; i < 34; i++) {
      if ((i < 32) && (testSize == 0)) System.out.println("testSize: Error for i=" + i);
      if ((i >= 32) && (testSize != 0)) System.out.println("testSize: Error for i=" + i);
      testSize *= 2;
    }

    // test if >> and << work the way they should
    int testValue = 1;
    for (int i = 0; i < 31; i++)
      testValue <<= 1;
    if (testValue >= 0) System.out.println("testShift: Error with " + testValue);
    testValue = -1;
    for (int i = 0; i < 32; i++)
      testValue >>= 1;
    if (testValue != -1) System.out.println("testShift: Error with " + testValue);

    // test setBit
    cur.setBit(3);
    if (!cur.checkBit(3)) System.out.println("Error 1.1");
    cur.clearBit(3);
    if (cur.checkBit(3)) System.out.println("Error 1.2");
    cur.setBit(66);
    if (!cur.checkBit(66)) System.out.println("Error 1.3");
    cur.clearBit(66);
    if (cur.checkBit(66)) System.out.println("Error 1.4");

    for (int i = 0; i < repeats; i++) {
      int bit = rndBit(rnd);
      cur.setBit(bit);
      if (!cur.checkBit(bit)) System.out.println("Error 1.5 at " + bit);
    }
    for (int i = 0; i < repeats; i++) {
      int bit = rndBit(rnd);
      cur.clearBit(bit);
      if (cur.checkBit(bit)) System.out.println("Error 1.6 at " + bit);
    }

    // check isElement
    cur = new Field2_n(rnd);
    cur.number[size-1] = 1 << (degree % 32);
    if (cur.isElement()) System.out.println("Error 2.1a");
    cur = new Field2_n(rnd);
    if (!cur.isElement()) System.out.println("Error 2.1b");

    // check isEqual
    for (int i = 0; i < repeats; i++) {
      cur1 = new Field2_n();
      cur2 = new Field2_n();
      int bit = rndBit(rnd);
      cur1.setBit(bit);
      if (cur1.isEqual(cur2)) System.out.println("Error 2.2");
      cur2.setBit(bit);
      if (!cur1.isEqual(cur2)) System.out.println("Error 2.3");
    }

    // test add/sub
    for (int i = 0; i < repeats; i++) {
      cur1.rnd(rnd);
      cur2.rnd(rnd);

      cur.add(cur1,cur2);
      if (!cur.isElement()) System.out.println("Error 3.1");
      for (int j = 0; j < degree; j++) {
        if (cur.checkBit(j) != cur1.checkBit(j) ^ cur2.checkBit(j)) System.out.println("Error 3.2");
      }
    }
    for (int i = 0; i < repeats; i++) {
      cur1.rnd(rnd);
      cur2.rnd(rnd);

      cur.add(cur1,cur2);
      cur1.add(cur2);
      if (!cur.isEqual(cur1)) System.out.println("Error 3.3");
    }
    for (int i = 0; i < repeats; i++) {
      cur1.rnd(rnd);
      cur2.rnd(rnd);

      cur.add(cur1,cur2);
      cur3.sub(cur1,cur2);
      if (!cur.isEqual(cur3)) System.out.println("Error 3.4");
    }
    for (int i = 0; i < repeats; i++) {
      cur1.rnd(rnd);
      cur2.rnd(rnd);

      cur.add(cur1,cur2);
      cur1.sub(cur2);
      if (!cur.isEqual(cur1)) System.out.println("Error 3.5");
    }

    // test shiftL operator
    cur = new Field2_n();
    cur.setBit(3);
    cur.shiftL();
    if (!cur.checkBit(4)) System.out.println("Error 4.1a");
    if (cur.checkBit(3)) System.out.println("Error 4.1b");

    cur = new Field2_n();
    cur.setBit(31);
    cur.shiftL();
    if (!cur.checkBit(32)) System.out.println("Error 4.2a");
    if (cur.checkBit(31)) System.out.println("Error 4.2b");

    cur = new Field2_n();
    cur.setBit(63);
    cur.shiftL();
    if (!cur.checkBit(64)) System.out.println("Error 4.3a");
    if (cur.checkBit(63)) System.out.println("Error 4.3b");
    cur1 = new Field2_n();
    cur1.setBit(64);
    if (!cur.isEqual(cur1)) System.out.println("Error 4.3c");

    cur = new Field2_n();
    cur.setBit(1);
    for (int i = 0; i < 65; i++) {
      cur.shiftL();
    }
    cur1 = new Field2_n();
    cur1.setBit(66);
    if (!cur1.isEqual(cur)) System.out.println("Error 4.4 with " + cur.stringIt() + " and " + cur1.stringIt());

    // check setZero
    cur = new Field2_n();
    cur1 = new Field2_n();
    cur1.setBit(2);
    if (cur.isEqual(cur1)) System.out.println("Error 4.7");
    cur1.setZero();
    if (!cur.isEqual(cur1)) System.out.println("Error 4.8");

    // check multiplication
    cur1 = new Field2_n();
    cur2 = new Field2_n();
    cur1.setBit(2);  // x^2
    cur2.setBit(2);  // x^2
    cur3.mul(cur1,cur2);
    cur = new Field2_n();
    cur.setBit(4);  // x^4
    if (!cur.isEqual(cur3)) System.out.println("Error 5.1 with " + cur.stringIt() + " and " + cur3.stringIt());

    cur1 = new Field2_n();
    cur2 = new Field2_n();
    cur1.setBit(12);  // x^12
    cur2.setBit(13);  // x^13
    cur3.mul(cur1,cur2);
    cur = new Field2_n();
    cur.setBit(25);  // x^25
    if (!cur.isEqual(cur3)) System.out.println("Error 5.2 with " + cur.stringIt() + " and " + cur3.stringIt());

    cur1 = new Field2_n();
    cur2 = new Field2_n();
    cur1.setBit(degree-2);   // these two together are x^(degree)
    cur2.setBit(2);
    cur3.mul(cur1,cur2);
    if (cur3.number[0] != reductionTerm) System.out.println("Error 5.3 with " + cur3.stringIt());

    cur1 = new Field2_n();
    cur2 = new Field2_n();
    cur1.setBit(degree-2);   // these two together are x^(degree+1)
    cur2.setBit(3);
    cur3.mul(cur1,cur2);
    if (cur3.number[0] != (2*reductionTerm)) System.out.println("Error 5.4 with " + cur3.stringIt());

    cur1 = new Field2_n();
    cur2 = new Field2_n();
    cur1.setBit(degree-3);   // these two together are x^(degree-1)
    cur2.setBit(2);
    cur3.mul(cur1,cur2);
    cur = new Field2_n();
    cur.setBit(degree-1);  // x^(degree-1)
    if (!cur.isEqual(cur3)) System.out.println("Error 5.5 with " + cur.stringIt() + " and " + cur3.stringIt());

    cur1 = new Field2_n();
    cur2 = new Field2_n();
    cur1.setBit(2);   // these two together are x^(degree-1)
    cur2.setBit(degree-3);
    cur3.mul(cur1,cur2);
    cur = new Field2_n();
    cur.setBit(degree-1);  // x^(degree-1)
    if (!cur.isEqual(cur3)) System.out.println("Error 5.6 with " + cur.stringIt() + " and " + cur3.stringIt());

    // test mul / add exploiting distributive law
    // a(b+c) = ab + ac
    for (int i = 0; i < repeats; i++) {
      cur1.rnd(rnd);
      cur2.rnd(rnd);
      cur3.rnd(rnd);

      // left hand side
      cur4.add(cur2,cur3);
      cur6.mul(cur1,cur4);
      // right hand side
      cur4.mul(cur1,cur2);
      cur5.mul(cur1,cur3);
      cur5.add(cur4);

      if (!cur6.isEqual(cur5)) System.out.println("Error 5.7 with " + cur1.stringIt() + ", " + cur2.stringIt() + ", "
                               + cur3.stringIt() + ", "+ cur4.stringIt() + ", "+ cur5.stringIt() + " and " + cur6.stringIt());
    }

    // test functions which were added for division
    for (int i = 0; i < repeats; i++) {
      cur.rnd(rnd);
      if (cur.even() && ((cur.number[0] % 2) == 1)) System.out.println("Error 6.1a with " + cur.stringIt());
      if (!cur.even() && ((cur.number[0] % 2) == 0)) System.out.println("Error 6.1b with " + cur.stringIt());
    }


    cur1.setZero();
    cur2.rnd(rnd);
    cur2.copy(cur1);
    if (!cur1.isEqual(cur2)) System.out.println("Error 6.2a with " + cur1.stringIt() + ", " + cur2.stringIt());

    cur1.setZero();
    cur2.rnd(rnd);
    cur1.copy(cur2);
    if (!cur1.isEqual(cur2)) System.out.println("Error 6.2b with " + cur1.stringIt() + ", " + cur2.stringIt());

    cur1.setZero();
    cur2.setZero();
    cur1.setBit(31);
    cur2.setBit(30);
    cur1.shiftR();
    if (!cur1.isEqual(cur2)) System.out.println("Error 6.3a with " + cur1.stringIt() + ", " + cur2.stringIt());

    cur1.setZero();
    cur2.setZero();
    cur1.setBit(63);
    cur2.setBit(62);
    cur1.shiftR();
    if (!cur1.isEqual(cur2)) System.out.println("Error 6.3b with " + cur1.stringIt() + ", " + cur2.stringIt());

    for (int i = 1; i < degree; i++) {
      cur1.setZero();
      cur2.setZero();
      cur1.setBit(i);
      cur2.setBit(i-1);
      cur1.shiftR();
      if (!cur1.isEqual(cur2)) System.out.println("Error 6.3c at " + i + " with " + cur1.stringIt() + ", " + cur2.stringIt());
    }

    for (int i = 0; i < repeats; i++) {
      cur1.rnd(rnd);
      cur2.copy(cur1);
      cur1.shiftL();
      cur1.shiftR();
      if (!cur1.isEqual(cur2)) System.out.println("Error 6.3d at " + i + " with " + cur1.stringIt() + ", " + cur2.stringIt());
    }

    cur2.setZero();
    cur2.setBit(1);  // set x, i.e. polynom with _one_ single term
    for (int i = 0; i < repeats; i++) {
      cur1.rnd(rnd);
      cur1.clearBit(degree-1);
      cur3.mul(cur1,cur2);
      cur1.shiftL();      // exploits the fact that multiplication with x is the same as a left shift _if_ we don't need reduction
      if (!cur1.isEqual(cur3)) System.out.println("Error 6.3e at " + i + " with " + cur1.stringIt() + ", " + cur3.stringIt());
    }

    cur1.setZero();
    if (cur1.cntDegree() != -1) System.out.println("Error 6.4a with " + cur1.cntDegree());
    for (int i = 0; i < degree; i++) {
      cur1.setZero();
      cur1.setBit(i);
      if (cur1.cntDegree() != i) System.out.println("Error 6.4b at " + i + " with " + cur1.cntDegree() + " and " + cur1.stringIt());
    }

    // test addModul
    cur.setZero();
    cur.addModul();
    if (!cur.checkBit(degree)) System.out.println("Error 6.5a with " + cur.stringIt());
    if (cur.number[0] != reductionTerm) System.out.println("Error 6.5b with " + cur.stringIt());
    for (int i = 1; i < size-2; i++)
      if (cur.number[i] != 0) System.out.println("Error 6.5c with " + cur.stringIt());

    // test div
    cur1.setZero();
    cur2.setZero();
    cur1.setBit(3);
    cur2.setBit(3);
    cur3.div(cur1,cur2);
    cur.setZero();
    cur.setBit(0);
    if (!cur.isEqual(cur3)) System.out.println("Error 6.6a with " + cur.stringIt() + " and " + cur3.stringIt());

    cur1.setZero();
    cur2.setZero();
    cur1.setBit(3); cur1.setBit(0);
    cur2.setBit(3); cur2.setBit(0);
    cur3.div(cur1,cur2);
    cur.setZero();
    cur.setBit(0);
    if (!cur.isEqual(cur3)) System.out.println("Error 6.6a with " + cur.stringIt() + " and " + cur3.stringIt());

    // test division together with multiplication
    for (int i = 0; i < repeats; i++) {
      cur1.rnd(rnd);
      cur2.rnd(rnd);
      cur.mul(cur1,cur2);
      cur3.div(cur,cur1);
      if (!cur3.isEqual(cur2)) System.out.println("Error 6.7a with " + cur.stringIt() + ", "
                               + cur1.stringIt() + ", " + cur2.stringIt() + " and " + cur3.stringIt());
      cur3.div(cur,cur2);
      if (!cur3.isEqual(cur1)) System.out.println("Error 6.7b with " + cur.stringIt() + ", "
                               + cur1.stringIt() + ", " + cur2.stringIt() + " and " + cur3.stringIt());
    }

    // additional functionality
    cur.setZero();
    if (!cur.isZero()) System.out.println("Error 7.1a with " + cur.stringIt());
    cur.setBit(0);
    if (cur.isZero()) System.out.println("Error 7.1b with " + cur.stringIt());
    cur.setZero();
    cur.setBit(degree-1);
    if (cur.isZero()) System.out.println("Error 7.1c with " + cur.stringIt());

    for (int i = 0; i < repeats; i++) {
      cur1.rnd(rnd);
      cur2.rnd(rnd);
      cur3.mul(cur1,cur2);
      cur1.mul(cur2);
      if (!cur1.isEqual(cur3)) System.out.println("Error 7.2a with " + cur1.stringIt() + ", " + cur2.stringIt()
                               + " and " + cur3.stringIt());

      cur1.rnd(rnd);
      cur2.rnd(rnd);
      cur3.div(cur1,cur2);
      cur1.div(cur2);
      if (!cur1.isEqual(cur3)) System.out.println("Error 7.2b with " + cur1.stringIt() + ", " + cur2.stringIt()
                               + " and " + cur3.stringIt());
    }

    cur.setOne();
    cur1.setZero();
    if (cur.isEqual(cur1)) System.out.println("Error 7.3a with " + cur.stringIt() + ", " + cur1.stringIt());
    cur1.setBit(0);
    if (!cur.isEqual(cur1)) System.out.println("Error 7.3b with " + cur.stringIt() + ", " + cur1.stringIt());

    // test set and get
    int [] bitVec = new int[degree];
    cur1.setOne();
    cur2.setZero();
    cur1.getElement(bitVec);
    cur2.setElement(bitVec);
    if (!cur1.isEqual(cur2)) System.out.println("Error 8.1a");

    cur1.rnd(rnd);
    cur2.setZero();
    cur1.getElement(bitVec);
    cur2.setElement(bitVec);
    if (!cur1.isEqual(cur2)) System.out.println("Error 8.1b");

    cur1.setZero();
    cur2.rnd(rnd);
    cur1.getElement(bitVec);
    cur2.setElement(bitVec);
    if (!cur1.isEqual(cur2)) System.out.println("Error 8.1c");

    for (int i = 0; i < repeats; i++) {
      cur1.rnd(rnd);
      cur2.rnd(rnd);
      cur1.getElement(bitVec);
      cur2.setElement(bitVec);
      if (!cur1.isEqual(cur2)) System.out.println("Error 8.1b");
    }

    // test leftShiftMod
    cur1 = new Field2_n(2);  // x
    for (int i = 0; i < repeats; i++) {
      cur2.rnd(rnd);
      cur3.mul(cur1,cur2);  // compute cur2*x using multiplication
      cur2.shiftLMod();     // compute the same using shiftLMod
      if (!cur2.isEqual(cur3)) System.out.println("Error 8.2 with " + cur2.stringIt() + " and " + cur3.stringIt());
    }

    // test add one
    cur1 = new Field2_n(1);
    for (int i = 0; i < repeats; i++) {
      cur2.rnd(rnd);
      cur3.add(cur2,cur1);
      cur2.addOne();
      if (!cur2.isEqual(cur3)) System.out.println("Error 8.3 with " + cur2.stringIt() + " and " + cur3.stringIt());
    }

    // test xorBit
    for (int i = 0; i < repeats; i++) {
      cur1.rnd(rnd);
      cur2.copy(cur1);
      int pos = rndBit(rnd);
      cur2.xorBit(pos);
      if (cur1.checkBit(pos) & cur2.checkBit(pos)) System.out.println("Error 8.4 with " + cur1.stringIt() + " and " + cur2.stringIt());
    }

    System.out.println("...finished testing.");
  }

  /** Tests everything in this class
      @see #testIt
    */
  public static void main(String[] args) {
    testIt();
  }
}
