package HFE;

import java.util.*;
//-----------------------------------------------------------------------------
// Project:     Implementing HFE at UCC as Project CS4095
// Supervisors: Simon Foley (Computer Science) and Patrick Fitzpatrick (Maths)
//-----------------------------------------------------------------------------

/**
<p>
This class encapsulates the operations for a finite field
with 4 elements. The current field is generated by the
polynomial f(x) = x^2 + x + 1 <br>
Most operations are performed using a look-up table,
add and sub are using xor for integers.
<p>
@author Christopher Wolf (hfe@Christopher-Wolf.de)
@version 0.1
*/

//Copyright (C) 2002    Christopher Wolf
//
//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License (file LICENCE) for more details.


final class Field4 {

  /** number of elements in this field
   @serial constant for users of that class
  */
  public final static int elements = 4;


  /** Checks if a given integer is within the range of
      a Field4 element.
    @return is elem in the range of an Field4 element
  */
  public static boolean isElement(int elem) {
    return (elem < elements) && (elem >= 0);
  }

  /** Checks if a given vector consists only
      of Field4 elements.
    @return true iff all elements of the vector
            are Field4 elements.
  */
  public static boolean isElement(int[] elems) {
    for (int i = 0; i < elems.length; i++)
      if (!isElement(elems[i])) return false;
    return true;
  }

  /** Checks if a given matrix consists only
      of Field4 elements.
    @return true iff all elements of the matrix
            are Field4 elements.
  */
  public static boolean isElement(int[][] elems) {
    // uses the vector version of isElement
    for (int i = 0; i < elems.length; i++)
      if (!isElement(elems[i])) return false;
    return true;
  }

  /** Adds two numbers using the xor function for intergers.
      <i>Note: This works for any extionsion field over p=2.</i>
   @param ter1 element of F_4
   @param ter2 element of F_4
   @return     (ter1 + ter2) over F_4
  */
  public final static int add(int ter1, int ter2) {
    Check.assert(isElement(ter1) && isElement(ter2));
    return ter1 ^ ter2;
  }


  /** Subtracts two numbers using the xor function for intergers.
      <i>Note: This works for any extionsion field over p=2</i>
   @param ter1 element of F_4
   @param ter2 element of F_4
   @return     (ter1 - ter2) over F_4
  */
  public final static int sub(int ter1, int ter2) {
    Check.assert(isElement(ter1) && isElement(ter2));
    return ter1 ^ ter2;
  }

  /** look-up table for multiplication
   @serial constant for internal use
  */
  private static final int[][] mulTab = {
    {0, 0, 0, 0},
    {0, 1, 2, 3},
    {0, 2, 3, 1},
    {0, 3, 1, 2}
  };


  /** Multiplies two numbers using finite field arithmetic in F_4.
      For the sake of speed I store the results in an lookup table.
   @param fac1 element of F_4
   @param fac2 element of F_4
   @return     (fac1 * fac2) over F_4
  */
  public final static int mul(int fac1, int fac2) {
    Check.assert(isElement(fac1) && isElement(fac2));
    return mulTab[fac1][fac2];
  }


  // look-up table for division
  /** look-up table for multiplication
   @serial constant for internal use
  */
  private static final int[][] divTab = {
    {0, 0, 0, 0},
    {0, 1, 3, 2},
    {0, 2, 1, 3},
    {0, 3, 2, 1}
  };


  /** Divides two numbers using finite field arithmetic in F_4.
      For the sake of speed I store the results in an lookup table
   @param whom which number should be divised
   @param by   which number divides
   @return     (whom/by) over F_4
  */
  public final static int div(int whom, int by) {
    Check.assert(isElement(whom) && isElement(by));
    Check.assert(by != 0);
    return divTab[whom][by];
  }

  /** returns a random element from Field4.
  */
  public final static int rnd(Random rnd) {
    return rnd.nextInt(elements);
  }


  /** Tests the functionality of the whole class.
      If there occurs any error it will be printed on stdout.
  */
  public final static void testIt () {
    System.out.println("start testing Field4 ...");

    // test add
    if (add(1,1) != 0) System.out.println("Error 1.1");
    if (add(2,2) != 0) System.out.println("Error 1.2");
    if (add(3,3) != 0) System.out.println("Error 1.3");
    if (add(2,3) != 1) System.out.println("Error 1.4");
    if (add(1,3) != 2) System.out.println("Error 1.5");

    // test mul
    if (mul(1,3) != 3) System.out.println("Error 2.1");
    if (mul(2,3) != 1) System.out.println("Error 2.2");
    if (mul(1,2) != 2) System.out.println("Error 2.3");
    if (mul(2,2) != 3) System.out.println("Error 2.4");

    // test div
    if (div(1,1) != 1) System.out.println("Error 3.1");
    if (div(1,2) != 3) System.out.println("Error 3.2");
    if (div(1,3) != 2) System.out.println("Error 3.3");
    if (div(2,1) != 2) System.out.println("Error 3.4");
    if (div(2,2) != 1) System.out.println("Error 3.5");
    if (div(2,3) != 3) System.out.println("Error 3.6");
    if (div(3,1) != 3) System.out.println("Error 3.7");
    if (div(3,2) != 2) System.out.println("Error 3.8");
    if (div(3,3) != 1) System.out.println("Error 3.9");

    // test the symmetry of addition an multiplication
    for (int a = 0; a < elements; a++) {
      for (int b = 0; b < elements; b++) {
        if (add(a,b) != add(b,a)) System.out.println("Error 4.1 with " + a + " " + b);
        if (mul(a,b) != mul(b,a)) System.out.println("Error 4.2 with " + a + " " + b);
      }
    }

    // test add and mul using the distributive law
    // a(b+c) = ab + ac
    for (int a = 0; a < elements; a++) {
      for (int b = 0; b < elements; b++) {
        for (int c = 0; c < elements; c++) {
          if (mul(a,add(b,c)) != add(mul(a,b),mul(a,c))) System.out.println("Error 4.3 with " + a + " " + b + " " + c);
        }
      }
    }

    // test mul and div using the property
    // (a*b)/b = a  with b != 0
    for (int a = 0; a < elements; a++) {
      for (int b = 1; b < elements; b++) {
        if (a != div(mul(a,b),b)) System.out.println("Error 4.4 with " + a + " " + b);
      }
    }

    // test isElement
    if (isElement(5)) System.out.println("Error 5.1");
    if (!isElement(1)) System.out.println("Error 5.2");

    System.out.println("...finished testing.");
  }

  /** Tests everything in this class
      @see #testIt
    */
  public static void main(String[] args) {
    testIt();
  }
}
