package HFE;

import java.lang.*;
import java.io.*;
import java.util.*;
import java.security.*;
import sun.misc.*;
import HFE.*;

//-----------------------------------------------------------------------------
// Project:     Implementing HFE at UCC as Project CS4095
// Supervisors: Simon Foley (Computer Science) and Patrick Fitzpatrick (Maths)
//-----------------------------------------------------------------------------

/**
<p>
This class is the "final version" of HFE which works on
n=67,129,257 bits over the field GF(2) ({@link HFE.Field2}) which
extents to GF(2^n) ({@link HFE.Field2_n}).
<p>
@author Christopher Wolf (hfe@Christopher-Wolf.de)
@version 0.1
*/

//Copyright (C) 2002    Christopher Wolf
//
//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License (file LICENCE) for more details.


class HFE2_n implements CheckRoot {

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// Central switches to select E = F^67, F^129 or F^257

  /** Central switch to change generating polynomial.
      Saves only the powers as the coefficients must be 1.
      The same polynomial (but in a different format) is stored in
      {@link #generatingPolynomialAsMaxPoly}
  */
  final static int[] generatingPolynomial = new int[] {67, 5, 2, 1, 0}; // f(x) = x^67 + x^5 + x^2 + x + 1
  //final static int[] generatingPolynomial = new int[] {129, 5,0};       // f(x) = x^129 + x^5 + 1
  //final static int[] generatingPolynomial = new int[] {257,12,1};       // f(x) = x^257 + x^12 + 1

  /** Central switch to change the reduction term,
     i.e. the part of the polynomial with is not the degree and not zero
  */
  final static int reductionTerm = 39;   //  f(x) = x^67 + x^5 + x^2 + x + 1 (i.e. 39 = 32+4+2+1 as reduction factor)
  //final static int reductionTerm = 33;   // f(x) = x^129 + x^5 + 1 (i.e. 33 = 32+1 as reduction factor)
  //final static int reductionTerm = 4097; // f(x) = x^257 + x^12 + 1 (i.e. 4097 = 4096+1 as reduction factor)

  /** Holds the degree of the irreducible polynomial and thereby also
      the degree of the extension field
  */
  final static int extensionDegree = 67;
  //final static int extensionDegree = 129;
  //final static int extensionDegree = 257;

//!!!!!!!!!!!!!!!!!!!!!!!!!!!

  /** Polynomial set in variable {@link #generatingPolynomial}
      but as a Poly2Max polynomial.
      Does not need to be changed for different {@link Field2_n}.
  */
  final static Poly2Max generatingPolynomialAsMaxPoly = new Poly2Max(generatingPolynomial);

  /** Holds the number of bits for the hash value
  */
  final static int hashLen = 24;

  /** Holds the number of public equations which are replaced by random equations.
      Must be 0 for a signature scheme for internal reasons. In this case "padBits"
      replace paramK. paramK should be less than 10 for a decryption scheme as
      increasing paramK by one means to double the time for decryption.
  */
  final static int paramK = 4;

  /** Holds the number of bits for an encrypted message, i.e.
  {@link #hashLen} + {@link #extensionDegree}.
  */
  final static int messageLength = hashLen + extensionDegree;

  /** Holds the number of bits for a message in terms of a signature.
      This is smaller than messageLength as we need some "extra bits" for
      random padding. padBits can be very large in a signature scheme and
      should be bigger than 10 - otherwise the change to find a root is
      too poor.
  */
  final static int padBits = 20;

  /** Holds the number of bits for a message in terms of a signature.
      This is smaller than messageLength as we need some "extra bits" for
      random padding.
  */
  final static int messageSigLength = extensionDegree - padBits;

  /** Holds the number of bits for a signature.

  */
  final static int signatureLength = extensionDegree;

  /** Signals if we are in decryption (1) or signature generation mode (-1).
      Is used in method {@link #checkRoot} to signal the correct result
      to class {@link RootFinding}.
  */
  int modeDecSig = 0;

  /** Stores the private polynomial over GF(2^n). Is used for decryption.
    @serial Private Key
  */
  private Poly2_nMax privatePoly;

  /** Stores the public key in an array of polynomials.
      Is used for encryption
    @serial Poly2MultiSquare
  */
  private Poly2MultiSquare[] publicPoly;

  /** Transformation S in HFE
  */
  private Transform2 transS;

  /** Transformation T in HFE
  */
  private Transform2 transT;

  /** Root finding algorithm
  */
  private RootFinding rootFinder = new RootFinding();

  /** Message as a vector. For the sake of speed is it a static.
    @serial vector
  */
  private static int[] textVec = new int[extensionDegree];

  /** Hash of message as a vector. For the sake of speed is it a static.
      Is used in {@link #genSign}
    @serial vector
  */
  private static int[] sigTmpVec = new int[extensionDegree];

  /** Message as a vector. For the sake of speed is it a static.
    @serial vector
  */
  private static int[] resVec = new int[extensionDegree];

  /** Message as a vector. For the sake of speed is it a static.
      Is set in {@link #checkRoot} and used in {@link #decrypt}
      as well as {@link #genSign}.
    @serial vector
  */
  private static int[] rootVec = new int[extensionDegree];

  /** Message in a special format as a vector. For the sake of speed is it a static.
    @serial vector
  */
  private static int[] valueVec = new int[extensionDegree+1];

  /** Cipher as a vector. For the sake of speed static.
  */
  private static int[] codeVec = new int[Poly2MultiSquare.tau(extensionDegree)];

  /** Quadratic polynomials for computing hash of message m
  */
  Poly2MultiSquare[] hashPoly = new Poly2MultiSquare[hashLen];

  /** hash value of the current message
  */
  int[] hashValue = new int[hashLen];

  /** Temporary register [element of GF(2^n)]. Static for the sake of speed
  */
  Field2_n tmpElem = new Field2_n();

  /** Random number generator for signature generation
  */
  Random rnd = new Random();


  /** Initializes HFE2_n by generating the
      public and the private key from the given seed. Also generates the
      message digest.
     @param seed Seed for the cryptographic random number generator.
  */
  HFE2_n (byte[] seed) {
    generateKeys(seed);
    setHashPoly();
  }


  /** Generates a private key for HFE.
      This polynomial has the form
        f(x) = x^33 + ax^32 + bx^24 + cx^20 +
        dx^18 + ex^17 + fx^16 + gx^12 + hx^10 +
        ix^9 + jx^8 + kx^6 + lx^5 + mx^4 + nx^3 +
        ox^2 + px^1 where a..p are from GF(2^n).
     @param  rndGen Secure random number generator
     @return private polynomial
  */
  private static Poly2_nMax genPolyPrivate(SecureRandom rndGen) {
    Poly2_nMax res = new Poly2_nMax();
    // set the leading coefficient to be 1
    res.setCoeff(new Field2_n(1), 33);
    // the rest are random elements
    res.setCoeff(rndGen,32);
    res.setCoeff(rndGen,24);
    res.setCoeff(rndGen,20);
    res.setCoeff(rndGen,18);
    res.setCoeff(rndGen,17);
    res.setCoeff(rndGen,16);
    res.setCoeff(rndGen,12);
    res.setCoeff(rndGen,10);
    res.setCoeff(rndGen,9);
    res.setCoeff(rndGen,8);
    res.setCoeff(rndGen,6);
    res.setCoeff(rndGen,5);
    res.setCoeff(rndGen,4);
    res.setCoeff(rndGen,3);
    res.setCoeff(rndGen,2);
    res.setCoeff(rndGen,1);
    return res;
  }

  /** Generats the public polynomials from the private polynomial
     @param  rndGen Secure random number generator
     @return public polynomials
  */
  private Poly2MultiSquare[] genPolyPublic(SecureRandom rndGen) {
    int[] coeffVec = new int[extensionDegree];
    Field2_n coeff = new Field2_n();
    Multi2Power powers = new Multi2Power(generatingPolynomialAsMaxPoly, transS.getTransform(), 5);
    Multi2Mul calc = new Multi2Mul(generatingPolynomialAsMaxPoly);

    Poly2MultiSquare[] res = new Poly2MultiSquare[extensionDegree];
    for (int i = 0; i < extensionDegree; i++)
      res[i] = new Poly2MultiSquare(extensionDegree);

    Poly2MultiSquare[] resMulti = new Poly2MultiSquare[extensionDegree];
    for (int i = 0; i < extensionDegree; i++)
      resMulti[i] = new Poly2MultiSquare(extensionDegree);
    Poly2MultiLinear[] resLin = new Poly2MultiLinear[extensionDegree];
    for (int i = 0; i < extensionDegree; i++)
      resLin[i] = new Poly2MultiLinear(extensionDegree);

    // compute the square terms
    int[] squareTerms = new int []{5,0, 4,3, 4,2, 4,1, 4,0, 3,2, 3,1, 3,0, 2,1, 2,0, 1,0};
    //int[] squareTerms = new int[]{5,0, 4,3};
    for (int i = 0; i < squareTerms.length; i+= 2) {
      int i1 = squareTerms[i];
      int i2 = squareTerms[i+1];
      int pos = (1 << i1) + (1 << i2);
      privatePoly.getCoeff(coeff,pos);
      coeff.getElement(coeffVec);
      calc.mul(resMulti,coeffVec,powers.givePower(i1),powers.givePower(i2));
      for (int j = 0; j < extensionDegree; j++)
        res[j].add(resMulti[j]);
    }

    // compute the linear terms
    int[] linearTerms = new int []{5, 4, 3, 2, 1, 0};
    //int[] linearTerms = new int []{1, 0};
    for (int i = 0; i < linearTerms.length; i++) {
      int i1 = linearTerms[i];
      int pos = 1 << i1;
      privatePoly.getCoeff(coeff,pos);
      coeff.getElement(coeffVec);
      calc.mul(resLin,coeffVec,powers.givePower(i1));
      for (int j = 0; j < extensionDegree; j++)
        res[j].add(resLin[j]);
    }

    // take paramK into account and hence overwrite some polynomials by random
    for (int i = 0; i < paramK; i++)
      res[i].rnd(rndGen);

    // overwrite or result vector with the T transformation
    res = transT.transform(res);

    return res; // after "a few steps" we finally got our public polynomials
  }


  /** Generates the public and the private key from
      a given seed and stores them in the object's fields.
      For cryptographic purposes, the seed is too short.
     @param seed Seed for the random number generator. Must be "quite long" to be secure.
  */
  void generateKeys(byte [] seed) {
    SecureRandom rndGen = new SecureRandom();
    rndGen.setSeed(seed);
    privatePoly = genPolyPrivate(rndGen);
    transS = new Transform2(rndGen);
    transT = new Transform2(rndGen);
    publicPoly = genPolyPublic(rndGen);
  }

  /** Sets the hash polynomials using a constant seed.
  */
  private void setHashPoly() {
    Random rnd = new Random(7577);
    for (int i = 0; i < hashLen; i++)
      hashPoly[i] = new Poly2MultiSquare(extensionDegree,rnd);
  }

  /** Encryptes a given, n bit plain text.
      <br>
      Uses <b>private</b> key.
     @param plain plaintext as vector of n bits
     @param res result as a vector of n+16 bits
  */
  private void encryptPrivate(int[] plain, int [] res) {
    //Check.assert(plain.length == extensionDegree, "plaintext too big " + plain.length);
    //Check.assert(res.length == messageLength, "result too big " + res.length);

    Field2_n tmpVal = new Field2_n();
    Field2_n resVal = new Field2_n();
    int[] resVec2 = new int[extensionDegree];

    // prepare for multivariante polynomials
    System.arraycopy(plain,0,valueVec,1,extensionDegree);
    Poly4MultiSquare.getValues(valueVec,codeVec);

    transS.transform(plain,resVec);
    tmpVal.setElement(resVec);
    resVal = privatePoly.applyPoly(tmpVal);
    resVal.getElement(resVec);
    transT.transform(resVec,resVec2);
    for (int i = 0; i < extensionDegree; i++)
      res[i] = resVec2[i];

    // prepare for multivariante polynomials
    System.arraycopy(plain,0,valueVec,1,extensionDegree);
    Poly4MultiSquare.getValues(valueVec,codeVec);
    // after encrypting message plain we compute its hash
    for (int i = 0; i < hashLen; i++)
      res[extensionDegree+i] = hashPoly[i].apply(codeVec);
  }

  /** Encryptes a given, n bit plain text.
      <br>
      Uses public polynomials
     @param plaintext as vector of n bits
     @param res result as a vector of n+16 bits
  */
  public void encrypt(int[] plain, int [] res) {
    //Check.assert(plain.length == extensionDegree, "plaintext has wrong length " + plain.length);
    //Check.assert(res.length == messageLength, "result has wrong length " + res.length);

    System.arraycopy(plain,0,valueVec,1,extensionDegree);
    Poly4MultiSquare.getValues(valueVec,codeVec);
    for (int i = 0; i < extensionDegree; i++)
      res[i] = publicPoly[i].apply(codeVec);

    // here we compute the hash of message plain
    for (int i = 0; i < hashLen; i++)
      res[extensionDegree+i] = hashPoly[i].apply(codeVec);
  }

  /** Decrypts a given, n bit cipher text.
      <br>
      Uses private key (S,P,T)
     @param ciphertext as bit-vector
     @param res result as a bit-vector
  */
  public void decrypt(int[] cipher, int[] res) {
    //Check.assert(cipher.length == messageLength, "ciphertext has wrong length " + cipher.length);
    //Check.assert(res.length == extensionDegree, "result has wrong length " + res.length);
    modeDecSig = 1; // we are in decryption mode

    // take part which is _not_ hash
    System.arraycopy(cipher,0,textVec,0,extensionDegree);
    // take part which _is_ hash
    System.arraycopy(cipher,extensionDegree,hashValue,0,hashLen);
    // reverse transformation T - this changes textVec!
    transT.untransform(textVec,resVec);

    // try all values which are possible in respect of paramK, start with 0..0
    for (int i = 0; i < paramK; i++)
      resVec[i] = 0;
    for (int i = 0; i < (1 << paramK); i++) {
      // use root finding algorithm on element of GF(2^n)
      tmpElem.setElement(resVec);
      privatePoly.setCoeff(tmpElem,0);  // exploit the fact that coeff0 of privatePoly is zero
      if (rootFinder.findRoots(privatePoly,this) == true) {
        // copy to res-vector which was set in checkRoot
        System.arraycopy(rootVec,0,res,0,extensionDegree);
        modeDecSig = 0; // we are no longer in any mode!
        return;  // we are done!
      }
      // check next root!
      boolean carry = true;
      int j = 0;
      while (carry) {
        resVec[j] ^= 1; // add one in GF(2)
        carry = (resVec[j] == 0);
        j++;
      }
    }
    // if we ever reach this part we are in trouble!
    // it's not the correct exception be we don't have to declare this one...
    throw new IllegalArgumentException("Did not find a suitable message in decrypt.");
  }

  /** Generates signature for given message using (S,P,T).
    @param message to sign (in general: hash of this message with {@link #messageSigLength} bits)
    @param res signature for this message with {@link #signatureLength} bits
  */
  public void genSign(int[] message, int[] res) {
    //Check.assert(message.length == messageSigLength, "message has wrong length: " + message.length);
    //Check.assert(res.length == signatureLength, "result has wrong length: " + res.length);
    //Check.assert(paramK == 0, "paramK is too big: " + paramK);

    // we are in signature generation mode
    modeDecSig = -1;

    // copy message and set the padding bits by random
    System.arraycopy(message,0,sigTmpVec,0,messageSigLength);
    for (int i = messageSigLength; i < signatureLength; i++)
      sigTmpVec[i] = Field2.rnd(rnd);

    for (int i = 0; i < (1 << padBits); i++) {
      System.arraycopy(sigTmpVec,0,rootVec,0,signatureLength);  // bad style, but anyway: rootVec it not needed in this step!
      transT.untransform(rootVec,resVec);

      // use root finding algorithm on element of GF(2^n)
      tmpElem.setElement(resVec);
      privatePoly.setCoeff(tmpElem,0);  // exploit the fact that coeff0 of privatePoly is zero
      if (rootFinder.findRoots(privatePoly,this)) {
        // copy to res-vector which was set in checkRoot
        System.arraycopy(rootVec,0,res,0,signatureLength);
        modeDecSig = 0; // we are no longer in any mode!
        return;
      }
      // check next padding
      boolean carry = true;
      int j = messageSigLength;
      while (carry && (j < signatureLength)) {
        sigTmpVec[j] ^= 1; // add one in GF(2)
        carry = (sigTmpVec[j] == 0);
        j++;
      }
    }
    // if we ever reach this part we are in trouble!
    // it's not the correct exception be we don't have to declare this one...
    throw new IllegalArgumentException("Did not find a suitable root for signature.");
  }

  /** Validates if given signature <code>sig</code> is valid or not.
    @param message which was signed (in general: hash of this message with {@link #messageSigLength} bits) using method {@link #genSign}
    @param sig signature for this message with {@link #signatureLength} bits
    @return true iff sig is valid for current public key
  */
  public boolean valSign(int[] message, int[] sig) {
    //Check.assert(message.length == messageSigLength, "message has wrong length: " + message.length);
    //Check.assert(sig.length == signatureLength, "signature has wrong length: " + sig.length);
    //Check.assert(paramK == 0, "paramK is too big: " + paramK);

    System.arraycopy(sig,0,valueVec,1,signatureLength);
    Poly4MultiSquare.getValues(valueVec,codeVec);
    for (int i = 0; i < messageSigLength; i++)
      if (message[i] != publicPoly[i].apply(codeVec)) return false;

    return true;
  }

  /** Returns always true in signature mode (see variable {@link #modeDecSig}).
      In decryption mode checks if the root has the correct hash value after
      transformation S is applied. If so, returns true.
     @return true iff the root is correct.
  */
  public boolean checkRoot(Field2_n root) {
    //Check.assert((modeDecSig == 1) || (modeDecSig == -1), "modeDecSig has wrong value, namely " + modeDecSig);
    // decryption mode
    if (modeDecSig == 1) {
      // reverse transformation S from root
      root.getElement(textVec);
      transS.untransform(textVec,rootVec);

      // compute hash of message
      System.arraycopy(rootVec,0,valueVec,1,extensionDegree);
      Poly4MultiSquare.getValues(valueVec,codeVec);
      for (int i = 0; i < hashLen; i++)
        if (hashValue[i] != hashPoly[i].apply(codeVec)) return false;  // wrong message...
      // be paranoic: We don't want to get a second root!
      modeDecSig = 0;
      return true; // we got the correct one!
    }
    // signature mode: we are happy with _any_ root!
    if (modeDecSig == -1) {
      root.getElement(textVec);
      transS.untransform(textVec,rootVec);
      // be paranoic: We don't want to get a second root!
      modeDecSig = 0;
      return true;
    }
    // we are in trouble if we get to this place.
    // it's not the correct exception be we don't have to declare this one...
    throw new IllegalArgumentException("modeDecSig is wrong in HFE2_n.checkRoot: " + modeDecSig);
  }


  /** Tests the functionality of the whole class.
      If there occurs any error it will be printed on stdout.
  */
  public static void testIt() {
    System.out.println("Start testing HFE2_n...");
System.out.println("System parameter: ");
System.out.println("  Extension degree: " + extensionDegree);
System.out.println("  paramK: " + paramK);

    HFE2_n testObj = new HFE2_n(new byte[] {1, 2, 3, 4, 5, 6, 7, 8});

    Random rnd = new Random();
    Field2_n rndElem = new Field2_n(rnd);
    int[] plain = new int[extensionDegree];
    int[] sig = new int[messageSigLength];
    int[] dec = new int[extensionDegree];
    int[] enc = new int[messageLength];

    for (int i = 0; i < 100; i++) {
System.out.println("Run " + i);
      rndElem.rnd(rnd);
      rndElem.getElement(plain);

      for (int k = 0; k < sig.length; k++)
        sig[k] = Field2.rnd(rnd);

      testObj.encrypt(plain,enc);
      testObj.decrypt(enc,dec);
      //if (testObj.valSign(sig,dec)) System.out.println("Signature valid!");
      //testObj.genSign(sig,dec);
      //if (!testObj.valSign(sig,dec)) System.out.println("Signature not valid!");

      for (int j = 0; j < extensionDegree; j++)
        if (plain[j] != dec[j]) System.out.println("plain != dec at " + i);
    }
    System.out.println("...ended testing.");
  }

  public final static void speedIt() {
    //Check.assert(false);  // make sure we do not have any assertion checkers
    long overallTime = System.currentTimeMillis();

    System.out.println("Start speed measurement HFE2_n...");
    System.out.print("System parameter: ");
    System.out.print(" Extension degree: " + extensionDegree);
    System.out.println(" paramK: " + paramK);
    Random rnd = new Random();
    byte [] seed = new byte[] {1, 2, 3, 4, 5, 6, 7, 8};
    HFE2_n mySpeed = new HFE2_n(seed);
    long curTime;
    int paramM;
    int repeats;
    int[] plainVec = new int[extensionDegree];
    int[] encVec = new int[messageLength];

    /*
    System.gc();
    // key generation
    paramM = 102;
    System.out.println("n=" + paramM);
    System.out.println("r=1");
    System.out.println("Key Generation [E=GF(2^" + extensionDegree +  "), paramK=" + paramK + "]");
    for (int i = 0; i < paramM; i++) {
      Integer tmpInt = new Integer(i);
      seed = new byte[] {1, 2, 3, 4, 5, 6, 7, 8, 9, tmpInt.byteValue()};
      curTime = System.currentTimeMillis();
      mySpeed.generateKeys(seed);
      curTime = System.currentTimeMillis() - curTime;
      System.out.println(curTime + " ms");
    }
    System.out.println();  /**/

    System.gc();
    // encryption
    paramM = 102;
    repeats = 100;
    System.out.println("n=" + paramM);
    System.out.println("r=" + repeats);
    System.out.println("Encryption [E=GF(2^" + extensionDegree +  "), paramK=" + paramK + "]");
    for (int i = 0; i < paramM; i++) {
      for (int j = 0; j < plainVec.length; j++)
        plainVec[j] = Field2.rnd(rnd);
      curTime = System.currentTimeMillis();
      for (int j = 0; j < repeats; j++) {
        mySpeed.encrypt(plainVec,encVec);
      }
      curTime = System.currentTimeMillis() - curTime;
      System.out.println(curTime + " ms");
    }
    System.out.println(); /**/

    System.gc();
    // decryption
    paramM = 102;
    int[][] manyEnc = new int[paramM][messageLength];
    // generate messages for decryption
    for (int i = 0; i < paramM; i++) {
      manyEnc[i] = new int[messageLength];
      for (int j = 0; j < plainVec.length; j++)
        plainVec[j] = Field2.rnd(rnd);
      mySpeed.encrypt(plainVec,manyEnc[i]);
    }
    System.out.println("n=" + paramM);
    System.out.println("r=1");
    System.out.println("Decryption [E=GF(2^" + extensionDegree +  "), paramK=" + paramK + "]");
    for (int i = 0; i < paramM; i++) {
      curTime = System.currentTimeMillis();
      mySpeed.decrypt(manyEnc[i],plainVec);
      curTime = System.currentTimeMillis() - curTime;
      System.out.println(curTime + " ms");
    }
    System.out.println();  /**/

    System.gc();
    if (paramK != 0) {
      System.out.println("n=0");  // finis!
      System.out.println();
      long overall2 = System.currentTimeMillis() - overallTime;
      System.out.println("Overall time: " + overall2 + " ms (about " + (((float)overall2) / (1000*60))+ " min)");
      return;
    }
    // signature
    paramM = 102;
    System.out.println("n=" + paramM);
    System.out.println("r=1");
    System.out.println("Signature [E=GF(2^" + extensionDegree +  "), paramK=" + paramK + "]");
    int[][] manySig = new int[paramM][signatureLength];
    int[][] sigVec = new int[paramM][messageSigLength];
    for (int i = 0; i < paramM; i++) {
      sigVec[i] = new int[messageSigLength];
      for (int j = 0; j < sigVec[i].length; j++)
        sigVec[i][j] = Field2.rnd(rnd);
      curTime = System.currentTimeMillis();
      mySpeed.genSign(sigVec[i],manySig[i]);
      curTime = System.currentTimeMillis() - curTime;
      System.out.println(curTime + " ms");
    }
    System.out.println();  /**/

    System.gc();
    // signature verification
    repeats = 100;
    System.out.println("n=" + paramM);
    System.out.println("r=" + repeats);
    System.out.println("Verification [E=GF(2^" + extensionDegree +  "), paramK=" + paramK + "]");
    for (int i = 0; i < paramM; i++) {
      curTime = System.currentTimeMillis();
      for (int j = 0; j < repeats; j++) {
        mySpeed.valSign(sigVec[i],manySig[i]);
      }
      curTime = System.currentTimeMillis() - curTime;
      System.out.println(curTime + " ms");
    }
    System.out.println(); /**/

    System.out.println("n=0");  // finis!
    System.out.println();
    long overall2 = System.currentTimeMillis() - overallTime;
    System.out.println("Overall time: " + overall2 + " ms (about " + (((float)overall2) / (1000*60))+ " min)");
  }

  /** Tests everything in this class
      @see #testIt
  */
  public static void main(String[] args) {
    // init static variables in our classes
    Multi2Mul.initAll();
    Poly2_nMax.initAll();


    //testIt();
   speedIt();
  }
}
