package HFE;

import java.lang.*;
import java.io.*;
import java.util.*;
import java.security.*;
import sun.misc.*;
import HFE.*;

//-----------------------------------------------------------------------------
// Project:     Implementing HFE at UCC as Project CS4095
// Supervisors: Simon Foley (Computer Science) and Patrick Fitzpatrick (Maths)
//-----------------------------------------------------------------------------

/**
<p>
This class is the "toy version" of HFE which works on
6 bits over the field F_4 ({@link HFE.Field4}) which
extents to F_64.
<p>
@author Christopher Wolf (hfe@Christopher-Wolf.de)
@version 0.1
*/

//Copyright (C) 2002    Christopher Wolf
//
//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License (file LICENCE) for more details.


class HFE4_3 {

  /** Stores the private polynomial over F_64. Is used for decryption.
    @serial Private Key
  */
  private Poly64Max privatePoly;

  /** Stores the public key matrix of polynomials over GF(4).
      They are used for encryption.
    @serial Matrix
  */
  private Poly4MultiSquare[] publicPoly;

  /** Stores a copy of the message digest which is used to
      get one unique solution
     @serial Message Digest
  */
  private final MessageDigest digest;

  /** Transformation S in HFE
  */
  private Transform4 transS;

  /** Transformation T in HFE
  */
  private Transform4 transT;

  /** Message as a vector. For the sake of speed is it a static.
    @serial vector
  */
  private static int[] textVec = new int[Poly64Max.extension];

  /** Message as a vector. For the sake of speed is it a static.
    @serial vector
  */
  private static int[] resVec = new int[Poly64Max.extension];

  /** Message in a special format as a vector. For the sake of speed is it a static.
    @serial vector
  */
  private static int[] valueVec = new int[Poly64Max.extension+1];

  /** Cipher as a vector. For the sake of speed static.
  */
  private static int[] codeVec = new int[Poly4MultiSquare.tau(Poly64Max.extension)];


  /** Initializes HFE4_3 by generating the
      public and the private key from the given seed. Also generates the
      message digest.
     @param seed Seed for the cryptographic random number generator.
            <b>Caution! This seed is to short for cryptographic purposes!</b>
  */
  HFE4_3 (long seed) throws NoSuchAlgorithmException {
    generateKeys(seed);
    digest = MessageDigest.getInstance("SHA1");
  }


  /** Produces a 16 bit hash for the given plaintext.
    @param plain Plaintext in the lower 6 bits
  */
  int giveHash(int plain) {
    // add redundency to the message
    try {
      String plainText = plain + ":" + plain;
      byte[] data = plainText.getBytes("UTF8");
      byte[] hashData = digest.digest(data);

      return hashData[0] & 65535;
    }
    catch(IOException ioe) {
      ioe.printStackTrace();
      return 0;
    }
  }


  /** computes an vector of integers out of one integer
  */
  private static void int2vec(int in, int[] out) {
    Check.assert(out.length == 3, "Out has wrong length.");
    Check.assert((0 <= in) && (in < 64));
    out[0] = in & 3;
    out[1] = (in >> 2) & 3;
    out[2] = (in >> 4) & 3;
  }

  /** computes an vector of integers out of one integer
  */
  private static int vec2int(int[] in) {
    Check.assert(in.length == 3, "In has wrong length.");
    Check.assert(Field4.isElement(in));
    return in[0] + (in[1] << 2) + (in[2] << 4);
  }


  /** Generates a private key for HFE. This polynomial has the form
      f(x) = ax^20 + bx^17 + cx^16 + dx^4 + e where a, b, c, d, e are
      elements from F_64.
     @param  coeff   Coefficients for the private polynomial
     @return private polynomial
  */
  private static Poly64Max genPolyPrivate(int [] coeff) {
    Check.assert(coeff.length == 5, "Wrong number of coefficients.");
    return new Poly64Max(new int[] {coeff[0],20, coeff[1],17, coeff[2], 16, coeff[3],4, coeff[4],0});
  }

  /** Generates the 3 public polynomial from the
      object's private polynomial.
     @param  coeff   Coefficients for the private polynomial
     @return 3 Polynomials in a matrix over F_4. The matrix has the size n*{@link Poly4MultiSquare#tau(int) tau(n)} [here: n=3, tau(3)=10]
  */
  private Poly4MultiSquare[] genPolyPublic(int [] coeff) {
    Check.assert(coeff.length == 5, "Wrong number of coefficients.");
    int[] vec = new int[3];
    Multi4Power powers = new Multi4Power(Poly64Max.genPoly, transS.getTransform(), 2);
    Multi4Mul calc = new Multi4Mul(Poly64Max.genPoly);

    Poly4MultiSquare[] res20 = new Poly4MultiSquare[3];
    int2vec(coeff[0],vec);
    res20 = calc.mul(vec,powers.givePower(1),powers.givePower(2));  // 20 = 4 + 16 = 4^1 + 4^2
    Poly4MultiSquare[] res17 = new Poly4MultiSquare[3];
    int2vec(coeff[1],vec);
    res17 = calc.mul(vec,powers.givePower(0),powers.givePower(2));  // 17 = 1 + 16 = 4^0 + 4^2
    Poly4MultiLinear[] res16 = new Poly4MultiLinear[3];
    int2vec(coeff[2],vec);
    res16 = calc.mul(vec,powers.givePower(2));  // 16 = 4^2
    Poly4MultiLinear[] res4 = new Poly4MultiLinear[3];
    int2vec(coeff[3],vec);
    res4 = calc.mul(vec,powers.givePower(1));  // 4 = 4^1

    // now take e into consideration
    int2vec(coeff[4],vec);

    Poly4MultiSquare[] res = new Poly4MultiSquare[3];
    for (int i = 0; i < 3; i++) {
      res[i] = new Poly4MultiSquare(3);
      res[i].add(res20[i]);
      res[i].add(res17[i]);
      res[i].add(res16[i]);
      res[i].add(res4[i]);
      res[i].add(vec[i],0);  // add the constant term
    }
    res = transT.transform(res);  // overwrite or result vector with the T transformation

    return res; // after "a few steps" we finally got our public polynomials
  }


  /** Generates the public and the private key from
      a given seed and stores them in the object's fields.
      For cryptographic purposes, the seed is too short.
     @param seed Seed for the random number generator
  */
  void generateKeys(long seed) {
    SecureRandom rndGen = new SecureRandom();
    rndGen.setSeed(seed);
    int a = rndGen.nextInt(64);
    int b = rndGen.nextInt(64);
    int c = rndGen.nextInt(64);
    int d = rndGen.nextInt(64);
    int e = rndGen.nextInt(64);
    transS = new Transform4(rndGen);
    transT = new Transform4(rndGen);
    int[] coeff = new int[] {a,b,c,d,e};
    privatePoly = genPolyPrivate(coeff);
    publicPoly = genPolyPublic(coeff);
  }

  // gets the roots of a given Polynominal
  /** Calculates the root for a given Polynomial and a given
      element of F_64. The equation is f(x) = a.
      Returns _all_ x which satisfy this equation.
      There are up to d (d: degree of the polynomial) solutions.<br>
      Currently, this method uses brute force.
     @return all x with f(x) = a
  */
  Collection getRoots(Poly64Max masterPoly, int a) {
    LinkedList res = new LinkedList();
    // solves this problem using brute force
    for (int i = 0; i < 64; i++) {
      if (masterPoly.applyPoly(i) == a) {
        // now invert transformation S on the current root
        int2vec(i,textVec);
        transS.untransform(textVec,resVec);
        res.add(new Integer(vec2int(resVec)));
      }
    }
    return res;
  }

  // findes out which is the correct root
  // returns the _first_ root which is correct
  // throws an exception in an later version if there are multiple roots
  /** Finds out which roots are correct acording to the given hash.
  */
  int filterRoots(Collection resColl, int hash) {
    Iterator resIt = resColl.iterator();
    while (resIt.hasNext()) {
      Integer curRoot = (Integer)resIt.next();
      // check, if there are _two_ roots!
      if (giveHash(curRoot.intValue()) == hash) return curRoot.intValue();
    }
    // should throw an exception here if there is no root
    return -1;
  }


  /** Encryptes a given, 6 bit plain text.
      <br>
      Is currently using the <b>private</b> polynomial.
     @param plain plaintext in the lower 6 bits
     @return encyrpted plaintext (lower 6 bits) and hash (16 bits at bit positions 6..21)
  */
  int encrypt(int plain) {
    Check.assert((0 <= plain) && (plain < 64), "plain text out of range: " + plain + ";");
    // first transform the message into a vector and get the right values
    int2vec(plain,textVec);
    System.arraycopy(textVec,0,valueVec,1,Poly64Max.extension);
    Poly4MultiSquare.getValues(valueVec,codeVec);
    // now apply the 3 polynomials
    for (int i = 0; i < 3; i++)
      textVec[i] = publicPoly[i].apply(codeVec);
    // last but now least get an integer from the vector
    int cipher = vec2int(textVec);

    return cipher | (giveHash(plain) << 6);
  }

  // takes 22 bit input and produces 6 bits output
  // the lower 6 bits are the cryptotext, the upper 16 bits are a hash
  /** Decryptes a given, 22 bit ciphertext / hash
     @param cipherHash 22 bit where the lower 6 bits are the cipher text and the
            bits 6..21 are hash of the plaintext.
     @return plaintext in the lower 6 bits (i.e. 0..5)
  */
  int decrypt(int cipherHash) {
    int hash = cipherHash >> 6;
    int cipher = cipherHash & 63;

    // unstrip transformation T
    int2vec(cipher,textVec);
    transT.untransform(textVec,resVec);
    cipher = vec2int(resVec);

    // and now get the correct value for deciphering
    int plain = filterRoots(getRoots(privatePoly,cipher),hash);
    return plain;
  }


  /** Tests the functionality of the whole class.
      If there occurs any error it will be printed on stdout.
  */
  public static void testIt() {
    System.out.println("Start testing HFE4_3...");
    Poly64Max function = new Poly64Max(new int[] {4,5, 3,4, 7,3, 4,2, 8,1, 4,0}); // function x
    try {
      HFE4_3 testObj = new HFE4_3(15);

      Random rnd = new Random();
      SecureRandom secRnd = new SecureRandom();
      int curRnd;

      // test encryption / decryption
      if (testObj.decrypt(testObj.encrypt(0)) != 0) System.out.println("Error 1.1");
      for (int i = 0; i < 64; i++) {
        if (i != testObj.decrypt(testObj.encrypt(i))) System.out.println("Error 1.2 with " + i);
      }

      // tests the getRoot and filterRoot functions
      testObj.transS = new Transform4();  // identity transformation
      Poly64Max curPoly = new Poly64Max(new int[] {4,5, 3,4, 7,3, 4,2, 8,1, 4,0});
      if (0 != testObj.filterRoots(testObj.getRoots(curPoly,curPoly.applyPoly(0)),testObj.giveHash(0))) System.out.print("Error 2.1a");
      if (3 != testObj.filterRoots(testObj.getRoots(curPoly,curPoly.applyPoly(3)),testObj.giveHash(3))) System.out.print("Error 2.1b");

      // test if the public and the private polynomial do the same
      testObj.generateKeys(123645);
      for (int i = 0; i < 64; i++) {
        if (i != testObj.decrypt(testObj.encrypt(i))) System.out.println("Error 3.1 with " + i);
      }

      int[] vec = new int[3];
      int2vec(5,vec);
      if ((vec[0] != 1) || (vec[1] != 1) || (vec[2] != 0)) System.out.println("Error 4.1a");
      if (vec2int(vec) != 5) System.out.println("Error 4.1b");

      int2vec(57,vec);
      if ((vec[0] != 1) || (vec[1] != 2) || (vec[2] != 3)) System.out.println("Error 4.2a");
      if (vec2int(vec) != 57) System.out.println("Error 4.2b");

      for (int i = 0; i < 64; i++) {
        int a = i;
        int2vec(a,vec);
        if (vec2int(vec) != a) System.out.println("Error 4.3 with " + a + " and " + vec2int(vec));
      }



    }
    catch(NoSuchAlgorithmException nsae) {
      nsae.printStackTrace();
      System.out.println("\n\nMAJOR ERROR - SHA1 Algorithm not available, Exiting...");
      System.exit(1);
    }
    System.out.println("...ended testing.");
  }

  public final static void speedIt() {
    Check.assert(false);  // make sure we do not have any assertion checkers

    final int repeats = 10000;
    Random rnd = new Random();
    try {
    HFE4_3 myTest = new HFE4_3(123);

    int[] X = new int[repeats], Y = new int[repeats];
    System.out.println();

    long curTime = System.currentTimeMillis();
    // create random polynomials first
    for (int i = 0; i < repeats; i++)
      X[i] = rnd.nextInt(64);

    curTime = System.currentTimeMillis() - curTime;
    System.out.println("Generation of random values took " + curTime + "ms (" + repeats + " operations)");

    SecureRandom rndGen = new SecureRandom();
    rndGen.setSeed(1234);
    int a, b, c, d, e;
    int[] coeff = new int[] {0,0,0,0,0};


    curTime = System.currentTimeMillis();
    // create random polynomials first
    for (int i = 0; i < repeats; i++) {
      a = rndGen.nextInt(64);
      b = rndGen.nextInt(64);
      c = rndGen.nextInt(64);
      d = rndGen.nextInt(64);
      e = rndGen.nextInt(64);
      myTest.transS = new Transform4(rndGen);
      myTest.transT = new Transform4(rndGen);
      coeff[0] = a;
      coeff[1] = b;
      coeff[2] = c;
      coeff[3] = d;
      coeff[4] = e;
      myTest.privatePoly = myTest.genPolyPrivate(coeff);
      myTest.publicPoly = myTest.genPolyPublic(coeff);
    }
    curTime = System.currentTimeMillis() - curTime;
    System.out.println("Key generation took " + curTime + "ms (" + repeats + " operations)");


    curTime = System.currentTimeMillis();
    // encrypt now
    for (int i = 0; i < repeats; i++) {
      Y[i] = myTest.encrypt(X[i]);
    }
    curTime = System.currentTimeMillis() - curTime;
    System.out.println("Encryption took " + curTime + "ms (" + repeats + " operations)");

    curTime = System.currentTimeMillis();
    // encrypt now
    for (int i = 0; i < repeats; i++) {
       myTest.decrypt(Y[i]);
    }
    curTime = System.currentTimeMillis() - curTime;
    System.out.println("Decryption took " + curTime + "ms (" + repeats + " operations)");

    } // from try
    catch (Exception e) {
      System.out.println("Did not work...");
    }
  }

  /** Tests everything in this class
      @see #testIt
  */
  public static void main(String[] args) {
    testIt();
//     speedIt();
//     System.out.println();
//     System.out.println("==============================");
//     System.out.println("           Run 2");
//     System.out.println("==============================");
//     System.out.println();
//     speedIt();
//     System.out.println();
//     System.out.println("==============================");
//     System.out.println("           Run 3");
//     System.out.println("==============================");
//     System.out.println();
//     speedIt();
  }
}
