package HFE;

import java.util.*;
import HFE.*;

//-----------------------------------------------------------------------------
// Project:     Implementing HFE at UCC as Project CS4095
// Supervisors: Simon Foley (Computer Science) and Patrick Fitzpatrick (Maths)
//-----------------------------------------------------------------------------

/**
<p>
This class encapsulates the operations for LUDecomposition
for matrices over the finite field {@link Field4 F_4}).
<br>
The class provides only operations which are needed in the
HFE package. The class is based upon the Jama package, which
was produced by The MathWorks, Inc. and the National
Institute of Standards and Technology.
Jama can be downloaded from the NIST page
<a href="http://math.nist.gov/javanumerics/jama/" target="new">http://math.nist.gov/javanumerics/jama/</a>.
<p>
The explenation of LU decomposition is copied from there documentation:
<p>
For an m-by-n matrix A with m >= n, the LU decomposition is an m-by-n
unit lower triangular matrix L, an n-by-n upper triangular matrix U,
and a permutation vector piv of length m so that A(piv,:) = L*U.
If m < n, then L is m-by-m and U is m-by-n.
<p>
The LU decompostion with pivoting always exists, even if the matrix is
singular, so the constructor will never fail.  The primary use of the
LU decomposition is in the solution of square systems of simultaneous
linear equations. This will fail if isNonsingular() returns false.
<p>
@author Christopher Wolf (chris@Christopher-Wolf.de)
@version 0.1
*/


public class LUDecomposition4 implements java.io.Serializable {

/* ------------------------
   Class variables
 * ------------------------ */

   /** Array for internal storage of decomposition.
   @serial internal array storage.
   */
   private int[][] LU;

   /** Row and column dimensions, and pivot sign.
   @serial column dimension.
   @serial row dimension.
   @serial pivot sign.
   */
   private int m, n;  // CW: The sign keeps the same in F_4:  ,  pivsign;

   /** Internal storage of pivot vector.
   @serial pivot vector.
   */
   private int[] piv;

/* ------------------------
   Constructor
 * ------------------------ */

   /** LU Decomposition
   @param  A   Rectangular matrix
   @return     Structure to access L, U and piv.
   */
   public LUDecomposition4 (Matrix4 A) {

     // Use a "left-looking", dot-product, Crout/Doolittle algorithm.

      LU = A.getArrayCopy();
      m = A.getRowDimension();
      n = A.getColumnDimension();
      piv = new int[m];
      for (int i = 0; i < m; i++) {
         piv[i] = i;
      }
      // CW: The sign keeps the same in F_4, therefore:
      // pivsign = 1;
      int[] LUrowi;
      int[] LUcolj = new int[m];

      // Outer loop.

      for (int j = 0; j < n; j++) {

         // Make a copy of the j-th column to localize references.

         for (int i = 0; i < m; i++) {
            LUcolj[i] = LU[i][j];
         }

         // Apply previous transformations.

         for (int i = 0; i < m; i++) {
            LUrowi = LU[i];

            // Most of the time is spent in the following dot product.

            int kmax = Math.min(i,j);
            int s = 0;
            for (int k = 0; k < kmax; k++) {
               s = Field4.add(s,Field4.mul(LUrowi[k],LUcolj[k]));
            }

            LUcolj[i] = Field4.sub(LUcolj[i],s);
            LUrowi[j] = LUcolj[i];
         }


         // Find pivot and exchange if necessary.
                  int p = j;
         for (int i = j+1; i < m; i++) {
            // CW: there are no negative numbers in GF(4), therefore this line simplifies
            //if (Math.abs(LUcolj[i]) > Math.abs(LUcolj[p])) {
            if (LUcolj[i] > LUcolj[p]) {
               p = i;
            }
         }
         if (p != j) {
            for (int k = 0; k < n; k++) {
               int t = LU[p][k]; LU[p][k] = LU[j][k]; LU[j][k] = t;
            }
            int k = piv[p]; piv[p] = piv[j]; piv[j] = k;
            // CW: The sign keeps the same in F_4
            //pivsign = -pivsign;
         }

         // Compute multipliers.

         if ((j < m) & (LU[j][j] != 0)) {
            for (int i = j+1; i < m; i++) {
               LU[i][j] = Field4.div(LU[i][j],LU[j][j]);
            }
         }
      }
   }

/* ------------------------
   Public Methods
 * ------------------------ */


   /** Generates a random nonsingular matrix of the given size.
    @param size Number of rows/columns.
    @param rnd  Random number generator.
    @return In position [0] of the array is the generated matrix, in position [1] its LU decomposition.
   */
   public static Object[] generateNonsingular(int size, Random rnd) {
     Check.assert(size > 0, "Size must be positive.");
     Matrix4 A = new Matrix4(size,size);
     LUDecomposition4 lu;
     do {
       A.set(rnd);
       lu = A.lu();
     } while (!lu.isNonsingular());
     return new Object[] {A, lu};
   }

   /** Is the matrix nonsingular?
   @return     true if U, and hence A, is nonsingular.
   */
   public boolean isNonsingular () {
     Check.assert(n == m, "Matrix must be square.");
     for (int j = 0; j < n; j++) {
       if (LU[j][j] == 0)
         return false;
     }
     return true;
   }

   /** Determinant
   @return     det(A)
   */
   public int det () {
      Check.assert(m == n, "Matrix must be square");
      int d = LU[0][0]; // CW: The sign keeps the same in F_4    (int) pivsign;
      for (int j = 1; j < n; j++)
         Field4.mul(d,LU[j][j]);
      return d;
   }

   /** Solve A*X = B
   @param  B   A Matrix with as many rows as A and any number of columns.
   @return     X so that L*U*X = B(piv,:)
   @exception  IllegalArgumentException Matrix row dimensions must agree.
   @exception  RuntimeException  Matrix is singular.
   */
   public Matrix4 solve (Matrix4 B) {
      Check.assert(B.getRowDimension() == m, "Matrix row dimensions must agree.");
      Check.assert(isNonsingular(), "Matrix is singular.");

      // Copy right hand side with pivoting
      int nx = B.getColumnDimension();
      Matrix4 Xmat = B.getMatrix(piv,0,nx-1);
      int[][] X = Xmat.getArray();

      // Solve L*Y = B(piv,:)
      for (int k = 0; k < n; k++) {
         for (int i = k+1; i < n; i++) {
            for (int j = 0; j < nx; j++) {
               X[i][j] = Field4.sub(X[i][j],Field4.mul(X[k][j],LU[i][k]));
            }
         }
      }
      // Solve U*X = Y;
      for (int k = n-1; k >= 0; k--) {
         for (int j = 0; j < nx; j++) {
            // CW: X[k][j] /= LU[k][k];
            X[k][j] = Field4.div(X[k][j],LU[k][k]);
         }
         for (int i = 0; i < k; i++) {
            for (int j = 0; j < nx; j++) {
               X[i][j] = Field4.sub(X[i][j],Field4.mul(X[k][j],LU[i][k]));
            }
         }
      }
      return Xmat;
   }

   /** Solve A*x = b, i.e. computes x so that L*U*x = b(piv,:)
   @param  b   A vector with as many rows as A; it is treated to be a n*1 matrix.
   */
   public void solve (int[] b, int[] x) {
     Check.assert(b.length == m, "Matrix row dimensions must agree.");
     Check.assert(b.length == x.length, "Both vectors must have the same size.");
     Check.assert(Field4.isElement(b), "b is not of Field4");
     Check.assert(isNonsingular(), "Matrix is singular.");

     // Copy right hand side with pivoting
     for (int i = 0; i < b.length; i++)
       x[i] = b[piv[i]];


     // Solve L*Y = B(piv,:)
     for (int k = 0; k < n; k++) {
        for (int i = k+1; i < n; i++) {
           x[i] = Field4.sub(x[i],Field4.mul(x[k],LU[i][k]));
        }
     }
     // Solve U*X = Y;
     for (int k = n-1; k >= 0; k--) {
       x[k] = Field4.div(x[k],LU[k][k]);
       for (int i = 0; i < k; i++)
         x[i] = Field4.sub(x[i],Field4.mul(x[k],LU[i][k]));
     }
   }

  /** Tests the functionality of the whole class.
      If there occurs any error it will be printed on stdout.
  */
  public static void testIt () {
    System.out.println("start testing LUDecompostion4 ...");

    final int repeats = 100;

    Matrix4 A, B, C, D;
    LUDecomposition4 lu;
    Random rnd = new Random();

    // test the det operation
    A = Matrix4.identity(4,4);
    lu = new LUDecomposition4(A);
    if (lu.det() != 1) System.out.println("Error 1.1");
    A = Matrix4.identity(5,5);
    lu = new LUDecomposition4(A);
    if (lu.det() != 1) System.out.println("Error 1.2");
    A = new Matrix4(new int[][] {{0, 1, 2}, {3, 0, 1}, {2, 3, 1}});
    lu = new LUDecomposition4(A);
    if (lu.det() != 3) System.out.println("Error 1.3");

    // test generate nonsingular
    for (int i = 1; i < 10; i++) {
      Object[] res = generateNonsingular(i,rnd);
      A = (Matrix4) res[0];
      lu = (LUDecomposition4) res[1];
      if (lu.det() == 0) System.out.println("Error 1.4");
    }

    // solve a simple equation
    A = new Matrix4(new int[][] {{0, 1, 2}, {3, 0, 1}, {2, 3, 1}});
    lu = new LUDecomposition4(A);
    B = new Matrix4(new int[][] {{1}, {0}, {0}});
    C = lu.solve(B);
    if (!C.isEqual(new Matrix4(new int[][] {{2}, {3}, {1}}))) System.out.println("Error 2.1");
    // solve some random equations
    for (int i = 0; i < repeats; i++) {
      int n = rnd.nextInt(10)+1;
      Object[] res = generateNonsingular(n,rnd);
      A = (Matrix4) res[0];
      lu = (LUDecomposition4) res[1];
      B = new Matrix4(n,n); B.set(rnd);
      C = A.mul(B);
      D = lu.solve(C);
      if (!B.isEqual(D)) System.out.println("Error 2.2");
    }

    // test the "solve" method for vectors
    for (int i = 0; i < repeats; i++) {
      int n = rnd.nextInt(10)+1;
      Object tmp[] = LUDecomposition4.generateNonsingular(n,rnd);
      A = (Matrix4)tmp[0];
      lu = (LUDecomposition4)tmp[1];
      B = new Matrix4(n,1); B.set(rnd);

      C = A.mul(B);
      int[] c = C.getColumnPackedCopy();
      D = lu.solve(C);
      if (!D.isEqual(B)) System.out.println("Error 3.1");
      int[] res = new int[n];
      lu.solve(c,res);
      for (int j = 0; j < n; j++) {
        if (res[j] != D.get(j,0)) System.out.println("Error 3.2 at " + j);
      }
    }

    System.out.println("...finished testing.");
  }

  /** Tests everything in this class
     @see #testIt
  */
  public static void main(String[] args) {
    testIt();
  }

}
