package HFE;

import java.util.*;
import java.security.*;
import HFE.*;

//-----------------------------------------------------------------------------
// Project:     Implementing HFE at UCC as Project CS4095
// Supervisors: Simon Foley (Computer Science) and Patrick Fitzpatrick (Maths)
//-----------------------------------------------------------------------------

/**
<P>
   The class Matrix4 provides matrix operations for matrices over the
   finite field F_4 which is specified in {@link HFE.Field4}.
   The class provides only operations which are needed in the
   HFE package. The class is based upon the Jama package, which
   was produced by The MathWorks, Inc. and the National Institute of Standards and Technology.
   Jama can be downloaded from the NIST page
   <a href="http://math.nist.gov/javanumerics/jama/" target="new">http://math.nist.gov/javanumerics/jama/</a>.
<P>
@author Christopher Wolf (chris@Christopher-Wolf.de)
@version 0.1
*/

public class Matrix4 implements Cloneable, java.io.Serializable {

/* ------------------------
   Class variables
 * ------------------------ */

   /** Array for internal storage of elements.
   @serial internal array storage.
   */
   private int[][] A;

   /** Row (m) and column (n) dimensions.
   @serial row dimension.
   @serial column dimension.
   */
   private int m, n;

/* ------------------------
   Constructors
 * ------------------------ */

   /** Construct an m-by-n matrix of zeros.
   @param m    Number of rows.
   @param n    Number of colums.
   */
   public Matrix4 (int m, int n) {
      this.m = m;
      this.n = n;
      A = new int[m][n];
   }

   /** Construct a matrix using the given array with checking.
       The constructor checks if all the elements are in F_4 and if the
       array is rectangular. It does not clone the given array.
   @param A    Elements of the matrix as a 2D array.
   */
   public Matrix4 (int[][] A) {
     m = A.length;
     n = A[0].length;
     for (int i = 0; i < m; i++) {
       if (A[i].length != n) {
         throw new IllegalArgumentException("All rows must have the same length.");
       }
     }
     for (int i = 0; i < m; i++)
       for (int j = 0; j < n; j++)
         if (A[i][j] >= Field4.elements) throw new IllegalArgumentException("Element (" + i + "," + j + "): " + A[i][j] + " is not in F_4");
     this.A = A;
   }


   /** Construct a matrix using the given array without checking.
       The constructor does not check if all the elements are in F_4 or if the
       array is rectangular but assumes that. It does neither clone this array.
   @param A    Elements of the matrix as a 2D array.
   @param m    Number of rows.
   @param n    Number of colums.
   */
   public Matrix4 (int[][] A, int m, int n) {
     this.A = A;
     this.m = m;
     this.n = n;
   }

    /** Generats an m-by-n random matrix.
     @param m    Number of rows.
     @param n    Number of colums.
     @param rnd  number generator
   */
   public Matrix4 (int m, int n, Random rnd) {
      this.m = m;
      this.n = n;
      A = new int[m][n];
      set(rnd);
   }

   /** Make a deep copy of a matrix
       Allocates new memory
   */
   public Matrix4 copy () {
      Matrix4 X = new Matrix4(m,n);
      int[][] C = X.getArray();
      for (int i = 0; i < m; i++) {
         for (int j = 0; j < n; j++) {
            C[i][j] = A[i][j];
         }
      }
      return X;
   }

   /** Clone the Matrix object.
   */

   public Object clone () {
      return this.copy();
   }

   /** Returns a pointer to the internal two-dimensional array.
   @return     Pointer to the two-dimensional array of matrix elements.
   */

   public int[][] getArray () {
      return A;
   }

   /** Copy the internal two-dimensional array.
   @return     Two-dimensional array copy of matrix elements.
   */

   public int[][] getArrayCopy () {
      int[][] C = new int[m][n];
      for (int i = 0; i < m; i++) {
         for (int j = 0; j < n; j++) {
            C[i][j] = A[i][j];
         }
      }
      return C;
   }

   /** Make a one-dimensional column packed copy of the internal array.
   @return     Matrix elements packed in a one-dimensional array by columns.
   */

   public int[] getColumnPackedCopy () {
      int[] vals = new int[m*n];
      for (int i = 0; i < m; i++) {
         for (int j = 0; j < n; j++) {
            vals[i+j*m] = A[i][j];
         }
      }
      return vals;
   }

   /** Make a one-dimensional row packed copy of the internal array.
   @return     Matrix elements packed in a one-dimensional array by rows.
   */

   public int[] getRowPackedCopy () {
      int[] vals = new int[m*n];
      for (int i = 0; i < m; i++) {
         for (int j = 0; j < n; j++) {
            vals[i*n+j] = A[i][j];
         }
      }
      return vals;
   }

   /** Get row dimension.
   @return     m, the number of rows.
   */

   public int getRowDimension () {
      return m;
   }

   /** Get column dimension.
   @return     n, the number of columns.
   */

   public int getColumnDimension () {
      return n;
   }

   /** Get a single element.
   @param i    Row index.
   @param j    Column index.
   @return     A(i,j)
   @exception  ArrayIndexOutOfBoundsException
   */

   public int get (int i, int j) {
      return A[i][j];
   }

   /** Get a submatrix.
   @param i0   Initial row index
   @param i1   Final row index
   @param j0   Initial column index
   @param j1   Final column index
   @return     A(i0:i1,j0:j1)
   @exception  ArrayIndexOutOfBoundsException Submatrix indices
   */

   public Matrix4 getMatrix (int i0, int i1, int j0, int j1) {
      Matrix4 X = new Matrix4(i1-i0+1,j1-j0+1);
      int[][] B = X.getArray();
      try {
         for (int i = i0; i <= i1; i++) {
            for (int j = j0; j <= j1; j++) {
               B[i-i0][j-j0] = A[i][j];
            }
         }
      } catch(ArrayIndexOutOfBoundsException e) {
         throw new ArrayIndexOutOfBoundsException("Submatrix indices");
      }
      return X;
   }

   /** Get a submatrix.
   @param r    Array of row indices.
   @param c    Array of column indices.
   @return     A(r(:),c(:))
   @exception  ArrayIndexOutOfBoundsException Submatrix indices
   */

   public Matrix4 getMatrix (int[] r, int[] c) {
      Matrix4 X = new Matrix4(r.length,c.length);
      int[][] B = X.getArray();
      try {
         for (int i = 0; i < r.length; i++) {
            for (int j = 0; j < c.length; j++) {
               B[i][j] = A[r[i]][c[j]];
            }
         }
      } catch(ArrayIndexOutOfBoundsException e) {
         throw new ArrayIndexOutOfBoundsException("Submatrix indices");
      }
      return X;
   }

   /** Get a submatrix.
   @param i0   Initial row index
   @param i1   Final row index
   @param c    Array of column indices.
   @return     A(i0:i1,c(:))
   @exception  ArrayIndexOutOfBoundsException Submatrix indices
   */

   public Matrix4 getMatrix (int i0, int i1, int[] c) {
      Matrix4 X = new Matrix4(i1-i0+1,c.length);
      int[][] B = X.getArray();
      try {
         for (int i = i0; i <= i1; i++) {
            for (int j = 0; j < c.length; j++) {
               B[i-i0][j] = A[i][c[j]];
            }
         }
      } catch(ArrayIndexOutOfBoundsException e) {
         throw new ArrayIndexOutOfBoundsException("Submatrix indices");
      }
      return X;
   }

   /** Get a submatrix.
   @param r    Array of row indices.
   @param i0   Initial column index
   @param i1   Final column index
   @return     A(r(:),j0:j1)
   @exception  ArrayIndexOutOfBoundsException Submatrix indices
   */

   public Matrix4 getMatrix (int[] r, int j0, int j1) {
      Matrix4 X = new Matrix4(r.length,j1-j0+1);
      int[][] B = X.getArray();
      try {
         for (int i = 0; i < r.length; i++) {
            for (int j = j0; j <= j1; j++) {
               B[i][j-j0] = A[r[i]][j];
            }
         }
      } catch(ArrayIndexOutOfBoundsException e) {
         throw new ArrayIndexOutOfBoundsException("Submatrix indices");
      }
      return X;
   }

   /** Set a single element.
   @param i    Row index.
   @param j    Column index.
   @param s    A(i,j).
   @exception  ArrayIndexOutOfBoundsException
   */
   public void set (int i, int j, int s) {
      A[i][j] = s;
   }

   /** Generats an m-by-n random matrix.
   @param rnd  number generator
   */
   public void set(Random rnd) {
     for (int i = 0; i < m; i++) {
       for (int j = 0; j < n; j++) {
          A[i][j] = Field4.rnd(rnd);
       }
     }
   }

   /** Set a submatrix.
   @param i0   Initial row index
   @param i1   Final row index
   @param j0   Initial column index
   @param j1   Final column index
   @param X    A(i0:i1,j0:j1)
   @exception  ArrayIndexOutOfBoundsException Submatrix indices
   */

   public void setMatrix (int i0, int i1, int j0, int j1, Matrix4 X) {
      try {
         for (int i = i0; i <= i1; i++) {
            for (int j = j0; j <= j1; j++) {
               A[i][j] = X.get(i-i0,j-j0);
            }
         }
      } catch(ArrayIndexOutOfBoundsException e) {
         throw new ArrayIndexOutOfBoundsException("Submatrix indices");
      }
   }

   /** Set a submatrix.
   @param r    Array of row indices.
   @param c    Array of column indices.
   @param X    A(r(:),c(:))
   @exception  ArrayIndexOutOfBoundsException Submatrix indices
   */

   public void setMatrix (int[] r, int[] c, Matrix4 X) {
      try {
         for (int i = 0; i < r.length; i++) {
            for (int j = 0; j < c.length; j++) {
               A[r[i]][c[j]] = X.get(i,j);
            }
         }
      } catch(ArrayIndexOutOfBoundsException e) {
         throw new ArrayIndexOutOfBoundsException("Submatrix indices");
      }
   }

   /** Set a submatrix.
   @param r    Array of row indices.
   @param j0   Initial column index
   @param j1   Final column index
   @param X    A(r(:),j0:j1)
   @exception  ArrayIndexOutOfBoundsException Submatrix indices
   */

   public void setMatrix (int[] r, int j0, int j1, Matrix4 X) {
      try {
         for (int i = 0; i < r.length; i++) {
            for (int j = j0; j <= j1; j++) {
               A[r[i]][j] = X.get(i,j-j0);
            }
         }
      } catch(ArrayIndexOutOfBoundsException e) {
         throw new ArrayIndexOutOfBoundsException("Submatrix indices");
      }
   }

   /** Set a submatrix.
   @param i0   Initial row index
   @param i1   Final row index
   @param c    Array of column indices.
   @param X    A(i0:i1,c(:))
   @exception  ArrayIndexOutOfBoundsException Submatrix indices
   */
   public void setMatrix (int i0, int i1, int[] c, Matrix4 X) {
      try {
         for (int i = i0; i <= i1; i++) {
            for (int j = 0; j < c.length; j++) {
               A[i][c[j]] = X.get(i-i0,j);
            }
         }
      } catch(ArrayIndexOutOfBoundsException e) {
         throw new ArrayIndexOutOfBoundsException("Submatrix indices");
      }
   }

   /** Set a row elementwise zero.
   */
   public void setRowZero(int i) {
     try {
       int [] rowI = A[i];
       for (int k = 0; k < n; k++)
         rowI[k] = 0;
     }
     catch (ArrayIndexOutOfBoundsException e) {
       throw new IllegalArgumentException("Row out of bound.");
     }
   }


   /** Matrix transpose.
   @return    A^T
   */
   public Matrix4 transpose () {
      Matrix4 X = new Matrix4(n,m);
      int[][] C = X.getArray();
      for (int i = 0; i < m; i++) {
        int[] rowI = A[i];
        for (int j = 0; j < n; j++) {
          C[j][i] = rowI[j];
        }
      }
      return X;
   }


   /** C = A + B
   @param B    another matrix
   @return     A + B
   */
   public Matrix4 add(Matrix4 B) {
      checkMatrixDimensions(B);
      Matrix4 X = new Matrix4(m,n);
      int[][] C = X.getArray();
      for (int i = 0; i < m; i++) {
         for (int j = 0; j < n; j++) {
            C[i][j] = Field4.add(A[i][j],B.A[i][j]);
         }
      }
      return X;
   }


   /** A[i] += A[j].
       Adds elementwise the two rows i,j
   */
   public void addRowsEquals(int i, int j) {
     try {
       int[] rowI, rowJ;
       rowI = A[i];
       rowJ = A[j];
       for (int k = 0; k < n; k++)
         rowI[k] = Field4.add(rowI[k],rowJ[k]);
     }
     catch (ArrayIndexOutOfBoundsException e) {
       throw new IllegalArgumentException("Row out of bound.");
     }
   }


   /** A[i] += f*A[j].
       Adds elementwise the two rows i,j. The row j is first multiplied with the scalar f
       @param f scalar
       @param i number of the destination row
       @param j number of the source row
   */
   public void addRowsEquals(int f, int i, int j) {
     int[] rowI, rowJ;
     rowI = A[i];
     rowJ = A[j];
     for (int k = 0; k < n; k++)
       rowI[k] = Field4.add(rowI[k],Field4.mul(f,rowJ[k]));
   }


   /** A += B
   @param B    another matrix
   @return     A + B
   */
   void addEquals (Matrix4 B) {
      checkMatrixDimensions(B);
      for (int i = 0; i < m; i++) {
         for (int j = 0; j < n; j++) {
            A[i][j] = Field4.add(A[i][j],B.A[i][j]);
         }
      }
   }

   /** C = A - B
   @param B    another matrix
   @return     A - B
   */

   public Matrix4 sub (Matrix4 B) {
      checkMatrixDimensions(B);
      Matrix4 X = new Matrix4(m,n);
      int[][] C = X.getArray();
      for (int i = 0; i < m; i++) {
         for (int j = 0; j < n; j++) {
            C[i][j] = Field4.sub(A[i][j],B.A[i][j]);
         }
      }
      return X;
   }

   /** A -= B
   @param B    another matrix
   @return     A - B
   */
   void subEquals (Matrix4 B) {
      checkMatrixDimensions(B);
      for (int i = 0; i < m; i++) {
         for (int j = 0; j < n; j++) {
            A[i][j] = Field4.sub(A[i][j],B.A[i][j]);
         }
      }
   }

   /** Linear algebraic matrix multiplication, A * B.
       Multiplication is not in place.
   @param B    another matrix with B.m == A.n (rows equals columns)
   @return     Matrix product (in a new matrix!), A * B
   @exception  IllegalArgumentException Matrix inner dimensions must agree.
   */
   public Matrix4 mul (Matrix4 B) {
      if (B.m != n) {
         throw new IllegalArgumentException("Matrix inner dimensions must agree.");
      }
      Matrix4 X = new Matrix4(m,B.n);
      int[][] C = X.getArray();
      int[] Bcolj = new int[n];
      for (int j = 0; j < B.n; j++) {
         for (int k = 0; k < n; k++) {
            Bcolj[k] = B.A[k][j];
         }
         for (int i = 0; i < m; i++) {
            int[] Arowi = A[i];
            int s = 0;
            for (int k = 0; k < n; k++) {
               s = Field4.add(s,Field4.mul(Arowi[k],Bcolj[k]));
            }
            C[i][j] = s;
         }
      }
      return X;
   }

   /** Linear algebraic matrix/vector multiplication, res = A * b.
       Multiplication is in place.
   @param b    a vector with b.length == A.n (rows equals columns)
   */
   public void mul (int[] b, int[] res) {
     Check.assert(b.length == n, "Columns must equal rows in b " + n + ", " + b.length);
     Check.assert(res.length == n, "Columns must equal rows in res " + n + ", " + b.length);
     Check.assert(Field4.isElement(b), "b is not completely Field4");

     for (int i = 0; i < n; i++) {
       int tmp = 0;
       for (int j = 0; j < n; j++) {
         tmp = Field4.add(tmp,Field4.mul(b[j],A[i][j]));
       }
       res[i] = tmp;
     }
   }
   /** LU Decomposition
   @return     LUDecomposition4
   @see LUDecomposition4
   */
   public LUDecomposition4 lu () {
      return new LUDecomposition4(this);
   }


   /** Generate identity matrix
   @param m    Number of rows.
   @param n    Number of colums.
   @return     An m-by-n matrix with ones on the diagonal and zeros elsewhere.
   */
   public static Matrix4 identity (int m, int n) {
      Matrix4 A = new Matrix4(m,n);
      int[][] X = A.getArray();
      for (int i = 0; i < m; i++) {
         for (int j = 0; j < n; j++) {
            X[i][j] = (i == j ? 1 : 0);
         }
      }
      return A;
   }


   /** Compares two matrices.
   @param B    another matrix
   @return     <code>true</code>: Matrices are equal. <br>
               <code>false</code>: Matrices are not equal (e.g. different size, different elements)
   */
   public boolean isEqual(Matrix4 B) {
     if ((B.m != m) || (B.n != n)) return false;
     int[][] X = B.getArray();
     for (int i = 0; i < m; i++)
       for (int j = 0; j > n; j++)
         if (A[i][j] != X[i][j]) return false;
     return true;
   }

   /*
   public static Matrix read (BufferedReader input) throws java.io.IOException {
      StreamTokenizer tokenizer= new StreamTokenizer(input);

      // Although StreamTokenizer will parse numbers, it doesn't recognize
      // scientific notation (E or D); however, Double.valueOf does.
      // The strategy here is to disable StreamTokenizer's number parsing.
      // We'll only get whitespace delimited words, EOL's and EOF's.
      // These words should all be numbers, for Double.valueOf to parse.

      tokenizer.resetSyntax();
      tokenizer.wordChars(0,255);
      tokenizer.whitespaceChars(0, ' ');
      tokenizer.eolIsSignificant(true);
      java.util.Vector v = new java.util.Vector();

      // Ignore initial empty lines
      while (tokenizer.nextToken() == StreamTokenizer.TT_EOL);
      if (tokenizer.ttype == StreamTokenizer.TT_EOF)
	throw new java.io.IOException("Unexpected EOF on matrix read.");
      do {
         v.addElement(Double.valueOf(tokenizer.sval)); // Read & store 1st row.
      } while (tokenizer.nextToken() == StreamTokenizer.TT_WORD);

      int n = v.size();  // Now we've got the number of columns!
      double row[] = new double[n];
      for (int j=0; j<n; j++)  // extract the elements of the 1st row.
         row[j]=((Double)v.elementAt(j)).doubleValue();
      v.removeAllElements();
      v.addElement(row);  // Start storing rows instead of columns.
      while (tokenizer.nextToken() == StreamTokenizer.TT_WORD) {
         // While non-empty lines
         v.addElement(row = new double[n]);
         int j = 0;
         do {
            if (j >= n) throw new java.io.IOException
               ("Row " + v.size() + " is too long.");
            row[j++] = Double.valueOf(tokenizer.sval).doubleValue();
         } while (tokenizer.nextToken() == StreamTokenizer.TT_WORD);
         if (j < n) throw new java.io.IOException
            ("Row " + v.size() + " is too short.");
      }
      int m = v.size();  // Now we've got the number of rows.
      double[][] A = new double[m][];
      v.copyInto(A);  // copy the rows out of the vector
      return new Matrix(A);
   }
   */


  /** Tests the functionality of the whole class.
      If there occurs any error it will be printed on stdout.
  */
  public static void testIt () {
    System.out.println("start testing Matrix4 ...");

    final int repeats = 100;
    Matrix4 A, B, C, D;
    Random rnd = new Random();

    // test the matrix addition
    A = new Matrix4(3,3);
    A.set(rnd);
    B = new Matrix4(3,3);
    B.set(rnd);
    C = A.add(B);
    for (int k = 0; k < 3; k++) {
      for (int j = 0; j < 3; j++) {
        if (C.get(k,j) != Field4.add(B.get(k,j),A.get(k,j)))
          System.out.println("Error 1.1 with " + A.get(k,j) + " " + B.get(k,j) + " " + C.get(k,j));
      }
    }

    // test matrix multiplication
    A = new Matrix4(new int[][] {{0, 1}, {2, 3}});
    B = new Matrix4(new int[][] {{0, 3}, {1, 2}});
    C = new Matrix4(new int[][] {{1, 2}, {3, 0}});
    D = A.mul(B);
    if (!C.isEqual(D)) System.out.println("Error 1.2");
    for (int i = 1; i < repeats; i++) {
      int n = rnd.nextInt(10)+1;
      A = new Matrix4(n,n);
      A.set(rnd);
      B = new Matrix4(n,n);
      B.set(rnd);
      C = identity(n,n);
      D = A.mul(C);  // multiplication from left
      C = C.mul(B);  // and multiplication from right
      if (!D.isEqual(A)) System.out.println("Error 1.3");
      if (!C.isEqual(B)) System.out.println("Error 1.4");
    }

    // test transposition
    for (int i = 0; i < repeats; i++) {
      A = new Matrix4(rnd.nextInt(10)+1,rnd.nextInt(10)+1);
      A.set(rnd);
      B = A.transpose().transpose();
      if (!A.isEqual(B)) System.out.println("Error 1.5");
    }

    // test addRowsEquals
    for (int i = 0; i < repeats; i++) {
      int m = rnd.nextInt(10)+1;
      int n = rnd.nextInt(10)+1;
      A = new Matrix4(m,n);
      B = A.copy();
      int k = rnd.nextInt(m);
      int l = rnd.nextInt(m);
      A.addRowsEquals(k,l);
      B.addRowsEquals(1,k,l);
      if (!A.isEqual(B)) System.out.println("Error 1.6a");
      if (k == l) {
        B.setRowZero(k);
        if (!A.isEqual(B)) System.out.println("Error 1.6b");
      }
    }

    // test vector mul
    for (int i = 0; i < repeats; i++) {
      Object tmp[] = LUDecomposition4.generateNonsingular(3,rnd);
      A = (Matrix4)tmp[0];
      LUDecomposition4 lu = (LUDecomposition4)tmp[1];
      B = new Matrix4(3,1); B.set(rnd);
      int[] res = new int[3];
      int[] b = B.getColumnPackedCopy();

      C = A.mul(B);
      A.mul(b,res);
      for (int j = 0; j < 3; j++) {
        if (res[j] != C.get(j,0)) System.out.println("Error 2.1 at " + j);
      }
      D = lu.solve(C);
      if (!D.isEqual(B)) System.out.println("Error 2.2");
    }

    System.out.println("...finished testing.");
  }

  /** Tests everything in this class
     @see #testIt
  */
  public static void main(String[] args) {
    testIt();
  }

/* ------------------------
   Private Methods
 * ------------------------ */

   /** Check if size(A) == size(B) **/

   private void checkMatrixDimensions (Matrix4 B) {
      if (B.m != m || B.n != n) {
         throw new IllegalArgumentException("Matrix dimensions must agree.");
      }
   }

}
