package HFE;

import java.util.*;
import HFE.*;

//-----------------------------------------------------------------------------
// Project:     Implementing HFE at UCC as Project CS4095
// Supervisors: Simon Foley (Computer Science) and Patrick Fitzpatrick (Maths)
//-----------------------------------------------------------------------------

/**
<p>
This class deals with multi-variable polynomials and
a general modulus provied to the constructor.
Basicly, this class calculates only powers of the form
q^k where q is the number of elements in the field (here: 2)
and k is a non-negative integer (e.g. 0, 1, 2)
All calculations are done in GF(2) ({@link HFE.Field2})
Every entry is a polynomial of the form
f_i(x1,x2,...) = g^i*(b_0 + b_1*x_1 + b_2*x_2 + ... + b_n*x_n)
The reduction is done in the variable g.
<p>
@author Christopher Wolf (hfe@Christopher-Wolf.de)
@version 0.1
*/

//Copyright (C) 2002    Christopher Wolf
//
//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License (file LICENCE) for more details.


class Multi2Power {

  /** Stores all the results.
      Every power has its own polynomials. In resArray[0] are the powers of q^0,
      in resArray[1] the powers of q^1, .... resArray[i][0] is the polynomial for the
      output for 1, resArray[i][1] for gamma, resArray[i][2] for gamma^2,...
    @serial 3D array
  */
  private Poly2MultiLinear[][] resArray;

  /** Initializes the internal fields
      @param modulus  polynomial over GF(2) for reduction
                      the number of variables must be 3 for our purpose
      @param start    multivariable polynomials over GF(2) for q^0.
      @param maxpower till which power shall we compute?
  */
  Multi2Power(Poly2Max modulus, Poly2MultiLinear[] start, int maxPower) {
    Check.assert(modulus.degree() == HFE2_n.extensionDegree);
    Check.assert(start.length == HFE2_n.extensionDegree);
    Check.assert(maxPower > 0);

    int variables = HFE2_n.extensionDegree; // number of variables in the multivariant polynomial

    // init resArray
    resArray = new Poly2MultiLinear[maxPower+1][variables];
    resArray[0] = start;  // get the first power (which is power 0)

    // now compute the reduction function where necessary
    // this is 1, 2, ..., n for our purpose in GF(2) -> GF(2^n)
    Poly2Max[] reduce = new Poly2Max[variables];
    reduce[0] = new Poly2Max(new int[] {0});  // set reduce to be constant polynomial 1, i.e. set coefficient_0 to be 1!
    for (int i = 1; i < variables; i++) {
      reduce[i] = new Poly2Max(reduce[i-1],Field2.elements);  // use the shift constructor to compute (reduce[i-1])^2
      reduce[i].mod(modulus);  // now reduce the result
    }

    // produce the entries for the resArray
    // one entry for every power
    for (int i = 1; i <= maxPower; i++) {
      // produce one entry for every power of gamma
      for (int j = 0; j < variables; j++)
        resArray[i][j] = new Poly2MultiLinear(variables);
      // copy the entries from the last power using the right reduction
      for (int j = 0; j < variables; j++)
        for (int k = 0; k < variables; k++)
          if (reduce[j].coeff(k) != 0)
            resArray[i][k].add(resArray[i-1][j]);
    }
  }


  /** Returns the polynomial vector for a given power which.
      Can only return values which have been computed by the constructor.
  */
  public Poly2MultiLinear[] givePower (int which) {
    Check.assert(which < resArray.length);
    return resArray[which];
  }

  /** Tests the functionality of the class.
      Is called by the main method.
  */
  public static void testIt () {
    System.out.println("start testing Multi2Power ...");

    Random rnd = new Random();
    int maxPow = 5;

    Transform2 trans = new Transform2();  // identity transformation
    Poly2MultiLinear[] start = trans.getTransform();
    Poly2MultiLinear[] power;
    Field2_n u = new Field2_n();
    Field2_n tmp = new Field2_n(), res = new Field2_n();
    int[] u_vec = new int[Field2_n.degree];
    int[] res_vec = new int[Field2_n.degree];
    int[] value_vec = new int[Field2_n.degree+1];

    u.getElement(u_vec);  // get u in terms of a bit vector
    // we need this as apply does not take value 0
    for (int i = 0; i < Field2_n.degree; i++)
      value_vec[i+1] = u_vec[i];

    // some pre-tests
    u = new Field2_n(2);
    tmp.copy(u);
    u.mul(tmp,tmp);  // compute the square of u
    tmp = new Field2_n(4);
    if (!tmp.isEqual(u)) System.out.println("Error 1.0  with " + tmp.stringIt() + " and " + u.stringIt());

    // now try the identity transformation with random u
    Multi2Power testObj = new Multi2Power(HFE2_n.generatingPolynomialAsMaxPoly,start,maxPow);  // compute powers up to 5
    u = new Field2_n(2);
    u.getElement(u_vec);  // get u in terms of a bit vector
    // we need this as apply does not take value 0
    for (int i = 0; i < Field2_n.degree; i++)
      value_vec[i+1] = u_vec[i];

    // compute different powers of u
    for (int pow = 0; pow <= maxPow; pow++) {
      power = testObj.givePower(pow);
      // apply powers to u_vec

      u.getElement(u_vec);  // get u in terms of a bit vector
      for (int i = 0; i < Field2_n.degree; i++) {
        res_vec[i] = power[i].apply(value_vec);
        if (res_vec[i] != u_vec[i])
          System.out.println("Error 1.1 at pos " + i + " with res " + res_vec[i] + " and u " + u_vec[i] + " for power 2^" + pow);
      }

      tmp.copy(u);
      u.mul(tmp,tmp);  // compute the square of u
    }

    // now try the identity transformation with random u
    // -> already defined! Multi2Power testObj = new Multi2Power(HFE2_n.generatingPolynomialAsMaxPoly,start,maxPow);  // compute powers up to 5
    u.rnd(rnd);
    u.getElement(u_vec);  // get u in terms of a bit vector
    // we need this as apply does not take value 0
    for (int i = 0; i < Field2_n.degree; i++)
      value_vec[i+1] = u_vec[i];

    // compute different powers of u
    for (int pow = 0; pow <= maxPow; pow++) {
      power = testObj.givePower(pow);
      // apply powers to u_vec

      u.getElement(u_vec);  // get u in terms of a bit vector
      for (int i = 0; i < Field2_n.degree; i++) {
        res_vec[i] = power[i].apply(value_vec);
        if (res_vec[i] != u_vec[i])
          System.out.println("Error 2.0 at " + i + " with " + res_vec[i] + " and " + u_vec[i] + " for " + pow);
      }

      tmp.copy(u);
      u.mul(tmp,tmp);  // compute the square of u
    }

    // last but not least a random transformation with random u
    trans = new Transform2(rnd);  // identity transformation
    start = trans.getTransform();
    testObj = new Multi2Power(HFE2_n.generatingPolynomialAsMaxPoly,start,maxPow);  // compute powers up to 5
    u.rnd(rnd);
    u.getElement(u_vec);  // get u in terms of a bit vector
    // we need this as Poly2MultiLinear.apply does not take value[0]
    for (int i = 0; i < Field2_n.degree; i++)
      value_vec[i+1] = u_vec[i];
    // now apply trans to u
    trans.transform(u_vec,res_vec);
    u.setElement(res_vec);

    // compute different powers of u
    for (int pow = 0; pow <= maxPow; pow++) {
      power = testObj.givePower(pow);
      // apply powers to u_vec

      u.getElement(u_vec);  // get u in terms of a bit vector
      for (int i = 0; i < Field2_n.degree; i++) {
        res_vec[i] = power[i].apply(value_vec);
        if (res_vec[i] != u_vec[i])
          System.out.println("Error 3.0 at " + i + " with " + res_vec[i] + " and " + u_vec[i] + " for " + pow);
      }

      tmp.copy(u);
      u.mul(tmp,tmp);  // compute the square of u
    }

    System.out.println("...finished testing.");
  }

  /** Calls the {@link #testIt} method
  */
  public static void main(String[] args) {
    testIt();
  }
}
