package HFE;

import java.util.*;
import java.lang.Math;
import HFE.*;

//-----------------------------------------------------------------------------
// Project:     Implementing HFE at UCC as Project CS4095
// Supervisors: Simon Foley (Computer Science) and Patrick Fitzpatrick (Maths)
//-----------------------------------------------------------------------------

/**
<p>
This class deals with multi-variable polynomials and
a general modulus provied to the constructor.
Basicly, this class calculates multiples of two polynomials
and one polynomial and one element of (F_4)^3.
All calculations are done in F_4 ({@link HFE.Field4})
Every entry is a polynomial of the form
f_i(x1,x2,...) = g^i*(b_0 + b_1*x_1 + b_2*x_2 + ... + b_n*x_n)
The reduction is done in the variable g.
<p>
@author Christopher Wolf (hfe@Christopher-Wolf.de)
@version 0.1
*/

//Copyright (C) 2002    Christopher Wolf
//
//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License (file LICENCE) for more details.


class Multi4Mul {

  /** Stores all the reduction steps.
      reduce[0] tells how to reduce the first polynomial,
      reduce[1] who to reduce the second polynomial, ...
    @serial 3D array
    @version 17.12.01
  */
  private Poly4Max[] reduce;

  /** Initializes the internal field {@link #reduce}
      @param modulus  polynomial over F_4 for reduction
                      the number of variables must be 3 for our purpose
  */
  Multi4Mul(Poly4Max modulus) {
    Check.assert(modulus.degree() == Poly64Max.extension);

    int degree = Poly64Max.extension; // number of variables in the multivariant polynomials we deal with
    int entryNum = (2*degree)-1;      // how many entries do we need?

    // init reduce array
    reduce = new Poly4Max[entryNum];
    reduce[0] = new Poly4Max(new int[] {1,0});  // our first entry is the constant polynomial 1
    for (int i = 1; i < entryNum; i++) {
      reduce[i] = new Poly4Max(reduce[i-1],1);  // use the last polynomial and multiply it by x using the shift-constructor
      reduce[i].mod(modulus);
    }
  }

  /** Computes res = a * b
      in terms of vector multiplication of polynomials. Uses {@link #reduce} to
      take the modulus into account.
  */
  public Poly4MultiSquare[] mul(Poly4MultiLinear[] a, Poly4MultiLinear[] b) {
    final int dimension = Poly64Max.extension;
    Check.assert(dimension == a.length);
    Check.assert(dimension == b.length);

    // get one temporary register
    Poly4MultiSquare tmp = new Poly4MultiSquare(dimension);

    // init our result vector
    Poly4MultiSquare[] res = new Poly4MultiSquare[dimension];
    for (int i = 0; i < dimension; i++)
      res[i] = new Poly4MultiSquare(dimension);

    // compute the result using the reduction polynomial
    // we go throuth the matrix from left upper to right lower
    // in diagonal manner. All results with the same g-power
    // are done in one step
    for (int i = 0; i < reduce.length; i++) {
      // init our start values
      int startValueB = Math.max(0,i-dimension+1);
      int aVal = Math.min(i,dimension-1);
      int bVal = startValueB;
      // get all values in diagonal manner
      tmp.setZero();
      while (aVal >= startValueB) {
        Check.assert(aVal+bVal==i);
        tmp.mulAdd(a[aVal],b[bVal]);
        // now produce a new aVal/bVal-pair
        aVal--;
        bVal++;
      }
      // now reduce them
      for (int j = 0; j < dimension; j++) {
        if (reduce[i].coeff(j) != 0)
          res[j].mulAdd(reduce[i].coeff(j),tmp);
      }
    }

    return res;
  }


  /** Computes res = con * a * b
      in terms of vector multiplication of polynomials. Uses {@link #reduce} to
      take the modulus into account.
    @param con constants as a vector, i.e. a vector of polynomials of degree 0 or -1
    @param a linear polynomial in dimension variables
    @param b linear polynimial in dimension variables
  */
  public Poly4MultiSquare[] mul(int[] con, Poly4MultiLinear[] a, Poly4MultiLinear[] b) {
    final int dimension = Poly64Max.extension;
    Check.assert(dimension == a.length, "a has wrong length, namely " + a.length);
    Check.assert(dimension == b.length, "b has wrong length, namely " + b.length);
    Check.assert(dimension == con.length, "con has wrong length, namely " + con.length);
    Check.assert(Field4.isElement(con), "con is not over Field4.");

    // get temporary registers
    Poly4MultiSquare tmpS = new Poly4MultiSquare(dimension);
    Poly4MultiLinear tmpL = new Poly4MultiLinear(dimension);

    // init our first result vector
    Poly4MultiLinear[] resL = new Poly4MultiLinear[dimension];
    for (int i = 0; i < dimension; i++)
      resL[i] = new Poly4MultiLinear(dimension);

    // compute the result using the reduction polynomial
    // we go throuth the matrix from left upper to right lower
    // in diagonal manner. All results with the same g-power
    // are done in one step
    for (int i = 0; i < reduce.length; i++) {
      // init our start values
      int startValueB = Math.max(0,i-dimension+1);
      int aVal = Math.min(i,dimension-1);
      int bVal = startValueB;
      // get all values in diagonal manner
      tmpL.setZero();
      while (aVal >= startValueB) {
        Check.assert(aVal+bVal==i);
        tmpL.mulAdd(con[bVal],a[aVal]);
        // now produce a new aVal/bVal-pair
        aVal--;
        bVal++;
      }
      // now reduce them
      for (int j = 0; j < dimension; j++) {
        if (reduce[i].coeff(j) != 0)
          resL[j].mulAdd(reduce[i].coeff(j),tmpL);
      }
    }

    // init our second result vector
    Poly4MultiSquare[] resS = new Poly4MultiSquare[dimension];
    for (int i = 0; i < dimension; i++)
      resS[i] = new Poly4MultiSquare(dimension);

    // compute the result using the reduction polynomial
    // we go throuth the matrix from left upper to right lower
    // in diagonal manner. All results with the same g-power
    // are done in one step
    for (int i = 0; i < reduce.length; i++) {
      // init our start values
      int startValueB = Math.max(0,i-dimension+1);
      int aVal = Math.min(i,dimension-1);
      int bVal = startValueB;
      // get all values in diagonal manner
      tmpS.setZero();
      while (aVal >= startValueB) {
        Check.assert(aVal+bVal==i);
        tmpS.mulAdd(resL[aVal],b[bVal]);
        // now produce a new aVal/bVal-pair
        aVal--;
        bVal++;
      }
      // now reduce them
      for (int j = 0; j < dimension; j++) {
        if (reduce[i].coeff(j) != 0)
          resS[j].mulAdd(reduce[i].coeff(j),tmpS);
      }
    }

    return resS;
  }


 /** Computes res = con * poly
      in terms of vector multiplication of polynomials. Uses {@link #reduce} to
      take the modulus into account.
    @param con constants as a vector, i.e. a vector of polynomials of degree 0 or -1
    @param poly linear polynomial in dimension variables
  */
  public Poly4MultiLinear[] mul(int[] con, Poly4MultiLinear[] poly) {
    final int dimension = Poly64Max.extension;
    Check.assert(dimension == poly.length, "poly has wrong length, namely " + poly.length);
    Check.assert(dimension == con.length, "con has wrong length, namely " + con.length);
    Check.assert(Field4.isElement(con), "con is not over Field4.");

    // get temporary registers
    Poly4MultiLinear tmp = new Poly4MultiLinear(dimension);

    // init our first result vector
    Poly4MultiLinear[] res = new Poly4MultiLinear[dimension];
    for (int i = 0; i < dimension; i++)
      res[i] = new Poly4MultiLinear(dimension);

    // compute the result using the reduction polynomial
    // we go throuth the matrix from left upper to right lower
    // in diagonal manner. All results with the same g-power
    // are done in one step
    for (int i = 0; i < reduce.length; i++) {
      // init our start values
      int startValueB = Math.max(0,i-dimension+1);
      int aVal = Math.min(i,dimension-1);
      int bVal = startValueB;
      // get all values in diagonal manner
      tmp.setZero();
      while (aVal >= startValueB) {
        Check.assert(aVal+bVal==i);
        tmp.mulAdd(con[bVal],poly[aVal]);
        // now produce a new aVal/bVal-pair
        aVal--;
        bVal++;
      }
      // now reduce them
      for (int j = 0; j < dimension; j++) {
        if (reduce[i].coeff(j) != 0)
          res[j].mulAdd(reduce[i].coeff(j),tmp);
      }
    }

    return res;
  }


  /** Computes res = a * b
      in terms of vector multiplication of polynomials. Uses {@link #reduce} to
      take the modulus into account. This is an independent implementation to
      {@link #mul(Poly4MultiLinear[], Poly4MultiLinear[])} and regarded to be slower.
  */
  public Poly4MultiSquare[] mul2(Poly4MultiLinear[] a, Poly4MultiLinear[] b) {
    final int dimension = Poly64Max.extension;
    Check.assert(dimension == a.length, "a has wrong length, namely " + a.length);
    Check.assert(dimension == b.length, "b has wrong length, namely " + b.length);

    // get one temporary register
    Poly4MultiSquare tmp = new Poly4MultiSquare(dimension);

    // init our result vector
    Poly4MultiSquare[] res = new Poly4MultiSquare[dimension];
    for (int i = 0; i < dimension; i++)
      res[i] = new Poly4MultiSquare(dimension);

    // compute the result using the reduction polynomial
    // first take care about the elements on the diagonal
    for (int i = 0; i < dimension; i++) {
      tmp.mul(a[i],b[i]);
      for (int j = 0; j < dimension; j++) {
        if (reduce[2*i].coeff(j) != 0)
          res[j].mulAdd(reduce[2*i].coeff(j),tmp);
      }
    }
    // now compute the non-diagonal elements
    for (int i = 0; i < (dimension-1); i++) {
      for (int j = i+1; j < dimension; j++) {
        tmp.mul(a[i],b[j]);
        tmp.mulAdd(a[j],b[i]);
        for (int k = 0; k < dimension; k++) {
          if (reduce[i+j].coeff(k) != 0)
            res[k].mulAdd(reduce[i+j].coeff(k),tmp);
        }
      }
    }

    return res;
  }


  /** Computes res = con * a * b
      in terms of vector multiplication of polynomials. Uses {@link #reduce} to
      take the modulus into account. This is an independent implementation to
      {@link #mul(int[], Poly4MultiLinear[], Poly4MultiLinear[])} and regarded to be slower.
  */
  public Poly4MultiSquare[] mul2(int[] con, Poly4MultiLinear[] a, Poly4MultiLinear[] b) {
    final int dimension = Poly64Max.extension;
    Check.assert(dimension == a.length, "a has wrong length, namely " + a.length);
    Check.assert(dimension == b.length, "b has wrong length, namely " + b.length);
    Check.assert(dimension == con.length, "con has wrong length, namely " + con.length);
    Check.assert(Field4.isElement(con), "con is not over Field4.");

    // get one temporary register
    Poly4MultiSquare tmp = new Poly4MultiSquare(dimension);

    // init our result vector
    Poly4MultiSquare[] res1 = new Poly4MultiSquare[dimension];
    for (int i = 0; i < dimension; i++)
      res1[i] = new Poly4MultiSquare(dimension);

    // compute the result using the reduction polynomial
    // first take care about the elements on the diagonal
    for (int i = 0; i < dimension; i++) {
      tmp.mul(a[i],b[i]);
      for (int j = 0; j < dimension; j++) {
        if (reduce[2*i].coeff(j) != 0)
          res1[j].mulAdd(reduce[2*i].coeff(j),tmp);
      }
    }
    // now compute the non-diagonal elements
    for (int i = 0; i < (dimension-1); i++) {
      for (int j = i+1; j < dimension; j++) {
        tmp.mul(a[i],b[j]);
        tmp.mulAdd(a[j],b[i]);
        for (int k = 0; k < dimension; k++) {
          if (reduce[i+j].coeff(k) != 0)
            res1[k].mulAdd(reduce[i+j].coeff(k),tmp);
        }
      }
    }

    // the last step is to compute the result
    // in terms of the constant polynomials
    // init our final result vector
    Poly4MultiSquare[] res2 = new Poly4MultiSquare[dimension];
    for (int i = 0; i < dimension; i++)
      res2[i] = new Poly4MultiSquare(dimension);

    // compute the result using the reduction polynomial
    // first take care about the elements on the diagonal
    for (int i = 0; i < dimension; i++) {
      tmp.setZero();
      tmp.mulAdd(con[i],res1[i]);
      for (int j = 0; j < dimension; j++) {
        if (reduce[2*i].coeff(j) != 0)
          res2[j].mulAdd(reduce[2*i].coeff(j),tmp);
      }
    }
    // now compute the non-diagonal elements
    for (int i = 0; i < (dimension-1); i++) {
      for (int j = i+1; j < dimension; j++) {
        tmp.setZero();
        tmp.mulAdd(con[i],res1[j]);
        tmp.mulAdd(con[j],res1[i]);
        for (int k = 0; k < dimension; k++) {
          if (reduce[i+j].coeff(k) != 0)
            res2[k].mulAdd(reduce[i+j].coeff(k),tmp);
        }
      }
    }
    return res2;
  }

  /** Tests the functionality of the class.
      Is called by the main method.
  */
  public static void testIt () {
    System.out.println("start testing Multi4Mul ...");

    Random rnd = new Random();
    final int repeats = 10;   // how often do we run our tests
    final int variables = 3;  // must be 3 at this configuration: 4^3 = 64

    // test if shift and power are equal
    final int maxPow = 20;

    Poly4Max[] shifts = new Poly4Max[maxPow];
    Poly4Max[] powers = new Poly4Max[maxPow];
    shifts[0] = new Poly4Max(new int[] {1,0});
    for (int i = 1; i < maxPow; i++) {
      shifts[i] = new Poly4Max(shifts[i-1],1);
      shifts[i].mod(Poly64Max.genPoly);
    }

    Poly4Max.setModulus(Poly64Max.genPoly);
    for (int i = 0; i < maxPow; i++) {
      powers[i] = new Poly4Max(new int[] {1,i});
      powers[i].mod(Poly64Max.genPoly);
    }

    for (int i = 0; i < maxPow; i++) {
      if (!powers[i].isEqual(shifts[i])) System.out.println("Error 1.1 at " + i);
    }

    Multi4Mul testObj = new Multi4Mul(Poly64Max.genPoly);
    // check if we get the right entries in reduce
    for (int i = 0; i < Poly64Max.extension; i++)
      if (!testObj.reduce[i].isEqual(powers[i])) System.out.println("Error 1.2 at " + i);


    // now start with different test cases for the multiplication procedure
    Poly4MultiLinear[] left = new Poly4MultiLinear[variables], right = new Poly4MultiLinear[variables],
                       degree0 = new Poly4MultiLinear[variables];
    Poly4MultiSquare[] res1 = new Poly4MultiSquare[variables], res2 = new Poly4MultiSquare[variables],
                       cmp = new Poly4MultiSquare[variables];

    // start with a very simple example
    left[0] = new Poly4MultiLinear(new int[] {0,0,1,0});  // x_2
    left[1] = new Poly4MultiLinear(new int[] {0,0,0,0});  // 0
    left[2] = new Poly4MultiLinear(new int[] {0,0,0,0});  // 0

    right[0] = new Poly4MultiLinear(new int[] {1,0,0,0}); // 1
    right[1] = new Poly4MultiLinear(new int[] {0,0,0,0}); // 0
    right[2] = new Poly4MultiLinear(new int[] {0,0,0,0});

    cmp[0] = new Poly4MultiSquare(new int[] {0,0,1,0,0,0,0,0,0,0});  // x_2
    cmp[1] = new Poly4MultiSquare(new int[] {0,0,0,0,0,0,0,0,0,0});  // 0
    cmp[2] = new Poly4MultiSquare(new int[] {0,0,0,0,0,0,0,0,0,0});  // 0

    res1 = testObj.mul(left,right);
    res2 = testObj.mul(right,left);
    for (int i = 0; i < Poly64Max.extension; i++) {
      if (!cmp[i].isEqual(res1[i])) System.out.println("Error 2.1a at " + i);
      if (!cmp[i].isEqual(res2[i])) System.out.println("Error 2.1b at " + i);
    }

    // next, a little more complicated example
    left[0] = new Poly4MultiLinear(new int[] {0,0,0,0});  // 0
    left[1] = new Poly4MultiLinear(new int[] {0,1,0,3});  // g(x_1 + 3x_3)
    left[2] = new Poly4MultiLinear(new int[] {0,0,0,0});  // 0

    right[0] = new Poly4MultiLinear(new int[] {0,0,0,0});  // 0
    right[1] = new Poly4MultiLinear(new int[] {0,1,2,0});  // g(x_1 + 2x_2)
    right[2] = new Poly4MultiLinear(new int[] {0,0,0,0});  // 0

    cmp[0] = new Poly4MultiSquare(new int[] {0,0,0,0,0,0,0,0,0,0});  // 0
    cmp[1] = new Poly4MultiSquare(new int[] {0,0,0,0,0,0,0,0,0,0});  // 0
    cmp[2] = new Poly4MultiSquare(new int[] {0,0,0,0,1,2,3,0,1,0});  // g^2 (x_1^2 + 2x_1x_2 + 3x_1x_3 + x_2x_3)

    res1 = testObj.mul(left,right);
    res2 = testObj.mul(right,left);
    for (int i = 0; i < Poly64Max.extension; i++) {
      if (!cmp[i].isEqual(res1[i])) {
        System.out.println("Error 2.2a at " + i);
        System.out.print("Got:      "); res1[i].writeString(); System.out.println();
        System.out.print("Expected: ");  cmp[i].writeString(); System.out.println();
      }
      if (!cmp[i].isEqual(res2[i])) {
        System.out.println("Error 2.2b at " + i);
        System.out.print("Got:      "); res2[i].writeString(); System.out.println();
        System.out.print("Expected: ");  cmp[i].writeString(); System.out.println();
      }
    }

    // an intermediate hand calculation example
    left[0] = new Poly4MultiLinear(new int[] {0,0,0,0});
    left[1] = new Poly4MultiLinear(new int[] {0,0,0,0});
    left[2] = new Poly4MultiLinear(new int[] {0,1,0,0});   // g^2(x_1)

    right[0] = new Poly4MultiLinear(new int[] {0,0,0,0});
    right[1] = new Poly4MultiLinear(new int[] {0,0,1,0});  // g(x_2)
    right[2] = new Poly4MultiLinear(new int[] {0,0,0,1});  // g^2(x_3)

    cmp[0] = new Poly4MultiSquare(new int[] {0,0,0,0,0,3,3,0,0,0});
    cmp[1] = new Poly4MultiSquare(new int[] {0,0,0,0,0,2,1,0,0,0});
    cmp[2] = new Poly4MultiSquare(new int[] {0,0,0,0,0,1,3,0,0,0});

    res1 = testObj.mul(left,right);
    res2 = testObj.mul(right,left);
    for (int i = 0; i < Poly64Max.extension; i++) {
      if (!cmp[i].isEqual(res1[i])) {
        System.out.println("Error 2.3a at " + i);
        System.out.print("Got:      "); res1[i].writeString(); System.out.println();
        System.out.print("Expected: ");  cmp[i].writeString(); System.out.println();
      }
      if (!cmp[i].isEqual(res2[i])) {
        System.out.println("Error 2.3b at " + i);
        System.out.print("Got:      "); res2[i].writeString(); System.out.println();
        System.out.print("Expected: ");  cmp[i].writeString(); System.out.println();
      }
    }

    // last but not least the most complicated thing I wanted to compute by hand
    left[0] = new Poly4MultiLinear(new int[] {0,0,0,0});
    left[1] = new Poly4MultiLinear(new int[] {0,0,0,0});
    left[2] = new Poly4MultiLinear(new int[] {0,0,1,1});

    right[0] = new Poly4MultiLinear(new int[] {0,0,0,0});
    right[1] = new Poly4MultiLinear(new int[] {0,0,1,2});
    right[2] = new Poly4MultiLinear(new int[] {0,3,2,0});

    cmp[0] = new Poly4MultiSquare(new int[] {0,0,0,0,0,2,2,2,3,1});
    cmp[1] = new Poly4MultiSquare(new int[] {0,0,0,0,0,3,3,0,3,3});
    cmp[2] = new Poly4MultiSquare(new int[] {0,0,0,0,0,2,2,0,2,2});

    res1 = testObj.mul(left,right);
    res2 = testObj.mul(right,left);
    for (int i = 0; i < Poly64Max.extension; i++) {
      if (!cmp[i].isEqual(res1[i])) {
        System.out.println("Error 2.4a at " + i);
        System.out.print("Got:      "); res1[i].writeString(); System.out.println();
        System.out.print("Expected: ");  cmp[i].writeString(); System.out.println();
      }
      if (!cmp[i].isEqual(res2[i])) {
        System.out.println("Error 2.4b at " + i);
        System.out.print("Got:      "); res2[i].writeString(); System.out.println();
        System.out.print("Expected: ");  cmp[i].writeString(); System.out.println();
      }
    }

    // test if mul == mul2
    for (int i = 0; i < repeats; i++) {
      int[] a = new int[variables+1];
      int[] b = new int[variables+1];
      int fac = Field4.rnd(rnd);
      for (int k = 0; k < variables; k++) {
        for (int j = 0; j < a.length; j++) {
          a[j] = Field4.rnd(rnd);
          b[j] = Field4.rnd(rnd);
        }
        left[k] = new Poly4MultiLinear(a);
        right[k] = new Poly4MultiLinear(b);
      }
      cmp = testObj.mul(left,right);       // mul
      res1 = testObj.mul2(left,right);     // mul2
      res2 = testObj.mul2(right,left);
      for (int j = 0; j < Poly64Max.extension; j++) {
        if (!cmp[j].isEqual(res1[j])) {
          System.out.println("Error 3.1 at " + j);
          System.out.print("Got:      "); res1[j].writeString(); System.out.println();
          System.out.print("Expected: ");  cmp[j].writeString(); System.out.println();
        }
        if (!cmp[j].isEqual(res2[j])) {
          System.out.println("Error 3.2 at " + j);
          System.out.print("Got:      "); res2[j].writeString(); System.out.println();
          System.out.print("Expected: ");  cmp[j].writeString(); System.out.println();
        }
      }
    }

    // test if the "triple mul" works
    int[] vec;

    // start with a very simple example

    vec = new int[] {1,0,0};    // 1

    left[0] = new Poly4MultiLinear(new int[] {0,0,1,0});  // x_2
    left[1] = new Poly4MultiLinear(new int[] {0,0,0,0});  // 0
    left[2] = new Poly4MultiLinear(new int[] {0,0,0,0});  // 0

    right[0] = new Poly4MultiLinear(new int[] {1,0,0,0}); // 1
    right[1] = new Poly4MultiLinear(new int[] {0,0,0,0}); // 0
    right[2] = new Poly4MultiLinear(new int[] {0,0,0,0});

    cmp[0] = new Poly4MultiSquare(new int[] {0,0,1,0,0,0,0,0,0,0});  // x_2
    cmp[1] = new Poly4MultiSquare(new int[] {0,0,0,0,0,0,0,0,0,0});  // 0
    cmp[2] = new Poly4MultiSquare(new int[] {0,0,0,0,0,0,0,0,0,0});  // 0

    res1 = testObj.mul(vec,left,right);
    for (int i = 0; i < Poly64Max.extension; i++) {
      if (!cmp[i].isEqual(res1[i])) {
        System.out.println("Error 4.1 at " + i);
        System.out.print("Got:      "); res1[i].writeString(); System.out.println();
        System.out.print("Expected: ");  cmp[i].writeString(); System.out.println();
      }
    }

    // get a little more sophisticated
    vec = new int[] {0,1,0};    // g

    left[0] = new Poly4MultiLinear(new int[] {0,0,1,0});  // x_2
    left[1] = new Poly4MultiLinear(new int[] {0,0,0,0});  // 0
    left[2] = new Poly4MultiLinear(new int[] {0,0,0,0});  // 0

    right[0] = new Poly4MultiLinear(new int[] {1,0,0,0}); // 1
    right[1] = new Poly4MultiLinear(new int[] {0,0,0,0}); // 0
    right[2] = new Poly4MultiLinear(new int[] {0,0,0,0}); // 0

    cmp[0] = new Poly4MultiSquare(new int[] {0,0,0,0,0,0,0,0,0,0});  // 0
    cmp[1] = new Poly4MultiSquare(new int[] {0,0,1,0,0,0,0,0,0,0});  // x_2
    cmp[2] = new Poly4MultiSquare(new int[] {0,0,0,0,0,0,0,0,0,0});  // 0

    res1 = testObj.mul(vec,left,right);
    for (int i = 0; i < Poly64Max.extension; i++) {
      if (!cmp[i].isEqual(res1[i])) {
        System.out.println("Error 4.2 at " + i);
        System.out.print("Got:      "); res1[i].writeString(); System.out.println();
        System.out.print("Expected: ");  cmp[i].writeString(); System.out.println();
      }
    }


    // we test reduction now
    vec = new int[] {0,1,0};    // g

    left[0] = new Poly4MultiLinear(new int[] {0,0,0,0});  // 1 (0)
    left[1] = new Poly4MultiLinear(new int[] {0,0,1,0});  // g (x_2)
    left[2] = new Poly4MultiLinear(new int[] {0,0,0,0});  // g^2 (0)

    right[0] = new Poly4MultiLinear(new int[] {0,0,0,0}); // 0
    right[1] = new Poly4MultiLinear(new int[] {1,0,0,0}); // 1
    right[2] = new Poly4MultiLinear(new int[] {0,0,0,0}); // 0

    cmp[0] = new Poly4MultiSquare(new int[] {0,0,3,0,0,0,0,0,0,0});  // 3x_2
    cmp[1] = new Poly4MultiSquare(new int[] {0,0,2,0,0,0,0,0,0,0});  // 2x_2
    cmp[2] = new Poly4MultiSquare(new int[] {0,0,1,0,0,0,0,0,0,0});  // x_2

    res1 = testObj.mul(vec,left,right);
        for (int i = 0; i < Poly64Max.extension; i++) {
      if (!cmp[i].isEqual(res1[i])) {
        System.out.println("Error 4.3 at " + i);
        System.out.print("Got:      "); res1[i].writeString(); System.out.println();
        System.out.print("Expected: ");  cmp[i].writeString(); System.out.println();
      }
    }

    // next, a little more complicated example
    vec = new int[] {1,0,0};    // 1

    left[0] = new Poly4MultiLinear(new int[] {0,0,0,0});  // 0
    left[1] = new Poly4MultiLinear(new int[] {0,1,0,3});  // g(x_1 + 3x_3)
    left[2] = new Poly4MultiLinear(new int[] {0,0,0,0});  // 0

    right[0] = new Poly4MultiLinear(new int[] {0,0,0,0});  // 0
    right[1] = new Poly4MultiLinear(new int[] {0,1,2,0});  // g(x_1 + 2x_2)
    right[2] = new Poly4MultiLinear(new int[] {0,0,0,0});  // 0

    cmp[0] = new Poly4MultiSquare(new int[] {0,0,0,0,0,0,0,0,0,0});  // 0
    cmp[1] = new Poly4MultiSquare(new int[] {0,0,0,0,0,0,0,0,0,0});  // 0
    cmp[2] = new Poly4MultiSquare(new int[] {0,0,0,0,1,2,3,0,1,0});  // g^2 (x_1^2 + 2x_1x_2 + 3x_1x_3 + x_2x_3)
    res1 = testObj.mul(vec,left,right);
    for (int i = 0; i < Poly64Max.extension; i++) {
      if (!cmp[i].isEqual(res1[i])) {
        System.out.println("Error 4.4 at " + i);
        System.out.print("Got:      "); res1[i].writeString(); System.out.println();
        System.out.print("Expected: ");  cmp[i].writeString(); System.out.println();
      }
    }

    // the most complicated thing I wanted to compute by hand
    vec = new int[] {1,0,0};    // 1

    left[0] = new Poly4MultiLinear(new int[] {0,0,0,0});
    left[1] = new Poly4MultiLinear(new int[] {0,0,0,0});
    left[2] = new Poly4MultiLinear(new int[] {0,0,1,1});

    right[0] = new Poly4MultiLinear(new int[] {0,0,0,0});
    right[1] = new Poly4MultiLinear(new int[] {0,0,1,2});
    right[2] = new Poly4MultiLinear(new int[] {0,3,2,0});

    cmp[0] = new Poly4MultiSquare(new int[] {0,0,0,0,0,2,2,2,3,1});
    cmp[1] = new Poly4MultiSquare(new int[] {0,0,0,0,0,3,3,0,3,3});
    cmp[2] = new Poly4MultiSquare(new int[] {0,0,0,0,0,2,2,0,2,2});
    res1 = testObj.mul(vec,left,right);
    for (int i = 0; i < Poly64Max.extension; i++) {
      if (!cmp[i].isEqual(res1[i])) {
        System.out.println("Error 4.5 at " + i);
        System.out.print("Got:      "); res1[i].writeString(); System.out.println();
        System.out.print("Expected: ");  cmp[i].writeString(); System.out.println();
      }
    }

    // we concentrate on different values of vec now
    vec = new int[] {0,0,2};    // 2g^2

    left[0] = new Poly4MultiLinear(new int[] {0,1,0,0});  // x_1
    left[1] = new Poly4MultiLinear(new int[] {0,0,0,0});  // 0
    left[2] = new Poly4MultiLinear(new int[] {0,0,0,0});  // 0

    right[0] = new Poly4MultiLinear(new int[] {0,1,0,0}); // x_1
    right[1] = new Poly4MultiLinear(new int[] {0,0,0,0}); // 0
    right[2] = new Poly4MultiLinear(new int[] {0,0,0,0}); // 0

    cmp[0] = new Poly4MultiSquare(new int[] {0,0,0,0,0,0,0,0,0,0});
    cmp[1] = new Poly4MultiSquare(new int[] {0,0,0,0,0,0,0,0,0,0});
    cmp[2] = new Poly4MultiSquare(new int[] {0,0,0,0,2,0,0,0,0,0}); // g^2: 2(x_1)^2
    res1 = testObj.mul(vec,left,right);
    for (int i = 0; i < Poly64Max.extension; i++) {
      if (!cmp[i].isEqual(res1[i])) {
        System.out.println("Error 4.6 at " + i);
        System.out.print("Got:      "); res1[i].writeString(); System.out.println();
        System.out.print("Expected: ");  cmp[i].writeString(); System.out.println();
      }
    }

    // last but not least
    vec = new int[] {1,2,3};    // 1 + 2g + 3g^2

    left[0] = new Poly4MultiLinear(new int[] {0,1,0,0});  // x_1
    left[1] = new Poly4MultiLinear(new int[] {0,0,1,0});  // x_2
    left[2] = new Poly4MultiLinear(new int[] {0,0,0,0});  // 0

    right[0] = new Poly4MultiLinear(new int[] {0,0,0,0}); //
    right[1] = new Poly4MultiLinear(new int[] {0,0,0,1}); // x_3
    right[2] = new Poly4MultiLinear(new int[] {0,0,0,0}); // 0

    cmp[0] = new Poly4MultiSquare(new int[] {0,0,0,0,0,0,2,0,3,0}); //   1: 2x_1,3 + 3x_2_3
    cmp[1] = new Poly4MultiSquare(new int[] {0,0,0,0,0,0,0,0,0,0}); //   g: 0
    cmp[2] = new Poly4MultiSquare(new int[] {0,0,0,0,0,0,1,0,1,0}); // g^2: x_1,3 + x_2,3
    res1 = testObj.mul(vec,left,right);
    for (int i = 0; i < Poly64Max.extension; i++) {
      if (!cmp[i].isEqual(res1[i])) {
        System.out.println("Error 4.6 at " + i);
        System.out.print("Got:      "); res1[i].writeString(); System.out.println();
        System.out.print("Expected: ");  cmp[i].writeString(); System.out.println();
      }
    }

    // try random values for mul and mul2
    for (int i = 0; i < repeats; i++) {
      int[] a = new int[variables+1];
      int[] b = new int[variables+1];
      for (int k = 0; k < variables; k++) {
        vec[k] = Field4.rnd(rnd);
        for (int j = 0; j < a.length; j++) {
          a[j] = Field4.rnd(rnd);
          b[j] = Field4.rnd(rnd);
        }
        left[k] = new Poly4MultiLinear(a);
        right[k] = new Poly4MultiLinear(b);
      }

      res1 = testObj.mul(vec,left,right);      // first mul
      res2 = testObj.mul2(vec,left,right);     // then mul2

      for (int j = 0; j < Poly64Max.extension; j++) {
        if (!res1[j].isEqual(res2[j])) {
          System.out.println("Error 4.8 at " + j);
          System.out.print("Got:      "); res1[j].writeString(); System.out.println();
          System.out.print("Expected: ");  cmp[j].writeString(); System.out.println();
        }
      }
    }
    /**/
    System.out.println("...finished testing.");
  }

  /** Calls the {@link #testIt} method
  */
  public static void main(String[] args) {
    testIt();
  }
}
