package HFE;

import java.util.*;
import HFE.*;

//-----------------------------------------------------------------------------
// Project:     Implementing HFE at UCC as Project CS4095
// Supervisors: Simon Foley (Computer Science) and Patrick Fitzpatrick (Maths)
//-----------------------------------------------------------------------------

/**
<p>
This class deals with multi-variable polynomials and
a general modulus provied to the constructor.
Basicly, this class calculates only powers of the form
q^k where q is the number of elements in the field (here: 4)
and k is a non-negative integer (e.g. 0, 1, 2)
All calculations are done in F_4 ({@link HFE.Field4})
Every entry is a polynomial of the form
f_i(x1,x2,...) = g^i*(b_0 + b_1*x_1 + b_2*x_2 + ... + b_n*x_n)
The reduction is done in the variable g.
<p>
@author Christopher Wolf (hfe@Christopher-Wolf.de)
@version 0.1
*/

//Copyright (C) 2002    Christopher Wolf
//
//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License (file LICENCE) for more details.


class Multi4Power {

  /** Stores all the results.
      Every power has its own polynomials. In resArray[0] are the powers of q^0,
      in resArray[1] the powers of q^1, .... resArray[i][0] is the polynomial for the
      output for 1, resArray[i][1] for gamma, resArray[i][2] for gamma^2,...
    @serial 3D array
  */
  private Poly4MultiLinear[][] resArray;

  /** Initializes the internal fields
      @param modulus  polynomial over F_4 for reduction
                      the number of variables must be 3 for our purpose
      @param start    multivariable polynomials over F_4 for q^0.
      @param maxpower till which power shall we compute?
  */
  Multi4Power(Poly4Max modulus, Poly4MultiLinear[] start, int maxPower) {
    Check.assert(modulus.degree() == Poly64Max.extension);
    Check.assert(start.length == Poly64Max.extension);

    int variables = Poly64Max.extension; // number of variables in the multivariant polynomial

    // init resArray
    resArray = new Poly4MultiLinear[maxPower+1][variables];
    resArray[0] = start;  // get the first power (which is power 0)

    // now compute the reduction function where necessary
    // this is 1, 2 for our purpose in F_4 -> F_64
    Poly4Max[] reduce = new Poly4Max[variables];
    reduce[0] = new Poly4Max(new int[] {1,0});
    for (int i = 1; i < variables; i++) {
      reduce[i] = new Poly4Max(reduce[i-1],Field4.elements);  // use the shift constructor to multiply by g^4
      reduce[i].mod(modulus);  // now reduce the result
    }

    // produce the entries for the resArray
    // one entry for every power
    for (int i = 1; i <= maxPower; i++) {
      // produce one entry for every power of gamma
      for (int j = 0; j < variables; j++)
        resArray[i][j] = new Poly4MultiLinear(variables);
      // copy the entries from the last power using the right reduction
      for (int j = 0; j < variables; j++)
        for (int k = 0; k < variables; k++)
          if (reduce[j].coeff(k) != 0)
            resArray[i][k].mulAdd(reduce[j].coeff(k),resArray[i-1][j]);
    }
  }


  /** Returns the polynomial vector for a given power which.
      Can only return values which have been computed by the constructor.
  */
  public Poly4MultiLinear[] givePower (int which) {
    Check.assert(which < resArray.length);
    return resArray[which];
  }

  /** Tests the functionality of the class.
      Is called by the main method.
  */
  public static void testIt () {
    System.out.println("start testing Multi4Power ...");
    Poly4MultiLinear[] start = new Poly4MultiLinear[3];
    start[0] = new Poly4MultiLinear(new int[] {0,1,0,0});
    start[1] = new Poly4MultiLinear(new int[] {0,0,1,0});
    start[2] = new Poly4MultiLinear(new int[] {0,0,0,1});

    Multi4Power testObj = new Multi4Power(Poly64Max.genPoly,start,5);
    Poly4MultiLinear[] res = testObj.givePower(0);
    for (int i = 0; i < 3; i++) {
      if (!res[i].isEqual(start[i])) System.out.println("Error 1.1");
    }

    Poly4MultiLinear[] check = new Poly4MultiLinear[3];
    check[0] = new Poly4MultiLinear(new int[] {0,1,3,3});
    check[1] = new Poly4MultiLinear(new int[] {0,0,1,2});
    check[2] = new Poly4MultiLinear(new int[] {0,0,3,0});
    res = testObj.givePower(1);
    for (int i = 0; i < 3; i++) {
      if (!res[i].isEqual(check[i])) {
        System.out.println("Error 1.2 at " + i);
        System.out.print("Expect: "); check[i].writeString(); System.out.println();
        System.out.print("Got: "); res[i].writeString(); System.out.println();
        System.out.println();
      }
    }

    check[0] = new Poly4MultiLinear(new int[] {0,1,2,2});
    check[1] = new Poly4MultiLinear(new int[] {0,0,0,2});
    check[2] = new Poly4MultiLinear(new int[] {0,0,3,1});
    res = testObj.givePower(2);
    for (int i = 0; i < 3; i++) {
      if (!res[i].isEqual(check[i])) {
        System.out.println("Error 1.2 at " + i);
        System.out.print("Expect: "); check[i].writeString(); System.out.println();
        System.out.print("Got: "); res[i].writeString(); System.out.println();
        System.out.println();
      }
    }

    System.out.println("...finished testing.");
  }

  /** Calls the {@link #testIt} method
  */
  public static void main(String[] args) {
    testIt();
  }
}
