package HFE;

import java.util.*;
import HFE.*;

//-----------------------------------------------------------------------------
// Project:     Implementing HFE at UCC as Project CS4095
// Supervisors: Simon Foley (Computer Science) and Patrick Fitzpatrick (Maths)
//-----------------------------------------------------------------------------

/**
<p>
This class provides polynomials of positive degree over
the finite field F_2 which is provided in the class {@link Field2}.
It also provides addition, multiplication and modulo-operations
The maximal degree of the polynomials is fixed by the constant
{@link #maxLen}
<p>
@author Christopher Wolf (hfe@Christopher-Wolf.de)
@version 0.1
*/

//Copyright (C) 2002    Christopher Wolf
//
//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License (file LICENCE) for more details.


final class Poly2Max {

    /** Maximal length of the polynomials
      @serial constant
    */
    //public final static int maxLen = 30;  // value for testIt
    public final static int maxLen = HFE2_n.extensionDegree+Field2.elements; // value for "real" system

    /** Modulus for the mod version of the operations
      @serial constant
    */
    private static Poly2Max modulus;

    /** Current value of the polynomial.
        The polynomial has at most degree (maxLen-1). Its coefficients are
        taken from F_2. The element value[i] is the coefficient for x^i so
        value[0] is the constant term for this polynomial.
    */
    private int value[];


    /** sets all the coefficients to zero
    */
    Poly2Max() {
      setZero();
    }

    /** Creates a new polynomial. {@link #int2poly}
    */
    Poly2Max(int initValue) {
      setZero();
      int2poly(initValue);
    }

    /** Creates a copy from the given polynomial
    */
    Poly2Max(final Poly2Max father) {
      value = (int[])father.value.clone();
    }

    /** Takes one polynomial and shifts it by the given shift.
        It basicly leaves out shift number of coefficients
        and starts copying at value[shift].
        Makes a copy of this polynomial.
    */
    // !! Exception
    Poly2Max(final Poly2Max father, int shift) {
      setZero();
      Check.assert((father.degree()+shift) < maxLen,"Poly2Max: shift out of range " + shift);
      for (int i = 0; i <= father.degree(); i++) {
        value[shift+i] = father.value[i];
      }
    }


    /** {@link #array2poly}
    */
    Poly2Max(final int[] inArray) {
      array2poly(inArray);
    }

    /** Sets the private field modulus.
        Does not clone the moduls but storing it. Checks if the degree
        of the moduls is less than (maxLen/2).
    */
    static void setModulus(final Poly2Max newModulus) {
      // security check: is the degree of modulus small enough?
      Check.assert(newModulus.degree()*2 <= maxLen, "Poly2Max.setModulus: modulus too big " + newModulus.degree());
      Check.assert(newModulus.degree() >= 0, "modulus is zero");
      modulus = newModulus;
    }


    /** Returns the degree of the object.
        The degree of a constant polynomial is 0, the degree of the zero
        polynomial is -1.
    */
    int degree() {
      for (int i = maxLen-1; i >= 0; i--)
        if (value[i] != 0) return i;
      return -1;
    }

    /** Returns te coefficient for a given degree which (i.e. coeff*x^which) within the given polynomial.
    */
    int coeff(int which) {
      return value[which];
    }


    /** Compares the current polynomial with the given one.
      @return true iff (this == whom) for all coefficients
    */
    boolean isEqual(final Poly2Max whom) {
      for (int i = maxLen-1; i >= 0; i--) {
        if (value[i] != whom.value[i]) return false;
      }
      return true;
    }

    /** Checks if a polynomial is the zero polynomial.
      @return true iff all coefficients are zero
    */
    boolean isZero() {
      for (int i = 0; i < maxLen; i++)
        if (value[i] != 0) return false;
      return true;
    }


    /** Sets all the coefficients to zero by allocating new memory
    */
    // !! Version without new memory
    void setZero() {
      value = new int[maxLen];
    }

    /** Sets the coefficients according to intValue.
        The bits 0/1 are the coefficient for
        x^0, bits 2/3 are the coefficient for x^1, ...
    */
    void int2poly(int intValue) {
      Check.assert(intValue >= 0);
      Check.assert(maxLen <= 30);
      for (int i = 0; i < maxLen; i++)
        value[i] =  ((intValue >> i) & 1);
    }


    /** Sets the coefficients using the given power array.
        The elements in this array have the form {p_0, p_1, ...}
        where x^(p_i). The coefficient is assumed to be 1.
        For all elements _not_ in this array, the coefficient
        is assumed to be 0.
    */
    // !! Exception for wrong array length, power to big
    void array2poly(final int[] inArray) {
      setZero();

      for (int i = 0; i < inArray.length; i++) {
        Check.assert(inArray[i] < maxLen, "inArray is too big at position " + i + " with " + inArray[i] + " and " + maxLen);
        value[inArray[i]] = 1;
      }
    }


    /** multiplies all coefficients by the given factor in F_2.
    */
    void mulFactor(int factor) {
      for (int i = 0; i < maxLen; i++)
        value[i] = Field2.mul(factor,value[i]);
    }


    /** calculates this = wen1 + wen2
        as polynomial addition
    */
    void add(final Poly2Max a, final Poly2Max b) {
      for (int i = 0; i < maxLen; i++)
        value[i] = Field2.add(a.value[i],b.value[i]);
    }


    /** calculate this += whom
        as polynomial addition
    */
    void add(final Poly2Max whom) {
      for (int i = 0; i < maxLen; i++)
        value[i] = Field2.add(whom.value[i],value[i]);
    }

    /** calculates this = a * b
        as polynomial addition.
        Does <b>not</b> check if the result is of degree < maxLen
    */
    void mul(final Poly2Max a, final Poly2Max b) {
      // first reset this
      setZero();
      for (int i = 0; i < maxLen; i++) {
        // only do something if it is necessary
        if (a.value[i] != 0) {
          for (int j = 0; j+i < maxLen; j++) {
            if (b.value[j] != 0) {
              value[i+j] = Field2.add(value[i+j],Field2.mul(a.value[i],b.value[j]));  // add at the correct place
            }
          }
        }
      }
    }


    /** calculates this = base^by.
        Does <b>not</br> check if the result is of degree < maxLen
    */
    void pow(final Poly2Max base, int by) {
      if (by == 0) {
        setZero();
        value[0] = 1;
        return;
      }
      if ((by % 2) == 1) this.value = (int[])base.value.clone(); else {setZero(); value[0] = 1;}
      by /= 2;
      Poly2Max square = new Poly2Max(base);
      Poly2Max tmp = new Poly2Max();
      do {
        tmp.mul(square,square);
        square.value = (int[])tmp.value.clone();
        if ((by % 2) == 1) {
          tmp.mul(square,this);
          this.value = (int[])tmp.value.clone();
        }
        by /= 2;
      } while (by != 0);
    }

    /** Calculates this %= moduls.
        Uses the static field modulus.
    */
    void mod () {
      mod(modulus);
    }

    /** Calculates this = a % b
    */
    void mod (final Poly2Max a, final Poly2Max b) {
      // copy before doing anything
      value = (int[])a.value.clone();
      mod(b);
    }

    /** Calculates this %= b
    */
    void mod (final Poly2Max b) {
      // do not divide by zero
      Check.assert(!b.isZero(), "Division by Zero Error in mod!");

      // work as long as there is something to do
      while (b.degree() <= this.degree()) {
        Poly2Max additive = new Poly2Max(b,this.degree()-b.degree());
        add(additive);
      }
    }

    /** Computes this = (a+b)%modul.
        Assumes a<modul, b<modul. Does <b>not</b> check this condition.
    */
    void addMod(final Poly2Max a, final Poly2Max b) {
      add(a,b);
      mod();
    }

    /** Computes this = (this+b)%modul.
        Assumes this<modul, b<modul. Does <b>not</b> check this condition.
    */
    void addMod(final Poly2Max b) {
      add(b);
      mod();
    }

    /** Computes this = (a*b)%modul.
        Assumes a<modul, b<modul. Does <b>not</b> check this condition.
    */
    void mulMod(final Poly2Max a, final Poly2Max b) {
      mul(a,b);
      mod();
    }


    /** Writes the polynomial as a string to stdout
    */
    void writeString() {
      for (int i = maxLen-1; i > 0; i--) {
        if (value[i] != 0) {
          if (value[i] != 1) System.out.print(value[i]);
          System.out.print("x^" + i + " + ");
        }
      }
      System.out.print(value[0]);
      System.out.println(" (" + polyInt() + ")");
    }

    /** Returns the polynomial as an integer.
       Bits 0/1 are the coefficient for x^0, Bits 2/3 are the coefficients for x^1, ...
    */
    int polyInt() {
      int power = 1;
      int res = 0;
      for (int i = 0; i < maxLen; i++) {
        res += value[i]*power;
        power *= 2;
      }
      return res;
    }

    /** Tests the functionality for the whole class.
        Reports errors to stdout.
        <i>Works only for maxLen == 30</i>
    */
    static void testIt () {
      System.out.println("start testing Poly2Max ...");

      Random rnd = new Random();
      Poly2Max cur = new Poly2Max(6), cur1 = new Poly2Max(), cur2 = new Poly2Max(),
               cur3 = new Poly2Max(), cur4 = new Poly2Max();
      int curRnd, curRnd1, curRnd2, curRnd3, curRes;

      // first test the constructors and the I/O functions
      Poly2Max first = new Poly2Max();
      Poly2Max second = new Poly2Max(3);
      Poly2Max third = new Poly2Max(second);

      if (second.polyInt() != 3) System.out.println("Error 1.1");
      if (third.polyInt() != 3) System.out.println("Error 1.2a");
      if (second.coeff(0) != 1) System.out.println("Error 1.2b");
      if (second.coeff(1) != 1) System.out.println("Error 1.2c");
      //second.writeString();

      cur = new Poly2Max(130);
      if (130 != cur.polyInt()) System.out.println("Error 1.3 with " + cur.polyInt());
      cur = new Poly2Max(1024);
      if (1024 != cur.polyInt()) System.out.println("Error 1.4 with " + cur.polyInt());
      for (int i = 0; i < 100; i++) {
        curRnd = rnd.nextInt(1 << maxLen);  // gives values in 0..2^30-1
        cur = new Poly2Max(curRnd);
        cur2 = new Poly2Max(cur);
        if (curRnd != cur.polyInt()) System.out.println("Error 1.5 with " + curRnd);
        if (curRnd != cur2.polyInt()) System.out.println("Error 1.6 with " + curRnd);
      }
      cur = new Poly2Max(new int [] {0,1});
      if (3 != cur.polyInt()) System.out.println("Error 1.7 with " + cur.polyInt());
      cur.array2poly(new int [] {1, 2, 3});
      if (14 != cur.polyInt()) System.out.println("Error 1.8 with " + cur.polyInt());

      // test if the "isNull" parameter works
      cur = new Poly2Max(0);
      if (!cur.isZero()) System.out.println("Error 2.4");
      cur = new Poly2Max(23);
      cur.setZero();
      if (!cur.isZero()) System.out.println("Error 2.5");

      //Test addition
      cur1 = new Poly2Max();
      cur1.add(new Poly2Max((1 << (maxLen-1)) + 1), new Poly2Max(3));
      if (cur1.polyInt() != ((1 << (maxLen-1)) + 2)) System.out.println("Error 3.1 with " + cur1.polyInt());
      for (int i = 0; i < 100; i++) {
        curRnd1 = rnd.nextInt(1 << maxLen);
        curRnd2 = rnd.nextInt(1 << maxLen);
        curRes = curRnd1 ^ curRnd2;
        cur1 = new Poly2Max();
        cur1.add(new Poly2Max(curRnd1), new Poly2Max(curRnd2));
        cur2 = new Poly2Max(curRnd1);
        cur2.add(new Poly2Max(curRnd2));
        if (!cur1.isEqual(cur2)) System.out.println("Error 3.2 with " + curRnd1 + " " + curRnd2);
        if (cur1.polyInt() != curRes) System.out.println("Error 3.3 with " + curRnd1 + " " + curRnd2);
      }

      // Test multiplication
      cur = new Poly2Max();
      cur.mul(new Poly2Max(2), new Poly2Max(5));
      if (cur.polyInt() != 10) System.out.println("Error 4.1 with " + cur.polyInt());
      cur.mul(new Poly2Max(3), new Poly2Max(3));
      if (cur.polyInt() != 5) System.out.println("Error 4.2 with " + cur.polyInt());
      cur.mul(new Poly2Max(10), new Poly2Max(21));
      if (cur.polyInt() != 130) System.out.println("Error 4.3 with " + cur.polyInt());

      // Test both addition and multiplication
      // using the distributive law a(b+c)=ab+ac
      for (int i = 0; i < 100; i++) {
        curRnd1 = rnd.nextInt(1 << (maxLen/2));
        curRnd2 = rnd.nextInt(1 << (maxLen/2));
        curRnd3 = rnd.nextInt(1 << (maxLen/2));
        // left hand side
        cur1 = new Poly2Max(curRnd2);
        cur1.add(new Poly2Max(curRnd3));
        cur2.mul(cur1,new Poly2Max(curRnd1));
        curRes = cur2.polyInt();
        // right hand side
        cur1.mul(new Poly2Max(curRnd1), new Poly2Max(curRnd2));
        cur2.mul(new Poly2Max(curRnd1), new Poly2Max(curRnd3));
        cur3.add(cur1,cur2);

        if (cur3.polyInt() != curRes) System.out.println("Error 4.3 with " + curRnd1 + " " + curRnd2 + " " + curRnd3);
      }

      // test the shift constructor
      cur1 = new Poly2Max(new Poly2Max(3),2);
      if (cur1.polyInt() != (3 << 2)) System.out.println("Error 4.4");
      cur1 = new Poly2Max(new Poly2Max(1 << 28),0);
      if (cur1.polyInt() != (1 << 28)) System.out.println("Error 4.5");
      cur1 = new Poly2Max(new Poly2Max(3),13);
      if (cur1.polyInt() != (3 << 13)) System.out.println("Error 4.6");

      // Test multiplication together with the shift constructor
      for (int i = 0; i < 100; i++) {
        curRnd1 = rnd.nextInt(maxLen-1)+1;
        curRnd2 = rnd.nextInt(1 << ((maxLen-1)/curRnd1));
        cur1 = new Poly2Max(curRnd2);
        cur2.mul(cur1,new Poly2Max(1 << (curRnd1)));
        cur3 = new Poly2Max(cur1,curRnd1);

        if (!cur3.isEqual(cur2)) System.out.println("Error 4.7 with " + curRnd1 + " " + curRnd2);
      }

      // test mod function
      cur1 = new Poly2Max(3);
      cur2 = new Poly2Max(1);
      cur3.mod(cur1,cur2);
      if (!cur3.isZero()) System.out.println("Error 5.1 with " + cur3.polyInt());

      cur1 = new Poly2Max(2048);  // x^11
      cur2 = new Poly2Max(1025);  // x^10 -> 1
      cur3.mod(cur1,cur2);
      cur4 = new Poly2Max(2);     // x
      if (!cur3.isEqual(cur4)) {
        System.out.println("Error 5.2 with " + cur3.polyInt());
        cur3.writeString();
      }

      for (int i = 0; i < 100; i++) {
        curRnd1 = rnd.nextInt((1 << (maxLen/2))-1) +1;
        curRnd2 = rnd.nextInt((1 << (maxLen/2))-1) +1;
        cur1 = new Poly2Max(curRnd1);
        cur2 = new Poly2Max(curRnd2);
        cur4 = new Poly2Max(1);
        cur3.mul(cur1,cur2);
        cur3.mod(cur1);
        if (!cur3.isZero()) System.out.println("Error 5.3");
      }

      // test addMod functions using the add and the mod function
      for (int i = 0; i < 100; i++) {
        curRnd1 = rnd.nextInt(1 << (maxLen));
        curRnd2 = rnd.nextInt(1 << (maxLen));
        curRnd3 = rnd.nextInt((1 << maxLen/2) - 1) + 1;
        setModulus(new Poly2Max(curRnd3));
        cur1.mod(new Poly2Max(curRnd1), new Poly2Max(curRnd3));
        cur2.mod(new Poly2Max(curRnd2), new Poly2Max(curRnd3));
        cur3.add(cur1,cur2);
        cur1.addMod(new Poly2Max(curRnd1), new Poly2Max(curRnd2));
        cur2 = new Poly2Max(curRnd1);
        cur2.addMod(new Poly2Max(curRnd2));
        if (!cur1.isEqual(cur3)) System.out.println("Error 6.1 with " + curRnd1 + " " + curRnd2 + " " + curRnd3);
        if (!cur1.isEqual(cur2)) System.out.println("Error 6.2 with " + curRnd1 + " " + curRnd2 + " " + curRnd3);
      }

      // test mulMod functions using the mul and the mod function
      for (int i = 0; i < 100; i++) {
        curRnd1 = rnd.nextInt(1 << maxLen);
        curRnd2 = rnd.nextInt(1 << maxLen);
        curRnd3 = rnd.nextInt((1 << (maxLen/2)) - 1) + 1;
        setModulus(new Poly2Max(curRnd3));
        cur1.mod(new Poly2Max(curRnd1), new Poly2Max(curRnd3));
        cur2.mod(new Poly2Max(curRnd2), new Poly2Max(curRnd3));
        cur3.mul(cur1,cur2);
        cur3.mod(new Poly2Max(curRnd3));
        cur.mulMod(cur1, cur2);
        if (!cur.isEqual(cur3)) System.out.println("Error 6.3 at " + i + " with " + curRnd1 + " " + curRnd2 + " " + curRnd3);
      }


      System.out.println("...finished testing.");
    }

    /** Mainly invoces the {@link #testIt} method.
    */
    public static void main(String[] args) {
        testIt();
    }
}
