package HFE;

import java.util.*;
import java.lang.Math;
import HFE.*;

//-----------------------------------------------------------------------------
// Project:     Implementing HFE at UCC as Project CS4095
// Supervisors: Simon Foley (Computer Science) and Patrick Fitzpatrick (Maths)
//-----------------------------------------------------------------------------

/**
<p>
This class provides polynomials of degree two in n independent variables.
The polynomials are over the finite field GF(2) which is
provided in the class {@link Field2}.
<p>
@author Christopher Wolf (hfe@Christopher-Wolf.de)
@version 0.1
*/

//Copyright (C) 2002    Christopher Wolf
//
//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License (file LICENCE) for more details.<p>
 

final class Poly2MultiSquare {

    /** The number of variables in this polynomial.
      @serial should be greater than 0
    */
    private int varNum;

    /** Coefficients for the variables.
        The variables are x_1..x_n where
        n is {@link #varNum}. Position 0
        is for the constant term. Position
        n+1 is for x_1*x_1, n+2 is for
        x_1*x_2, ..., n+(n-1) is for x_1*x_n,
        n+n is for x_2*x_2, n+n+1 is for
        x_2*x_3, ...
    */
    private int coeff[];

    /** Calulates the number of terms
        for a given number of variables n by using the formula ((n+1)*(n+2))/2.
    */
    public static int tau(int n) {
      Check.assert(n >= 0);
      return ((n+1)*(n+2))/2;
    }

    /** Calulates the number of variables
        for a given number of terms t.
    */
    public static int tau_inverse(int t) {
      Check.assert(t > 0);
      int res = Math.round((-3 + (float)Math.sqrt((double)9 - 4*(2-2*t))) / 2);
      Check.assert(t == tau(res));
      return res;
    }


    /** sets all the coefficients to zero
        assumes that the polynomial has no variables
    */
    public Poly2MultiSquare() {
      varNum = 0;
      initMem();
    }

    /** Sets all the coefficients to zero by allocating new memory.
        Assumes that the polynomial has varNum variables.
       @param varNum number of variables for that polynomial
    */
    public Poly2MultiSquare(int varNum) {
      Check.assert(varNum >= 0);
      this.varNum = varNum;
      initMem();
    }

    /** Sets all the coefficients to a random value and
        also allocates new memory for these coefficients.
        Assumes that the polynomial has varNum variables.
       @param varNum number of variables for that polynomial
    */
    public Poly2MultiSquare(int varNum, Random rnd) {
      Check.assert(varNum >= 0);
      this.varNum = varNum;
      initMem();
      for (int i = 0; i < coeff.length; i++)
        coeff[i] = Field2.rnd(rnd);
    }

    /** Creates a copy from the given polynomial.
        Uses memory copy for the coefficients.
    */
    public Poly2MultiSquare(final Poly2MultiSquare father) {
      varNum = father.varNum;
      coeff = (int[])father.coeff.clone();
    }


    /** Creates a copy from the given linear polynomial.
        Uses memory copy for the coefficients.
    */
    public Poly2MultiSquare(final Poly2MultiLinear father) {
      varNum = father.giveVarNum();
      initMem();
      int [] linearCoeffs = father.coeffs();
      for (int i = 0; i < linearCoeffs.length; i++)
        coeff[i] = linearCoeffs[i];
    }


    /** {@link #array2poly}
    */
    public Poly2MultiSquare(final int[] inArray) {
      Check.assert(inArray.length > 0, "inArray.length == 0");
      Check.assert(Field2.isElement(inArray), "inArray not in Field2");
      varNum = tau_inverse(inArray.length);
      array2poly(inArray);
    }

    /** Creates a new object by using inArray for the
        coefficients. Does no checking for arguments,
        no memory copy.
       @param varNum number of variables for this object
       @param inArray coefficients for the multivaribale polynomial
    */
    public Poly2MultiSquare(int varNum, final int[] inArray) {
      Check.assert((varNum >= 0) && (tau(varNum) == inArray.length));
      Check.assert(Field2.isElement(inArray));
      this.varNum = varNum;
      coeff = inArray;
    }

    /** Multiplies two lineare polynomials
        and stores the result in a new object. The two lineare polynomials must
        have the same number of variables. The result is a polynomial of degree 2.
        This constructor also allocates new memory and is therefore
        quite slow. The faster version is {@link #mul}
    */
    public Poly2MultiSquare(Poly2MultiLinear a, Poly2MultiLinear b) {
      Check.assert(a.giveVarNum() == b.giveVarNum());
      varNum = a.giveVarNum();
      initMem();
      mul(a,b);
    }

    /** returns the coefficient array
        without copying it.
    */
    public int[] coeffs() {
      return coeff;
    }

    /** Compares this and whom coefficientwise.
        If the two polynomials have different number of variables they
        can not be equal.
      @return true iff all the coefficients are the same
              both objects
    */
    public boolean isEqual(Poly2MultiSquare whom) {
      if (whom.varNum != varNum) return false;
      for (int i = 0; i < coeff.length; i++) {
        if (whom.coeff[i] != coeff[i]) return false;
      }
      return true;
    }

    /** Checks if the polynomial is the zero polynomial.
      @return true iff all coefficients are zero
    */
    public boolean isZero() {
      for (int i = 0; i < coeff.length; i++)
        if (coeff[i] != 0) return false;
      return true;
    }


    /** Sets all the coefficients to zero
        without allocating new memory.
    */
    public void setZero() {
      for (int i = 0; i < coeff.length; i++)
        coeff[i] = 0;
    }


    /** Allocates new memory for the coefficients.
        The coefficients are initalized with zero.
    */
    public void initMem() {
      coeff = new int[tau(varNum)];
    }


    /** Sets all the coefficients
        by the given array. Does neither memory
        copy nor argument checking.
    */
    public void setCoeffs(int [] inArray) {
      Check.assert(tau(varNum) == inArray.length);
      Check.assert(Field2.isElement(inArray));
      coeff = inArray;
    }


    /** Sets the coefficients using the given coefficient.
        Copies the array and checks if the array
        has the right length.
       @throws IllegalArgumentException: array has the wrong length, needs to be (varNum+1)
    */
    public void array2poly(final int[] inArray) {
      Check.assert(Field2.isElement(inArray));
      if (inArray.length != tau(varNum)) new IllegalArgumentException("Need " + varNum + " coefficients");
      coeff = (int[])inArray.clone();
    }

    /** Sets all the coefficients to a random value
        without allocating new memory for these coefficients.
        @param rnd random generator
    */
    public void rnd(Random rnd) {
      for (int i = 0; i < coeff.length; i++)
        coeff[i] = Field2.rnd(rnd);
    }

    /** calculates this = a + b
        as polynomial addition
    */
    public void add(final Poly2MultiSquare a, final Poly2MultiSquare b) {
      Check.assert((a.varNum == varNum) && (b.varNum == varNum));
      for (int i = 0; i < coeff.length; i++)
        coeff[i] = Field2.add(a.coeff[i],b.coeff[i]);
    }


    /** calculate this += whom
        as polynomial addition
    */
    public void add(final Poly2MultiSquare whom) {
      Check.assert(whom.varNum == varNum);
      for (int i = 0; i < coeff.length; i++)
        coeff[i] = Field2.add(whom.coeff[i],coeff[i]);
    }

    /** calculate this[i] += whom
        in terms of coefficient addition
    */
    public void add(final int whom, final int i) {
      Check.assert((i < varNum) && (i >= 0), "i is not valid: " + i);
      Check.assert(Field2.isElement(whom));
      coeff[i] = Field2.add(whom,coeff[i]);
    }

    /** calculate this += whom
        as polynomial addition
    */
    public void add(final Poly2MultiLinear whom) {
      Check.assert(whom.giveVarNum() == varNum);
      int[] coeffsLinear = whom.coeffs();
      for (int i = 0; i < coeffsLinear.length; i++)
        coeff[i] = Field2.add(coeffsLinear[i],coeff[i]);
    }

    /** calculates this = a - b
        as polynomial subtraction
    */
    public void sub(final Poly2MultiSquare a, final Poly2MultiSquare b) {
      Check.assert((a.varNum == varNum) && (b.varNum == varNum));
      for (int i = 0; i < coeff.length; i++)
        coeff[i] = Field2.sub(a.coeff[i],b.coeff[i]);
    }


    /** calculate this -= whom
        as polynomial subtraction
    */
    public void sub(final Poly2MultiSquare whom) {
      Check.assert(whom.varNum == varNum);
      for (int i = 0; i < coeff.length; i++)
        coeff[i] = Field2.sub(coeff[i],whom.coeff[i]);
    }

    /** Calculates the square values
        using the linear values. I.e. uses a vector
        with the values for x_1, ... x_n (position 0 is not used)
        and calulates a vector with values for x_1, ... x_n, x_1^2, x_1*x_2, ...
        The position 0 in the output vector is not used.
    */
    public static void getValues(int[] rawValues, int[] res) {
      int myTau = res.length;
      int varNum = rawValues.length-1;
      Check.assert(Field2.isElement(rawValues));
      Check.assert(tau(varNum) == myTau, "not the same number of variables: " + tau(varNum) + " ("+  varNum + ") " + myTau);

      // first copy the linear values
      System.arraycopy(rawValues,1,res,1,rawValues.length-1);

      // now compute the square values
      int k = 1;
      int j = 1;
      for (int i = rawValues.length; i < myTau; i++) {
        res[i] = Field2.mul(rawValues[k],rawValues[j]);
        j++;
        if (j > varNum) {k++; j = k;}
      }
    }

    /** Evaluate the polynomial using the given value vector.
        The entry value[0] is _not_ uses, value[1] is the first valid entry.
        This format is produces by {@link #getValues}.
      @param values array with tau(varNum) entries. {@link #tau}
      @return value of the current polynimial.
    */
    public int apply(int[] values) {
      Check.assert(values.length == coeff.length, "Length are different: " + values.length + " and " + coeff.length);
      Check.assert(Field2.isElement(coeff) && Field2.isElement(values));
      int res = coeff[0];
      for (int i = 1; i < values.length; i++)
        res = Field2.add(res,Field2.mul(coeff[i],values[i]));

      return res;
    }

    /** Evaluate the polynomial using the given value vector
        and the given coefficient vector. The entry value[0] is _not_ uses,
        value[1] is the first valid entry.
      @param coeffs coefficients of a multivariable, linear polynomial
      @param values values for the variables (and there squares) of the polynomial, value[1] is the value for the first variable, ...
      @return value of the current polynimial.
    */
    public static int apply(int [] coeffs, int[] values) {
      Check.assert(values.length == coeffs.length);
      Check.assert(Field2.isElement(coeffs) && Field2.isElement(values));
      int res = coeffs[0];
      for (int i = 1; i < values.length; i++)
        res = Field2.add(res,Field2.mul(coeffs[i],values[i]));

      return res;
    }

    /** Calculates this = a*b
        in terms of polynomial multiplication. a and b must have the same number
        of varialbes. The current value of
        this is discared.
    */
    public void mul(Poly2MultiLinear a, Poly2MultiLinear b) {
      Check.assert(a.giveVarNum() == b.giveVarNum());
      Check.assert(a.giveVarNum() == varNum);

      int [] A, B;
      A = a.coeffs();
      B = b.coeffs();

      // assume that we are working on a (varNum+1)*(varNum+1) matrix
      // and reduce that to a triangle matrix for the terms which are
      // not on the diagonal
      int k = 0;    // k <= j all the time
      int j = 0;
      for (int i = 0; i < coeff.length; i++) {
        // do we have a quadratic term?
        if (k == j)
          coeff[i] = Field2.mul(A[k],B[j]);
        else
          coeff[i] = Field2.add(Field2.mul(A[k],B[j]),Field2.mul(A[j],B[k]));
        j++;
        if (j > varNum) {
          k++;
          j = k;
        }
      }
    }

    /** this += factor*whom.
        Multiplies whom with the scalar factor and adds the result to this.
    */
    public void mulAdd(int factor, Poly2MultiSquare whom) {
      Check.assert(Field2.isElement(factor));
      Check.assert(whom.varNum == varNum);

      for (int i = 0; i < coeff.length; i++)
        coeff[i] = Field2.add(coeff[i],Field2.mul(factor,whom.coeff[i]));
    }

    /** Calculates this += a*b
        in terms of polynomial multiplication.
    */
    public void mulAdd(Poly2MultiLinear a, Poly2MultiLinear b) {
      Check.assert(a.giveVarNum() == b.giveVarNum());
      Check.assert(a.giveVarNum() == varNum);

      int [] A, B;
      A = a.coeffs();
      B = b.coeffs();

      // multiply A,B using a kind of matrices
      int k = 0;    // k <= j all the time
      int j = 0;
      for (int i = 0; i < coeff.length; i++) {
        // do we have a quadratic term?
        if (k == j)
          coeff[i] = Field2.add(coeff[i],Field2.mul(A[k],B[j]));
        else
          coeff[i] = Field2.add(coeff[i],Field2.add(Field2.mul(A[k],B[j]),Field2.mul(A[j],B[k])));
        j++;
        if (j > varNum) {
          k++;
          j = k;
        }
      }
    }


    /** Writes the polynomial as a string to stdout
    */
    public void writeString() {
      // first write the single terms
      for (int i = 1; i <= varNum; i++) {
        if (coeff[i] != 0) {
          if (coeff[i] != 1) System.out.print(coeff[i]);
          System.out.print("x_" + i + " + ");
        }
      }
      // now print the mixed terms
      int k = 1;    // k < j all the time
      int j = 1;
      for (int i = varNum+1; i < coeff.length; i++) {
        if (coeff[i] != 0) {
          if (coeff[i] != 1) System.out.print(coeff[i]);
          if (k == j) System.out.print("x_" + k + "^2" + " + ");
            else System.out.print("x_" + k + "," + j + " + ");
        }
        j++;
        if (j > varNum) {
          k++;
          j = k;
        }
      }

      // finaly write the constant
      System.out.print(coeff[0]);
    }


    /** Tests the functionality for the whole class.
        Reports errors to stdout.
    */
    public static void testIt () {
      System.out.println("start testing Poly2MultiSquare ...");

      final int varsAll = 257;
      final int repeats = 10;

      Random rnd = new Random();
      Poly2MultiSquare cur = new Poly2MultiSquare(), cur1 = new Poly2MultiSquare(),
                       cur2 = new Poly2MultiSquare(), cur3 = new Poly2MultiSquare();
      int curRnd, curRnd1, curRnd2, curRnd3, curRes;

      // test the tau-function
      if (tau(0) != 1) System.out.println("Error 1.1a");
      if (tau(1) != 3) System.out.println("Error 1.1b");
      if (tau(2) != 6) System.out.println("Error 1.1c");
      if (tau(3) != 10) System.out.println("Error 1.1d");
      if (tau(4) != 15) System.out.println("Error 1.1e");
      if (tau(5) != 21) System.out.println("Error 1.1f");

      // test its inverse
      if (tau_inverse(1) != 0) System.out.println("Error 1.2a");
      if (tau_inverse(3) != 1) System.out.println("Error 1.2b");
      if (tau_inverse(6) != 2) System.out.println("Error 1.2c");
      if (tau_inverse(10) != 3) System.out.println("Error 1.2d");
      if (tau_inverse(15) != 4) System.out.println("Error 1.2e");
      if (tau_inverse(21) != 5) System.out.println("Error 1.2f");

      // do tests for bigger values of n, t
      for (int i = 0; i < 1000; i++)
        if (i != tau_inverse(tau(i))) System.out.println("Error 1.3 with " + i);

      cur1 = new Poly2MultiSquare();
      cur2 = new Poly2MultiSquare(3);
      if (cur1.isEqual(cur2)) System.out.println("Error 1.4");
      if (!cur1.isZero() || !cur2.isZero()) System.out.println("Error 1.5");
      cur3 = new Poly2MultiSquare(cur2);
      if (!cur2.isEqual(cur3) || !cur3.isZero())System.out.println("Error 1.6");
      cur1 = new Poly2MultiSquare(new int[] {1, 0,1, 1,0,1}); // 1 + x_2 + x_1^2 + x_2^2
      cur2 = new Poly2MultiSquare(2,cur1.coeffs());
      if (!cur1.isEqual(cur2)) System.out.println("Error 1.7");
      int [] c = cur2.coeffs();
      if ((c[0] != 1) || (c[1] != 0) || (c[2] != 1) ||
          (c[3] != 1) || (c[4] != 0) || (c[5] != 1)) System.out.println("Error 1.8");
      cur2.setCoeffs(new int[] {0, 0,0, 0,0,0});
      if (!cur2.isZero()) System.out.println("Error 1.9");
      cur2.setCoeffs(cur1.coeffs());
      if (!cur1.isEqual(cur2)) System.out.println("Error 1.10");
      cur2.setZero();
      if (!cur2.isZero()) System.out.println("Error 1.11");

      cur1 = new Poly2MultiSquare(new int[] {0, 1,1, 1,1,1});
      cur2 = new Poly2MultiSquare(new int[] {1, 1,1, 1,0,0});
      cur3 = new Poly2MultiSquare(2);
      cur3.add(cur1,cur2);
      if (!cur3.isEqual(new Poly2MultiSquare(new int[] {1, 0,0, 0,1,1}))) System.out.println("Error 2.1");
      cur3.sub(cur1,cur2);
      if (!cur3.isEqual(new Poly2MultiSquare(new int[] {1, 0,0, 0,1,1}))) System.out.println("Error 2.2");
      cur3 = new Poly2MultiSquare(cur2);
      cur3.add(cur1);
      if (!cur3.isEqual(new Poly2MultiSquare(new int[] {1, 0,0, 0,1,1}))) System.out.println("Error 2.3");
      cur3 = new Poly2MultiSquare(cur2);
      cur3.sub(cur1);
      if (!cur3.isEqual(new Poly2MultiSquare(new int[] {1, 0,0, 0,1,1}))) System.out.println("Error 2.4");

      for (int i = 0; i < repeats; i++) {
        final int vars = varsAll;
        int rndPos = rnd.nextInt(tau(vars));
        int res = 0;
        int[] a = new int[tau(vars)],b = new int[tau(vars)];
        for (int j = 0; j < a.length; j++) {
          int ra = Field2.rnd(rnd);
          int rb = Field2.rnd(rnd);
          a[j] = ra;
          b[j] = rb;
          if (j == rndPos) res = Field2.add(ra,rb);
        }
        cur1 = new Poly2MultiSquare(vars,a);
        cur2 = new Poly2MultiSquare(vars,b);
        cur3 = new Poly2MultiSquare(vars);
        cur3.add(cur1,cur2);
        if (cur3.coeffs()[rndPos] != res) System.out.println("Error 2.5 with " + res);
        cur3.sub(cur1);
        if (!cur3.isEqual(cur2)) System.out.println("Error 2.6");

        cur3.sub(cur1,cur2);
        if (cur3.coeffs()[rndPos] != res) System.out.println("Error 2.7");
        cur3.add(cur1);
        if (!cur3.isEqual(cur2)) System.out.println("Error 2.8");
      }

      // test the getValue function
      int[] t = new int[tau(3)];
      getValues(new int[] {0, 1,1,0}, t);  // computes the value vector for 3 variables
      if ((t[0] != 0) || (t[1] != 1) || (t[2] != 1) || (t[3] != 0) || (t[4] != 1) ||
          (t[5] != 1) || (t[6] != 0) || (t[7] != 1) || (t[8] != 0) || (t[9] != 0)) System.out.println("Error 3.1");

      cur1 = new Poly2MultiSquare(new int[] {1, 1,0, 1,1,1});
      t = new int[tau(2)];
      getValues(new int[] {0, 1,0},t);
      if (cur1.apply(t) != 1) System.out.println("Error 3.2");

      // if two polynomials are added componentwise, the result should be p(t)+q(t),
      // if the same values are applied to both polynomials p,q
      for (int i = 0; i < repeats; i++) {
        final int vars = varsAll;
        int[] a = new int[tau(vars)],b = new int[tau(vars)];
        int[] testValues = new int[vars+1];
        for (int j = 0; j < a.length; j++) {
          a[j] = Field2.rnd(rnd);
          b[j] = Field2.rnd(rnd);
        }
        for (int j = 0; j < testValues.length; j++)
          testValues[j] = Field2.rnd(rnd);

        cur1 = new Poly2MultiSquare(vars,a);
        cur2 = new Poly2MultiSquare(vars,b);
        cur3 = new Poly2MultiSquare(vars);
        getValues(testValues,a);
        int res = Field2.add(cur1.apply(a), cur2.apply(a));
        cur3.add(cur1,cur2);
        if (cur3.apply(a) != res) System.out.println("Error 4.1");
      }


      // random test for mul
      for (int i = 0; i < repeats; i++) {
        final int vars = varsAll;
        int res;
        int[] a = new int[vars+1],b = new int[vars+1], v = new int[vars+1];
        for (int j = 0; j < a.length; j++) {
          a[j] = Field2.rnd(rnd);
          b[j] = Field2.rnd(rnd);
          v[j] = Field2.rnd(rnd);
        }
        Poly2MultiLinear l1 = new Poly2MultiLinear(a);
        Poly2MultiLinear l2 = new Poly2MultiLinear(b);
        res = Field2.mul(l1.apply(v),l2.apply(v));
        cur1 = new Poly2MultiSquare(l1,l2);
        int[] allValues = new int[tau(vars)];
        getValues(v,allValues);
        if (res != cur1.apply(allValues)) System.out.println("Error 5.1");
      }

      // test if x_n is set correctly
      int res;
      int[] a = new int[varsAll+1],b = new int[varsAll+1], v = new int[varsAll+1];
      for (int j = 0; j < a.length; j++) {
        if (j < varsAll) a[j] = 0; else a[j] = 1;
        if (j < varsAll) b[j] = 0; else b[j] = 1;
        if (j < varsAll) v[j] = 0; else v[j] = 1;
      }
      Poly2MultiLinear l1 = new Poly2MultiLinear(a);
      Poly2MultiLinear l2 = new Poly2MultiLinear(b);
      res = Field2.mul(l1.apply(v),l2.apply(v));
      if (res != 1) System.out.println("Error 5.2");
      cur1 = new Poly2MultiSquare(l1,l2);
      int[] allValues = new int[tau(varsAll)];
      getValues(v,allValues);
      if (1 != cur1.apply(allValues)) System.out.println("Error 5.3");

      System.out.println("...finished testing.");
    }

    /** Mainly invoces the {@link #testIt()} method.
    */
    public static void main(String[] args) {
        testIt();
    }
}
