package HFE;

import HFE.*;
import java.util.*;

//-----------------------------------------------------------------------------
// Project:     Implementing HFE at UCC as Project CS4095
// Supervisors: Simon Foley (Computer Science) and Patrick Fitzpatrick (Maths)
//-----------------------------------------------------------------------------

/**
<p>
This class encapsulates the operations for polynomials
over the finite field F_64 which is an extension over
F_4. The generation polynomial is f(x) = x^3 + x^2 + 2x + 3
<br>
The polynomials are bounded by {@link #maxLen}.
<p>
@author Christopher Wolf (hfe@Christopher-Wolf.de)
@version 0.1
*/

//Copyright (C) 2002    Christopher Wolf
//
//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License (file LICENCE) for more details.


class Poly2_nMax {

    /** Maximal length of the polynomials
      @serial constant
    */
    final static int maxLen = 34;

    /** Modulus for the mod version of the operations
      @serial constant
    */
    static Poly2_nMax modulus;

    /** Current value of the polynomial.
        The polynomial has at most degree (maxLen-1). Its coefficients are
        taken from F_64. The element value[i] is the coefficient for x^i so
        value[0] is the constant term for this polynomial.
    */
    Field2_n value[];

    /** sets all the coefficients to zero and allocates new memory
    */
    Poly2_nMax() {
      allocMem();
    }

    /** Creates a copy from the given polynomial
      @param father polynomial to be copied
    */
    Poly2_nMax(final Poly2_nMax father) {
      allocMem();
      for (int i = 0; i < maxLen; i++)
        value[i].copy(father.value[i]);
    }

    /** Takes one polynomial and shifts it by the given shift.
        It basicly leaves out shift number of coefficients
        and starts copying at value[shift].
        Another point of view is that the polynomial
        constructed is father * x^(shift)
        Makes a copy of this polynomial.
       @param father polynomial to be copied
       @param shift  the number of elements to be shifted.
    */
    Poly2_nMax(final Poly2_nMax father, int shift) {
      allocMem();
      Check.assert((father.degree()+shift) < maxLen);
      Check.assert(shift >= 0);
      for (int i = 0; i <= father.degree(); i++) {
        value[shift+i].copy(father.value[i]);
      }
    }

    /** Generates a polynomial with random coefficients.
        The degree of this polynomial can be expected
        to be (maxLen-1).
       @param rnd Random source
    */
    Poly2_nMax(Random rnd) {
      allocMem();
      for (int i = 0; i < maxLen; i++) {
        value[i].rnd(rnd);
      }
    }

    /** Generates a polynomial with random coefficient.
        The degree of this polynomial is degree.
        Degree has to be non-negative (i.e. the zero
        polynomial can not be obtained using this function).
       @param rnd Random source
       @param degree degree of the resulting polynomial
    */
    Poly2_nMax(Random rnd, int degree) {
      Check.assert((degree >= 0) && (degree < maxLen), "Degree out of range " + degree);
      allocMem();
      for (int i = 0; i < degree-1; i++) {
        value[i].rnd(rnd);
      }
      // ensure that we really have the degree we wanted
      do {
        value[degree].rnd(rnd);
      } while (value[degree].isZero());
    }


    /** Sets the private field modulus.
        Does not clone the moduls but storing it. Checks if the degree
        of the moduls is less than (maxLen/2).
    */
    static void setModulus(final Poly2_nMax newModulus) {
      // security check: is the degree of modulus small enough?
      Check.assert((newModulus.degree()*2) <= maxLen, "degree of Modulus is too big");
      modulus = newModulus;
    }


    /** Returns the degree of the object.
        The degree of a constant polynomial is 0, the degree of the zero
        polynomial is -1.
    */
    int degree() {
      for (int i = maxLen-1; i >= 0; i--)
        if (!value[i].isZero()) return i;
      return -1;
    }

    /** Compares the current polynomial with the given one.
      @return true iff (this == whom) for all coefficients
    */
    boolean isEqual(final Poly2_nMax whom) {
      for (int i = maxLen-1; i >= 0; i--) {
        if (!value[i].isEqual(whom.value[i])) return false;
      }
      return true;
    }

    /** Checks if a polynomial is the zero polynomial.
      @return true iff all coefficients are zero
    */
    boolean isZero() {
      for (int i = 0; i < maxLen; i++)
        if (!value[i].isZero()) return false;
      return true;
    }

    /** Sets all the coefficients to zero without allocating new memory
    */
    void setZero() {
      for (int i = 0; i < maxLen; i++)
        value[i].setZero();
    }

    /** get coefficient at position pos
       by constructing a new object from type Field2_n
    */
    Field2_n getCoeff(int pos) {
      Check.assert((pos >=0) && (pos < maxLen), "Pos out of range " + pos);
      return new Field2_n(value[pos]);
    }

    /** get coefficient at position pos
       by copying it to res
    */
    void getCoeff(Field2_n res, int pos) {
      Check.assert((pos >=0) && (pos < maxLen), "Pos out of range " + pos);
      res.copy(value[pos]);
    }

    /** set coefficient at position pos _with_ copying it
    */
    void setCoeff(Field2_n whom, int pos) {
      Check.assert((pos >=0) && (pos < maxLen), "Pos out of range " + pos);
      value[pos].copy(whom);
    }

    /** set coefficient at position pos with a random value
    */
    void setCoeff(Random rnd, int pos) {
      Check.assert((pos >=0) && (pos < maxLen), "Pos out of range " + pos);
      value[pos].rnd(rnd);
    }

    /** Allocates new memory. As a side effect, sets everything to zero
    */
    void allocMem() {
      value = new Field2_n[maxLen];
      for (int i = 0; i < maxLen; i++)
        value[i] = new Field2_n();
    }

    /** Takes one polynomial and shifts it by the given shift.
        It basicly leaves out shift number of coefficients
        and starts copying at value[shift].
        Makes a copy of this polynomial.
       @param father Polynomial to be copied
       @param shift number of elements to be shifted.
    */
    void copyShift(final Poly2_nMax father, int shift) {
      Check.assert((father.degree()+shift) < maxLen);
      Check.assert(shift >= 0, "shift is too small, namely " + shift);
      setZero();
      // now copy everything into the value array
      for (int i = 0; i <= father.degree(); i++)
        value[shift+i].copy(father.value[i]);
    }

    /** Shifts the current polynomial by the given number of places.
      @param shift number of places to be shifted
    */
    void shiftL(int shift) {
      Check.assert(this.degree()+shift < maxLen, "to big shift " + shift);
      // copy the elements
      for (int i = this.degree(); i >= 0; i--)
        value[shift+i].copy(value[i]);
      // now reset the lower part of the polynomial
      for (int i = 0; i < shift; i++)
        value[i].setZero();
    }

    /** Takes one polynomial and copies it.
    */
    void copy(final Poly2_nMax father) {
      // now copy everything into the value array
      for (int i = 0; i < maxLen; i++)
        value[i].copy(father.value[i]);
    }

    /** multiplies all coefficients by the given factor in GF(2^n)
        uses static memory to store the result of each step
    */
    void mulFactor(Field2_n factor) {
      for (int i = 0; i < maxLen; i++)
        value[i].mul(factor);
    }

    /** calculates this = wen1 + wen2
        as polynomial addition
    */
    void add(final Poly2_nMax a, final Poly2_nMax b) {
      for (int i = 0; i < maxLen; i++)
        value[i].add(a.value[i],b.value[i]);
    }

    /** calculate this += whom
        as polynomial addition
    */
    void add(final Poly2_nMax whom) {
      for (int i = 0; i < maxLen; i++)
        value[i].add(whom.value[i]);
    }

    /** calculate this += 1
        as polynomial addition
    */
    void addOne() {
      value[0].addOne();
    }

    /** Static register for multiplication
    */

    static Field2_n mul_tmp = new Field2_n();

    /** calculates this = a * b
        as polynomial addition.
        Assumes that a.degree + b.degree < maxLen
    */
    void mul(final Poly2_nMax a, final Poly2_nMax b) {
      Check.assert((a.degree() + b.degree()) < maxLen, "a " + a.degree() + " and b " +  b.degree() + " are too big");
      // first reset this
      setZero();
      for (int i = 0; i < maxLen; i++) {
        // only do something if it is necessary
        if (!a.value[i].isZero()) {
          for (int j = 0; j+i < maxLen; j++) {
            if (!b.value[j].isZero()) {
              Check.assert(i+j < maxLen, "a and b are too big");
              mul_tmp.mul(a.value[i],b.value[j]);
              value[i+j].add(mul_tmp);  // add at the correct place
            }
          }
        }
      }
    }

    /** Calculates this %= moduls.
        Uses the static field modulus.
    */
    void mod () {
      mod(modulus);
    }


    /** Calculates this = a % b
    */
    void mod (final Poly2_nMax a, final Poly2_nMax b) {
      // copy before doing anything
      copy(a);
      mod(b);
    }

    /** static variable for our mod method
    */
    static Poly2_nMax modS_additive = new Poly2_nMax();
    /** second static variable for our mod method
    */
    static Field2_n modS_tmp = new Field2_n();

    /** Calculates this %= b
        slow but working implementation of mod
    */
    void modS (final Poly2_nMax b) {
      // do not divide by zero
      Check.assert(!b.isZero());

      // work so long as there is something to do
      while (b.degree() <= this.degree()) {
        modS_additive.copyShift(b,this.degree()-b.degree());
        modS_tmp.div(value[degree()],b.value[b.degree()]);
        mulAdd(modS_tmp,modS_additive);
      }
    }

    /** static variable for mod method
    */
    static Field2_n mod_tmp = new Field2_n();

    /** second static variable for our mod method
    */
    static Field2_n mod_divRes = new Field2_n();


    /** Calculates this %= b
    */
    void mod (final Poly2_nMax b) {
      // do not divide by zero
      Check.assert(!b.isZero());

      int degB = b.degree();
      int degT = this.degree();
      // work so long as there is something to do
      while (degB <= degT) {
        int degDiff = degT - degB;
        mod_divRes.div(value[degT],b.value[degB]);
        value[degT].setZero(); // is zero after this step
        for (int i = 0; i < degB; i++) {
          mod_tmp.mul(mod_divRes,b.value[i]);
          value[i+degDiff].add(mod_tmp);
        }
        // degree of this is smaller now by one at least
        degT--;
        if (degT != -1) {
          while (value[degT].isZero()) {  // maybe it is much smaller now!
            degT--;
            if (degT== -1) break;
          }
        }
        Check.assert(degB == b.degree(), "degree of b is wrong " + degB + " / " + b.degree());
        Check.assert(degT == this.degree(), "degree of this is wrong " + degT + " / " + this.degree());
      }
    }

    /** static variable for method div
    */
    static Poly2_nMax div_tmp = new Poly2_nMax();

    /** Calculates this = a / b
        and assumes that there is _no_ remainder!
        Slow implementation, use {@link #div} instead!
    */
    void divS(final Poly2_nMax a, final Poly2_nMax b) {
      // do not divide by zero
      Check.assert(!b.isZero());
      // init all registers
      this.setZero();
      div_tmp.copy(a);

      // work so long as there is something to do
      while (b.degree() <= div_tmp.degree()) {
        int diff = div_tmp.degree()-b.degree();
        Check.assert(diff >= 0, "Division: Difference is negative: " + diff);
        modS_additive.copyShift(b,diff);
        value[diff].div(div_tmp.value[div_tmp.degree()],b.value[b.degree()]);  // save the result in the right position
        div_tmp.mulAdd(value[diff],modS_additive);
      }
      Check.assert(div_tmp.degree() == -1, "Division was performed on not suitable polynomial.");
    }

    /** Calculates this = a / b
        and assumes that there is _no_ remainder, i.e. a is a multiple of b.
    */
    void div(final Poly2_nMax a, final Poly2_nMax b) {
      // do not divide by zero
      Check.assert(!b.isZero());

      // init all registers
      this.setZero();
      div_tmp.copy(a);

      int degB = b.degree();
      int degT = div_tmp.degree();
      // work so long as there is something to do
      while (degB <= degT) {
        int degDiff = degT - degB;
        Check.assert(degDiff >= 0, "Division: Difference is negative: " + degDiff);
        value[degDiff].div(div_tmp.value[degT],b.value[degB]);
        div_tmp.value[degT].setZero(); // is zero after this step
        for (int i = 0; i < degB; i++) {
          mod_tmp.mul(value[degDiff],b.value[i]);
          div_tmp.value[i+degDiff].add(mod_tmp);  // addition and substraction are the same in GF(2^n)
        }
        // degree of div_tmp is smaller now by one at least
        degT--;
        if (degT != -1) {
          while (div_tmp.value[degT].isZero()) {  // maybe it is much smaller now -> check!
            degT--;
            if (degT== -1) break;  // not too much, otherwise we are out of range!
          }
        }
        Check.assert(degB == b.degree(), "degree of b is wrong " + degB + " / " + b.degree());
        Check.assert(degT == div_tmp.degree(), "degree of this is wrong " + degT + " / " + this.degree());
      }
      Check.assert(div_tmp.degree() == -1, "Division was performed on not suitable polynomial.");
    }

    /** static variable for mulAdd
    */
    Field2_n mulAdd_coeff = new Field2_n();

    /** compute this = coeff*poly
        where coeff is from GF(2^n) and poly is an object of type Poly2_nMax
       @param coeff coefficient from GF(2^n) which is multiplied
       @param poly polynomial over GF(2^n) in x
    */
    void mulAdd (Field2_n coeff, Poly2_nMax poly) {
      for (int i = 0; i < maxLen; i++) {
        mulAdd_coeff.mul(poly.value[i],coeff);
        value[i].add(mulAdd_coeff);
      }
    }

    /** Initializes the static variables
    */
    static void initAll() {
      for (int i = 0; i < squareMul_tmp.length; i++)
        squareMul_tmp[i] = new Field2_n();
      for (int i = 0; i < gcd_tmp.length; i++)
        gcd_tmp[i] = new Poly2_nMax();
    }

    /** local variable for squareMod. Is glocal static for the sake of speed.
    */
    static Field2_n[] squareMul_tmp = new Field2_n[maxLen*2];

    /** local variable for squareMod. Is glocal static for the sake of speed.
    */
    static Field2_n squareMul_tmpCoeff = new Field2_n();

    /** Calculates this = in^2 mod modul
        and used the fact that characteristic p=2 for GF(2^n) which
        means that squaring in GF(2^n) is cheaper than multiplication.
        Assumes that in.degree() < modul.degree() and also that
        the leading coefficient of modul is 1.
       @param in Polynomial to be squared
       @param modul Modul which is used for reduction
       @param degree degree of the modul (only for the sake of speed)
    */
    void squareMod(Poly2_nMax in, Poly2_nMax modul, int degree) {
      Check.assert(modul.degree() == degree, "degree is wrong " + degree + " / " + modul.degree());
      Check.assert(!modul.isZero(), "modul is zero");
      Check.assert(in.degree() < degree, "degree of in is to big " + in.degree() + " / " + degree);
      Check.assert(modul.value[degree].isEqual(new Field2_n(1)), "modul has not 1 as leading coefficient" + modul.stringIt());

      // set the odd coefficients to zero
      for (int i = 1; i < degree*2; i += 2)
        squareMul_tmp[i].setZero();
      // exploit the fact that "in" has degree ("degree"-1) at most
      // And: Squaring is done here
      for (int i = 0; i < degree; i++)
        squareMul_tmp[i*2].mul(in.value[i],in.value[i]);

      // now compute reduction for our result
      for (int i = degree*2-2; i >= degree; i--) {
        for (int j = 0; j < degree; j++) {
          if (!modul.value[j].isZero()) {
            squareMul_tmpCoeff.mul(modul.value[j],squareMul_tmp[i]);
            squareMul_tmp[i-degree+j].add(squareMul_tmpCoeff);
          }
        }
      }

      // last but not least: copy the result to this
      for (int i = 0; i < degree; i++)
        value[i].copy(squareMul_tmp[i]);
      for (int i = degree; i < maxLen; i++)
        value[i].setZero();
    }

    /** local variable for gcd algorithm. Is static glocal for the sake of speed.
    */
    static private Poly2_nMax[] gcd_tmp = new Poly2_nMax[2];

    /** local variable for gcd algorithm. Is static glocal for the sake of speed.
    */
    static private Field2_n gcd_tmpCoeff = new Field2_n();

    /** Computes this = gcd(a,b)
        using an iterative gcd algorithm for polynomials. Assumes b != zero.
       @param a Polynomial
       @param b Polynomial
    */
    void gcd(Poly2_nMax a, Poly2_nMax b) {
      Check.assert(!b.isZero());

      gcd_tmp[0].copy(a);
      gcd_tmp[0].mod(b);
      if (gcd_tmp[0].isZero()) {
        // make result monic
        int deg = b.degree();
        gcd_tmpCoeff.copy(b.value[deg]);
        for (int i = 0; i < deg; i++)
          value[i].div(b.value[i],gcd_tmpCoeff);
        value[deg].setOne();
        for (int i = deg+1; i < maxLen; i++)
          value[i].setZero();
        return;
      }
      gcd_tmp[1].copy(b);

      int u = 1; int v = 0; int r;
      // we know that gcd_tmp[v] != 0 for our initial step
      while (true) {
        gcd_tmp[u].mod(gcd_tmp[v]);
        if (gcd_tmp[u].isZero()) {
          // make result monic
          int deg = gcd_tmp[v].degree();
          gcd_tmpCoeff.copy(gcd_tmp[v].value[deg]);
          for (int i = 0; i < deg; i++)
            value[i].div(gcd_tmp[v].value[i],gcd_tmpCoeff);
          value[deg].setOne();
          for (int i = deg+1; i < maxLen; i++)
            value[i].setZero();
            return;
        }
        // instead of copying polynomials, copy pointers to these polynomials!
        r = u;
        u = v;
        v = r;
      }
    }

    /** Uses the given polynomial as a function.
        Applies the value num (from F_64) to
        the polynomial.
       @return f(num)
    */
    Field2_n applyPoly(Field2_n num) {
      Field2_n res = new Field2_n(value[0]);  // result
      Field2_n cur = new Field2_n(1);
      Field2_n tmp = new Field2_n();
      for (int i = 1; i < maxLen; i++) {
        cur.mul(num);
        tmp.mul(value[i],cur);
        res.add(tmp);
      }
      return res;
    }

    /** transfer the polynomial to a string
    */
    String stringIt() {
      String erg = "";
      for (int i = maxLen-1; i > 0; i--) {
        if (!value[i].isZero())
          erg += value[i].stringIt() + "x^" + i + " + ";
      }
      erg += value[0].stringIt();
      return erg;
    }


    /** Tests the functionality for the whole class.
        Reports errors to stdout.
        <i>Works only for maxLen == 15</i>
    */
    static void testIt () {
      System.out.println("start testing Poly2_nMax ...");
      final int repeat = 100;
      Random rnd = new Random();

      Poly2_nMax cur = new Poly2_nMax(), cur1 = new Poly2_nMax(), cur2 = new Poly2_nMax(),
                cur3 = new Poly2_nMax(), cur4 = new Poly2_nMax(), cur5 = new Poly2_nMax(), cur6 = new Poly2_nMax();
      Field2_n coeff1 = new Field2_n(), coeff2 = new Field2_n(), coeff3 = new Field2_n(), coeff4 = new Field2_n();

      // first test the constructors
      coeff1 = new Field2_n(10);
      cur = new Poly2_nMax();
      cur.setCoeff(coeff1, 0);
      cur1 = new Poly2_nMax();
      if (cur.isEqual(cur1)) System.out.println("Error 0.0a with " + cur.stringIt() + " and " + cur1.stringIt());
      cur1 = new Poly2_nMax(cur);
      if (!cur.isEqual(cur1)) System.out.println("Error 0.0b with " + cur.stringIt() + " and " + cur1.stringIt());
      cur.setZero();
      if (!cur.isZero()) System.out.println("Error 0.0c with " + cur.stringIt());
      if (cur.isEqual(cur1)) System.out.println("Error 0.0d with " + cur.stringIt() + " and " + cur1.stringIt());
      if (cur1.isZero()) System.out.println("Error 0.0e with " + cur1.stringIt());
      coeff1 = cur1.getCoeff(2);
      if (!coeff1.isZero()) System.out.println("Error 0.0f with " + cur1.stringIt());

      // new test with new methods
      cur = new Poly2_nMax();
      if (!cur.isZero()) System.out.println("Error 0.1a with " + cur.stringIt());
      coeff1 = new Field2_n(300);
      cur.setCoeff(coeff1,2);
      if (cur.degree() != 2) System.out.println("Error 0.1b with " + cur.stringIt());
      cur.setCoeff(coeff1,maxLen-1);
      if (cur.degree() != maxLen-1) System.out.println("Error 0.1c with " + cur.stringIt());
      cur1 = new Poly2_nMax();
      if (cur1.degree() != -1) System.out.println("Error 0.1d with " + cur.stringIt());
      cur1.setCoeff(coeff1,1);
      cur = new Poly2_nMax(cur1,maxLen-2);
      if (cur.degree() != maxLen-1) System.out.println("Error 0.1e with " + cur.stringIt());
      cur2 = new Poly2_nMax();
      cur2.setCoeff(coeff1,maxLen-1);
      if (!cur.isEqual(cur2)) System.out.println("Error 0.1f with " + cur.stringIt() + " and " + cur2.stringIt());
      cur1.copyShift(cur,0);
      if (!cur1.isEqual(cur2)) System.out.println("Error 0.1g with " + cur1.stringIt() + " and " + cur2.stringIt());

      cur = new Poly2_nMax();
      cur.setCoeff(rnd,3);
      if (cur.degree() != 3) System.out.println("Error 0.2a with " + cur.stringIt());
      cur1 = new Poly2_nMax();
      cur1.setCoeff(rnd,0);
      cur.copyShift(cur1,maxLen-1);
      if (cur.degree() != maxLen-1) System.out.println("Error 0.2b with " + cur.stringIt());
      cur.setCoeff(new Field2_n(),maxLen-1);
      if (!cur.isZero()) System.out.println("Error 0.2c with " + cur.stringIt());

      // test addition
      for (int i = 0; i < repeat; i++) {
        cur1 = new Poly2_nMax(rnd);
        cur2 = new Poly2_nMax(rnd);
        int rndPos = rnd.nextInt(maxLen);
        coeff1 = cur1.getCoeff(rndPos);
        coeff2 = cur2.getCoeff(rndPos);
        coeff3.add(coeff1,coeff2);
        cur3.add(cur1,cur2);
        coeff4 = cur3.getCoeff(rndPos);
        if (!coeff4.isEqual(coeff3)) System.out.println("Error 0.3a at " + rndPos + " with " + cur1.stringIt() + ", "
                                     + cur2.stringIt() + ", " + cur3.stringIt());
        cur1.add(cur2);
        if (!cur1.isEqual(cur3)) System.out.println("Error 0.3b with " + cur1.stringIt() + ", "
                                     + cur2.stringIt() + ", " + cur3.stringIt());
      }

      // test multiplication
      cur1 = new Poly2_nMax();
      cur2 = new Poly2_nMax();
      coeff1 = new Field2_n(1);  // init it with 1
      cur1.setCoeff(coeff1,0);
      cur2.setCoeff(coeff1,0);
      cur3.mul(cur1,cur2);
      if (!cur1.isEqual(cur3)) System.out.println("Error 0.4a with " + cur1.stringIt() + ", "
                                     + cur2.stringIt() + ", " + cur3.stringIt());
      cur1 = new Poly2_nMax();
      cur2 = new Poly2_nMax();
      cur = new Poly2_nMax();
      coeff1 = new Field2_n(1);  // init it with 1
      cur1.setCoeff(coeff1,1);
      cur2.setCoeff(coeff1,1);
      cur.setCoeff(coeff1,2);
      cur3.mul(cur1,cur2);
      if (!cur.isEqual(cur3)) System.out.println("Error 0.4b with " + cur1.stringIt() + ", "
                                     + cur2.stringIt() + ", " + cur3.stringIt());

      cur1 = new Poly2_nMax();
      cur2 = new Poly2_nMax();
      cur = new Poly2_nMax();
      coeff1 = new Field2_n(1046);  // init it with 1
      cur1.setCoeff(coeff1,1);
      cur2.setCoeff(coeff1,1);
      coeff2.mul(coeff1,coeff1);
      cur.setCoeff(coeff2,2);
      cur3.mul(cur1,cur2);
      if (!cur.isEqual(cur3)) System.out.println("Error 0.4c with " + cur1.stringIt() + ", "
                                     + cur2.stringIt() + ", " + cur3.stringIt());
      if (!cur1.isEqual(cur2)) System.out.println("Error 0.4d with " + cur1.stringIt() + ", "
                                     + cur2.stringIt() + ", " + cur3.stringIt());

      // test mod
      cur1 = new Poly2_nMax();
      cur2 = new Poly2_nMax();
      cur = new Poly2_nMax();
      coeff1 = new Field2_n(1046);
      cur1.setCoeff(coeff1,1);
      cur2.setCoeff(coeff1,1);
      cur.setZero();
      cur3.mod(cur1,cur2);
      if (!cur.isEqual(cur3)) System.out.println("Error 1.0a with " + cur1.stringIt() + ", "
                                     + cur2.stringIt() + ", " + cur3.stringIt());
      if (!cur1.isEqual(cur2)) System.out.println("Error 1.0b with " + cur1.stringIt() + ", "
                                     + cur2.stringIt() + ", " + cur3.stringIt());

      cur1 = new Poly2_nMax();
      cur2 = new Poly2_nMax();
      cur = new Poly2_nMax();
      coeff1 = new Field2_n(1046);
      cur1.setCoeff(coeff1,2);
      cur2.setCoeff(coeff1,1);
      cur.setZero();
      cur1.mod(cur2);
      if (!cur.isEqual(cur1)) System.out.println("Error 1.1a with " + cur1.stringIt() + ", "
                                     + cur2.stringIt() + ", " + cur3.stringIt());


      cur1 = new Poly2_nMax();
      cur2 = new Poly2_nMax();
      cur = new Poly2_nMax();
      coeff1 = new Field2_n(1046);
      cur1.setCoeff(coeff1,1);
      cur2.setCoeff(coeff1,2);
      cur.copy(cur1);
      cur3.copy(cur2);
      cur1.mod(cur2);
      if (!cur.isEqual(cur1)) System.out.println("Error 1.2a with " + cur1.stringIt() + ", "
                                     + cur2.stringIt() + ", " + cur3.stringIt());
      if (!cur3.isEqual(cur2)) System.out.println("Error 1.2b with " + cur1.stringIt() + ", "
                                     + cur2.stringIt() + ", " + cur3.stringIt());


      cur1 = new Poly2_nMax();
      cur2 = new Poly2_nMax();
      cur = new Poly2_nMax();
      coeff1 = new Field2_n(1046);
      cur1.setCoeff(coeff1,1);
      coeff2 = new Field2_n(123);
      cur2.setCoeff(coeff2,1);
      cur.setZero();
      cur3.mod(cur1,cur2);
      if (!cur.isEqual(cur3)) System.out.println("Error 1.3a with " + cur1.stringIt() + ", "
                                     + cur2.stringIt() + ", " + cur3.stringIt());

      cur1 = new Poly2_nMax();
      cur2 = new Poly2_nMax();
      cur = new Poly2_nMax();
      coeff1 = new Field2_n(1046);
      cur1.setCoeff(coeff1,0);
      coeff2 = new Field2_n(123);
      cur2.setCoeff(coeff2,0);
      cur.setZero();
      cur3.mod(cur1,cur2);
      if (!cur.isEqual(cur3)) System.out.println("Error 1.4a with " + cur1.stringIt() + ", "
                                     + cur2.stringIt() + ", " + cur3.stringIt());
      // combine mul and mod
      for (int i = 0; i < 1; i++) {
        int rndDeg = rnd.nextInt(maxLen);
        cur1 = new Poly2_nMax(rnd,rndDeg);
        cur2 = new Poly2_nMax(rnd,maxLen-rndDeg-1);

        cur3.mul(cur1,cur2);
        cur3.mod(cur1);
        if (!cur3.isZero()) System.out.println("Error 1.5a");
        cur3.mul(cur1,cur2);
        cur3.mod(cur2);
        if (!cur3.isZero()) System.out.println("Error 1.5b");

        cur3.mul(cur1,cur2);
        cur4.copy(cur3);
        cur3.mod(cur4,cur2);
        if (!cur3.isZero()) System.out.println("Error 1.5c");
      }

      // test mulFactor
      for (int i = 0; i < repeat; i++) {
        cur1 = new Poly2_nMax(rnd);
        coeff1 = new Field2_n(rnd);
        int rndCoeff = rnd.nextInt(maxLen);
        cur1.getCoeff(coeff2,rndCoeff);
        cur1.mulFactor(coeff1);
        cur1.getCoeff(coeff3,rndCoeff);
        coeff2.mul(coeff1);
        if (!coeff2.isEqual(coeff3)) System.out.println("Error 1.6 at " + i);
      }

      // test additional functionality
      for (int i = 0; i < maxLen; i++) {
        cur1 = new Poly2_nMax(rnd,i);
        if (cur1.degree() != i) System.out.println("Error 2.0 at " + i + " with " + cur1.stringIt());
      }

      for (int i = 0; i < repeat; i++) {
        int rndDeg = rnd.nextInt(maxLen);
        cur1 = new Poly2_nMax(rnd,rndDeg);
        if (cur1.degree() != rndDeg) System.out.println("Error 2.1 at " + i + " with " + rndDeg + " and " + cur1.stringIt());
      }

      // test shift method
      cur1 = new Poly2_nMax();
      coeff1 = new Field2_n(1); // is one
      cur1.setCoeff(coeff1,1);  // is an x
      for (int i = 0; i < maxLen-1; i++) {
        cur2 = new Poly2_nMax(rnd,i);
        cur3.mul(cur1,cur2);
        cur2.shiftL(1);
        if (!cur2.isEqual(cur3)) System.out.println("Error 2.2 at " + i + " with " + cur2.stringIt() + " and " + cur3.stringIt());
      }


      for (int i = 0; i < repeat; i++) {
        int rndDeg = rnd.nextInt(maxLen);
        int rndShift;
        if (rndDeg < (maxLen-1))
          rndShift = rnd.nextInt(maxLen-rndDeg-1);
        else rndShift = 0;
        cur2 = new Poly2_nMax(rnd,rndDeg);
        cur1 = new Poly2_nMax();
        coeff1 = new Field2_n(1); // is one
        cur1.setCoeff(coeff1,rndShift);  // is an x^(rndShift)
        cur3.mul(cur1,cur2);
        cur2.shiftL(rndShift);
        if (!cur2.isEqual(cur3)) System.out.println("Error 2.3 at " + i + " with " + cur2.stringIt() + " and " + cur3.stringIt());
      }

      // test setZero
      cur1 = new Poly2_nMax(rnd);
      cur1.setZero();
      if (cur1.degree() != -1) System.out.println("Error 2.4 with " + cur1.stringIt());

      // test apply poly
      coeff1 = new Field2_n(1);  // set it to be one, i.e. we get the sum of all coefficients
      for (int i = 0; i < repeat; i++) {
        int deg = rnd.nextInt(maxLen);
        cur = new Poly2_nMax(rnd,deg);
        coeff2 = cur.applyPoly(coeff1);
        coeff3.setZero();
        for (int j = 0; j < maxLen; j++) {
          cur.getCoeff(coeff4,j);
          coeff3.add(coeff4);
        }
        if (!coeff2.isEqual(coeff3)) System.out.println("Error 2.5 at " + i);
      }

      // test if polynomial of degree 2 works
      coeff1 = new Field2_n(1);  // set it to be one
      cur = new Poly2_nMax();  // everything is zero
      cur.setCoeff(coeff1,2);
      for (int i = 0; i < repeat; i++) {
        coeff1 = new Field2_n(rnd);
        coeff2 = cur.applyPoly(coeff1);
        coeff3.mul(coeff1,coeff1);
        if (!coeff2.isEqual(coeff3)) System.out.println("Error 2.6 at " + i);
      }

      // test squareMod
      cur = new Poly2_nMax(rnd,2);
      cur.setCoeff(new Field2_n(1),2);
      cur1 = new Poly2_nMax(rnd,0);
      cur2.squareMod(cur1,cur,cur.degree());
      cur3.mul(cur1,cur1);
      if (!cur2.isEqual(cur3)) System.out.println("Error 3.1 with " + cur2.degree());

      cur = new Poly2_nMax(rnd,4);
      cur.setCoeff(new Field2_n(1),4);
      cur1 = new Poly2_nMax(rnd,1);
      cur2.squareMod(cur1,cur,cur.degree());
      cur3.mul(cur1,cur1);
      if (!cur2.isEqual(cur3)) System.out.println("Error 3.2 with " + cur2.degree());

      cur = new Poly2_nMax();  // x^4 + 1
      cur.setCoeff(new Field2_n(1),4);
      cur.setCoeff(new Field2_n(1),0);

      cur1 = new Poly2_nMax();
      cur1.setCoeff(new Field2_n(1),2);  // x^2
      cur2.squareMod(cur1,cur,cur.degree());  // result should be 1
      cur3.mul(cur1,cur1);
      cur3.mod(cur);        // result should be 1, too
      if (!cur2.isEqual(cur3)) {
        System.out.println("Error 3.3 with " + cur2.degree());
        System.out.println("  cur2 " + cur2.stringIt());
        System.out.println("  cur3 " + cur3.stringIt());
      }

      cur = new Poly2_nMax(rnd,4);
      cur.setCoeff(new Field2_n(1),4);
      cur1 = new Poly2_nMax(rnd,3);
      cur2.squareMod(cur1,cur,cur.degree());
      cur3.mul(cur1,cur1);
      cur3.mod(cur);
      if (!cur2.isEqual(cur3)) System.out.println("Error 3.4 with " + cur2.degree());

      for (int i = 1; i < maxLen/2; i++) {
        // produce modul with degree i
        cur = new Poly2_nMax(rnd,i-1);
        cur.setCoeff(new Field2_n(1),i);
        // produce input with random degree, but less than i
        int deg = rnd.nextInt(i);
        cur1 = new Poly2_nMax(rnd,deg);
        cur2.squareMod(cur1,cur,cur.degree());
        cur3.mul(cur1,cur1);
        cur3.mod(cur);
        if (!cur2.isEqual(cur3)) {
          System.out.println("Error 3.5 with " + cur2.degree());
          System.out.println("  cur  " + cur.stringIt());
          System.out.println("  cur1 " + cur1.stringIt());
          System.out.println("  cur2 " + cur2.stringIt());
          System.out.println("  cur3 " + cur3.stringIt());
        }
      }

      cur = new Poly2_nMax();
      cur.setCoeff(new Field2_n(1),0);
      cur1 = new Poly2_nMax(rnd,0);
      cur2 = new Poly2_nMax(rnd,0);
      cur3.gcd(cur1,cur2);
      if (!cur3.isEqual(cur)) System.out.println("Error 4.1 with " + cur3.stringIt());

      cur1 = new Poly2_nMax(rnd,3);
      cur2 = new Poly2_nMax(rnd,3);
      cur4.mul(cur1,cur2);
      cur3.gcd(cur4,cur2);
      cur3.mulFactor(cur2.value[3]);  // take into account that gcd is a _monic_ polynomial
      if (!cur3.isEqual(cur2)) System.out.println("Error 4.2 with " + cur3.degree());

      cur1 = new Poly2_nMax(rnd,3);
      cur2 = new Poly2_nMax(rnd,3);
      cur.mul(cur1,cur2);  // compute cur2^2 * cur1
      cur4.mul(cur,cur2);
      cur3.gcd(cur2,cur4); // get gcd(cur2^2*cur1,cur2) = cur2 (monic)
      cur3.mulFactor(cur2.value[3]);  // take into account that gcd is a _monic_ polynomial
      if (!cur3.isEqual(cur2)) System.out.println("Error 4.3 with " + cur3.degree());

      // test a more complicated case:
      cur = new Poly2_nMax(rnd,0);
      cur.setCoeff(new Field2_n(1),1);  // set coeff for x
      cur1 = new Poly2_nMax(rnd,0);
      cur2 = new Poly2_nMax(rnd,0);

      for (int i = 1; i < maxLen-1; i++) {
        cur.setCoeff(new Field2_n(i),0);  // set coeff for 1
        cur3.mul(cur1,cur);
        cur1.copy(cur3);

        cur.setCoeff(new Field2_n(i+maxLen),0);  // set coeff for 1
        cur3.mul(cur2,cur);
        cur2.copy(cur3);
      }
      cur = new Poly2_nMax(rnd,0);
      cur.setCoeff(new Field2_n(1),0);  // set coeff for 1
      cur3.gcd(cur1,cur2);
      if (!cur3.isEqual(cur)) System.out.println("Error 4.4 with " + cur3.degree());

      // test a more complicated case
      cur1 = new Poly2_nMax(rnd,0);
      cur2 = new Poly2_nMax(rnd,0);

      cur1 = new Poly2_nMax();
      cur1.setCoeff(new Field2_n(1),0);
      cur1.setCoeff(new Field2_n(1),1);
      cur2 = new Poly2_nMax();
      cur2.setCoeff(new Field2_n(2),0);
      cur2.setCoeff(new Field2_n(1),1);
      cur.mul(cur1,cur2);

      cur1.copy(cur);
      cur2.copy(cur);

      cur3.gcd(cur1,cur2);
      if (!cur3.isEqual(cur)) {
        System.out.println("Error 4.5 with " + cur3.degree());
        System.out.println("  cur  " + cur.stringIt());
        System.out.println("  cur1 " + cur1.stringIt());
        System.out.println("  cur2 " + cur2.stringIt());
        System.out.println("  cur3 " + cur3.stringIt());
      }

      // test a more complicated case:
      cur1 = new Poly2_nMax(rnd,0);
      cur2 = new Poly2_nMax(rnd,0);
      cur = new Poly2_nMax(rnd,0);
      cur.setCoeff(new Field2_n(1),1);  // set coeff for x

      cur1 = new Poly2_nMax();
      cur1.setCoeff(new Field2_n(1),0);
      cur1.setCoeff(new Field2_n(1),1);
      cur2 = new Poly2_nMax();
      cur2.setCoeff(new Field2_n(2),0);
      cur2.setCoeff(new Field2_n(1),1);
      cur4.mul(cur1,cur2);

      cur1.copy(cur4);
      cur2.copy(cur4);

      for (int i = 3; i < maxLen-1; i++) {
        cur.setCoeff(new Field2_n(i),0);  // set coeff for 1
        cur3.mul(cur1,cur);
        cur1.copy(cur3);

        cur.setCoeff(new Field2_n(i+maxLen),0);  // set coeff for 1
        cur3.mul(cur2,cur);
        cur2.copy(cur3);
      }
      cur = new Poly2_nMax(rnd,0);
      cur.setCoeff(new Field2_n(1),0);  // set coeff for 1
      cur3.gcd(cur1,cur2);
      if (!cur3.isEqual(cur4)) {
        System.out.println("Error 4.6 with " + cur3.degree());
        System.out.println("  cur  " + cur.stringIt());
        System.out.println("  cur1 " + cur1.stringIt());
        System.out.println("  cur2 " + cur2.stringIt());
        System.out.println("  cur3 " + cur3.stringIt());
        System.out.println("  cur4 " + cur4.stringIt());
      }

      // new test case: square factor!
      cur1 = new Poly2_nMax(rnd,0);
      cur2 = new Poly2_nMax(rnd,0);
      cur = new Poly2_nMax(rnd,0);
      cur.setCoeff(new Field2_n(1),1);  // set coeff for x

      cur1 = new Poly2_nMax();
      cur1.setCoeff(new Field2_n(1),0);
      cur1.setCoeff(new Field2_n(1),1);
      cur2 = new Poly2_nMax();
      cur2.setCoeff(new Field2_n(2),0);
      cur2.setCoeff(new Field2_n(1),1);
      cur4.mul(cur1,cur2);  // give degree 2
      cur.mul(cur4,cur4);   // gives degree 4

      cur1.copy(cur);
      cur2.copy(cur);
      cur4.copy(cur);

      cur = new Poly2_nMax(rnd,0);
      cur.setCoeff(new Field2_n(1),1);  // set coeff for x
      for (int i = 4; i < maxLen-1; i++) {
        cur.setCoeff(new Field2_n(i),0);  // set coeff for 1
        cur3.mul(cur1,cur);
        cur1.copy(cur3);

        cur.setCoeff(new Field2_n(i+maxLen),0);  // set coeff for 1
        cur3.mul(cur2,cur);
        cur2.copy(cur3);
      }
      cur = new Poly2_nMax(rnd,0);
      cur.setCoeff(new Field2_n(1),0);  // set coeff for 1
      cur3.gcd(cur1,cur2);
      if (!cur3.isEqual(cur4)) {
        System.out.println("Error 4.7 with " + cur3.degree());
        System.out.println("  cur  " + cur.stringIt());
        System.out.println("  cur1 " + cur1.stringIt());
        System.out.println("  cur2 " + cur2.stringIt());
        System.out.println("  cur3 " + cur3.stringIt());
        System.out.println("  cur4 " + cur4.stringIt());
      }

      // last but not least: two polynomials with x^n each
      cur1 = new Poly2_nMax();
      cur1.setCoeff(new Field2_n(rnd),maxLen-1);
      cur2 = new Poly2_nMax();
      cur2.setCoeff(new Field2_n(rnd),maxLen/2);
      cur = new Poly2_nMax();
      cur.setCoeff(new Field2_n(1),maxLen/2);  // result we want to have

      cur3.gcd(cur1,cur2);
      if (!cur3.isEqual(cur)) {
        System.out.println("Error 4.8 with " + cur3.degree());
        System.out.println("  cur  " + cur.stringIt());
        System.out.println("  cur1 " + cur1.stringIt());
        System.out.println("  cur2 " + cur2.stringIt());
        System.out.println("  cur3 " + cur3.stringIt());
        System.out.println("  cur4 " + cur4.stringIt());
      }

      // test mulAdd
      // compute cur1 += coeff1 * cur2
      for (int i = 0; i < repeat; i++) {
        cur1 = new Poly2_nMax(rnd);
        coeff1.rnd(rnd);
        cur2 = new Poly2_nMax(rnd);
        coeff1.rnd(rnd);

        cur3.copy(cur2);
        cur3.mulFactor(coeff1);
        cur4.copy(cur1);
        cur4.add(cur3);

        cur1.mulAdd(coeff1,cur2);
        if (!cur1.isEqual(cur4)) System.out.println("Error 5.1 with " + cur3.degree());
      }

      // test division
      // compute cur3 = cur1 * cur2 and then
      // cur4 = cur3 / cur1
      // cur4 = cur3 / cur2 and compare the results
      for (int i = 0; i < repeat; i++) {
        int deg = rnd.nextInt(maxLen-1);
        cur1 = new Poly2_nMax(rnd,deg);
        cur2 = new Poly2_nMax(rnd,rnd.nextInt(maxLen-deg));
        cur3.mul(cur1,cur2);
        cur4.div(cur3,cur1);
        if (!cur2.isEqual(cur4)) System.out.println("Error 5.2a with " + cur4.degree());
        cur4.div(cur3,cur2);
        if (!cur1.isEqual(cur4)) System.out.println("Error 5.2b with " + cur4.degree());
      }

      // test slow and "fast" mod operation
      for (int i = 0; i < repeat; i++) {
        cur1 = new Poly2_nMax(rnd,rnd.nextInt(maxLen));
        cur2 = new Poly2_nMax(rnd,rnd.nextInt(maxLen));
        cur3.copy(cur1);
        cur3.modS(cur2);
        cur1.mod(cur2);
        if (!cur1.isEqual(cur3)) System.out.println("Error 5.3");
      }
      // test slow and "fast" div operation
      for (int i = 0; i < repeat; i++) {
        int deg = rnd.nextInt(maxLen-1);
        cur1 = new Poly2_nMax(rnd,deg);
        cur2 = new Poly2_nMax(rnd,rnd.nextInt(maxLen-deg));
        cur3.mul(cur1,cur2);
        cur4.div(cur3,cur1);
        if (!cur2.isEqual(cur4)) System.out.println("Error 5.4a with " + cur4.degree());
        cur4.div(cur3,cur2);
        if (!cur1.isEqual(cur4)) System.out.println("Error 5.4b with " + cur4.degree());
        cur1.divS(cur3,cur2);
        if (!cur1.isEqual(cur4)) System.out.println("Error 5.4c with " + cur4.degree());
      }
      System.out.println("...finished testing.");
    }


    public final static void speedIt() {
    Check.assert(false);  // make sure we do not have any assertion checkers


      final int repeats = 100;
      Random rnd = new Random();

      Poly2_nMax[] X = new Poly2_nMax[repeats], Y = new Poly2_nMax[repeats];

      System.out.println();

      long curTime = System.currentTimeMillis();
      // create random polynomials first
      for (int i = 0; i < repeats; i++) {
        X[i] = new Poly2_nMax(rnd);
        Y[i] = new Poly2_nMax(rnd);
      }

      curTime = System.currentTimeMillis() - curTime;
      System.out.println("Generation of random values took " + curTime + "ms (" + repeats + " operations)");

      // temporary register for multiplication / division
      Poly2_nMax cur = new Poly2_nMax();

      curTime = System.currentTimeMillis();
      // create random polynomials first
      for (int i = 0; i < repeats; i++) {
        X[i].modS(Y[i]);
      }
      curTime = System.currentTimeMillis() - curTime;
      System.out.println("modS " + curTime + "ms (" + repeats + " operations)");

      curTime = System.currentTimeMillis();
      // create random polynomials first
      for (int i = 0; i < repeats; i++) {
        X[i] = new Poly2_nMax(rnd);
        Y[i] = new Poly2_nMax(rnd);
      }

      curTime = System.currentTimeMillis() - curTime;
      System.out.println("Generation of random values took " + curTime + "ms (" + repeats + " operations)");
      curTime = System.currentTimeMillis();
      // create random polynomials first
      for (int i = 0; i < repeats; i++) {
        X[i].mod(Y[i]);
      }
      curTime = System.currentTimeMillis() - curTime;
      System.out.println("mod " + curTime + "ms (" + repeats + " operations)");

      curTime = System.currentTimeMillis();
      // create random polynomials first
      for (int i = 0; i < repeats; i++) {
        int deg = rnd.nextInt(maxLen-1);
        cur = new Poly2_nMax(rnd);
        Y[i] = new Poly2_nMax(rnd);
        X[i].mul(cur,Y[i]);
      }

      curTime = System.currentTimeMillis() - curTime;
      System.out.println("Generation of random values took " + curTime + "ms (" + repeats + " operations)");

      curTime = System.currentTimeMillis();
      // create random polynomials first
      for (int i = 0; i < repeats; i++) {
        cur.divS(X[i],Y[i]);
      }
      curTime = System.currentTimeMillis() - curTime;
      System.out.println("divS " + curTime + "ms (" + repeats + " operations)");

      curTime = System.currentTimeMillis();
      cur = new Poly2_nMax();
      // create random polynomials first
      for (int i = 0; i < repeats; i++) {
        cur.div(X[i],Y[i]);
      }
      curTime = System.currentTimeMillis() - curTime;
      System.out.println("div " + curTime + "ms (" + repeats + " operations)");
    }

    /** Mainly invoces the {@link #testIt} method.
    */
    public static void main(String[] args) {
      initAll();
      testIt();
      //speedIt();
      //speedIt();
      //speedIt();
    }
}
