package HFE;

import java.util.*;
import HFE.*;

//-----------------------------------------------------------------------------
// Project:     Implementing HFE at UCC as Project CS4095
// Supervisors: Simon Foley (Computer Science) and Patrick Fitzpatrick (Maths)
//-----------------------------------------------------------------------------

/**
<p>
This class provides polynomials of positive degree over
the finite field F_4 which is provided in the class {@link Field4}.
It also provides addition, multiplication and modulo-operations
The maximal degree of the polynomials is fixed by the constant
{@link #maxLen}
<p>
@author Christopher Wolf (hfe@Christopher-Wolf.de)
@version 0.1
*/

//Copyright (C) 2002    Christopher Wolf
//
//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License (file LICENCE) for more details.
 

final class Poly4Max {

    /** Maximal length of the polynomials
      @serial constant
    */
    public final static int maxLen = 21;

    /** Modulus for the mod version of the operations
      @serial constant
    */
    private static Poly4Max modulus;

    /** Current value of the polynomial.
        The polynomial has at most degree (maxLen-1). Its coefficients are
        taken from F_4. The element value[i] is the coefficient for x^i so
        value[0] is the constant term for this polynomial.
    */
    private int value[];


    /** sets all the coefficients to zero
    */
    Poly4Max() {
      setZero();
    }

    /** Creates a new polynomial. {@link #int2poly}
    */
    Poly4Max(int initValue) {
      setZero();
      int2poly(initValue);
    }

    /** Creates a copy from the given polynomial
    */
    Poly4Max(final Poly4Max father) {
      value = (int[])father.value.clone();
    }

    /** Takes one polynomial and shifts it by the given shift.
        It basicly leaves out shift number of coefficients
        and starts copying at value[shift].
        Makes a copy of this polynomial.
    */
    // !! Exception
    Poly4Max(final Poly4Max father, int shift) {
      setZero();
      if ((father.degree()+shift) >= maxLen) {
        System.out.println("Poly4Max: shift out of range");
        return;
      }
      for (int i = 0; i <= father.degree(); i++) {
        value[shift+i] = father.value[i];
      }
    }


    /** {@link #array2poly}
    */
    // !! Exception
    Poly4Max(final int[] inArray) {
      array2poly(inArray);
    }

    /** Sets the private field modulus.
        Does not clone the moduls but storing it. Checks if the degree
        of the moduls is less than (maxLen/2).
    */
    // !! Exception
    static void setModulus(final Poly4Max newModulus) {
      // security check: is the degree of modulus small enough?
      if ((newModulus.degree()*2) > maxLen) {
        System.out.println("Poly4Max.setModulus: modulus too big");
      }
      modulus = newModulus;
    }


    /** Returns the degree of the object.
        The degree of a constant polynomial is 0, the degree of the zero
        polynomial is -1.
    */
    int degree() {
      for (int i = maxLen-1; i >= 0; i--)
        if (value[i] != 0) return i;
      return -1;
    }

    /** Returns te coefficient for a given degree which (i.e. coeff*x^which) within the given polynomial.
    */
    int coeff(int which) {
      return value[which];
    }


    /** Compares the current polynomial with the given one.
      @return true iff (this < whom)
    */
    boolean isLess(final Poly4Max whom) {
      for (int i = maxLen-1; i >= 0; i--) {
        if (value[i] < whom.value[i]) return true;
        if (value[i] > whom.value[i]) return false;
      }
      return false;
    }

    /** Compares the current polynomial with the given one.
      @return true iff (this > whom)
    */
    boolean isBigger(final Poly4Max whom) {
      for (int i = maxLen-1; i >= 0; i--) {
        if (value[i] < whom.value[i]) return false;
        if (value[i] > whom.value[i]) return true;
      }
      return false;
    }


    /** Compares the current polynomial with the given one.
      @return true iff (this == whom) for all coefficients
    */
    boolean isEqual(final Poly4Max whom) {
      for (int i = maxLen-1; i >= 0; i--) {
        if (value[i] != whom.value[i]) return false;
      }
      return true;
    }

    /** Checks if a polynomial is the zero polynomial.
      @return true iff all coefficients are zero
    */
    boolean isZero() {
      for (int i = 0; i < maxLen; i++)
        if (value[i] != 0) return false;
      return true;
    }


    /** Sets all the coefficients to zero by allocating new memory
    */
    // !! Version without new memory
    void setZero() {
      value = new int[maxLen];
    }

    /** Sets the coefficients according to intValue.
        The bits 0/1 are the coefficient for
        x^0, bits 2/3 are the coefficient for x^1, ...
    */
    void int2poly(int intValue) {
      Check.assert(intValue >= 0);
      Check.assert(maxLen <= 15);
      for (int i = 0; i < maxLen; i++)
        value[i] =  ((intValue >> (i*2)) & 3);
    }


    /** Sets the coefficients using the given coefficient / power array.
        The elements in this array have the form {c_0,p_0, c_1,p_1, ...}
        where c_i is the coefficient for x^(p_i). If there are doublicate
        values for p_i, the last one is taken.
    */
    // !! Exception for wrong array length, power to big
    void array2poly(final int[] inArray) {
      setZero();
      if (inArray.length % 2 == 1) {
        System.out.println("Poly4Max: wrong array length.");
        return;
      }
      for (int i = 0; i < inArray.length; i+=2)
        value[inArray[i+1]] = inArray[i];
    }


    /** multiplies all coefficients by the given factor in F_4.
    */
    void mulFactor(int factor) {
      for (int i = 0; i < maxLen; i++)
        value[i] = Field4.mul(factor,value[i]);
    }


    /** calculates this = wen1 + wen2
        as polynomial addition
    */
    void add(final Poly4Max a, final Poly4Max b) {
      for (int i = 0; i < maxLen; i++)
        value[i] = Field4.add(a.value[i],b.value[i]);
    }


    /** calculate this += whom
        as polynomial addition
    */
    void add(final Poly4Max whom) {
      for (int i = 0; i < maxLen; i++)
        value[i] = Field4.add(whom.value[i],value[i]);
    }

    /** calculates this = a * b
        as polynomial addition.
        Does <b>not</b> check if the result is of degree < maxLen
    */
    void mul(final Poly4Max a, final Poly4Max b) {
      // first reset this
      setZero();
      for (int i = 0; i < maxLen; i++) {
        // only do something if it is necessary
        if (a.value[i] != 0) {
          for (int j = 0; j+i < maxLen; j++) {
            if (b.value[j] != 0) {
              value[i+j] = Field4.add(value[i+j],Field4.mul(a.value[i],b.value[j]));  // add at the correct place
            }
          }
        }
      }
    }


    /** calculates this = base^by.
        Does <b>not</br> check if the result is of degree < maxLen
    */
    void pow(final Poly4Max base, int by) {
      if (by == 0) {
        setZero();
        value[0] = 1;
        return;
      }
      if ((by % 2) == 1) this.value = (int[])base.value.clone(); else {setZero(); value[0] = 1;}
      by /= 2;
      Poly4Max square = new Poly4Max(base);
      Poly4Max tmp = new Poly4Max();
      do {
        tmp.mul(square,square);
        square.value = (int[])tmp.value.clone();
        if ((by % 2) == 1) {
          tmp.mul(square,this);
          this.value = (int[])tmp.value.clone();
        }
        by /= 2;
      } while (by != 0);
    }

    /** Calculates this %= moduls.
        Uses the static field modulus.
    */
    void mod () {
      mod(modulus);
    }

    /** Calculates this = a % b
    */
    void mod (final Poly4Max a, final Poly4Max b) {
      // copy before doing anything
      value = (int[])a.value.clone();
      mod(b);
    }

    /** Calculates this %= b
    */
    void mod (final Poly4Max b) {
      // do not divide by zero
      if (b.isZero()) {
        System.out.println("Division by Zero Error in mod!");
        setZero();
      }

      // work so long as there is something to do
      while (b.degree() <= this.degree()) {
        Poly4Max additive = new Poly4Max(b,this.degree()-b.degree());
        additive.mulFactor(Field4.div(value[degree()], b.value[b.degree()]));
        add(additive);
      }
    }

    /** Computes this = (a+b)%modul.
        Assumes a<modul, b<modul. Does <b>not</b> check this condition.
    */
    void addMod(final Poly4Max a, final Poly4Max b) {
      add(a,b);
      mod();
    }

    /** Computes this = (this+b)%modul.
        Assumes this<modul, b<modul. Does <b>not</b> check this condition.
    */
    void addMod(final Poly4Max b) {
      add(b);
      mod();
    }

    /** Computes this = (a*b)%modul.
        Assumes a<modul, b<modul. Does <b>not</b> check this condition.
    */
    void mulMod(final Poly4Max a, final Poly4Max b) {
      mul(a,b);
      mod();
    }

    /** Computes this = (base^by)%modul.
        Assumes base<modul. Does <b>not</b> check this condition.
    */
    void powMod(final Poly4Max base, int by) {
      if (by == 0) {
        setZero();
        value[0] = 1;
        return;
      }
      if ((by % 2) == 1) this.value = (int[])base.value.clone(); else {setZero(); value[0] = 1;}
      by /= 2;
      Poly4Max square = new Poly4Max(base);
      Poly4Max tmp = new Poly4Max();
      do {
        tmp.mulMod(square,square);
        square.value = (int[])tmp.value.clone();
        if ((by % 2) == 1) {
          tmp.mulMod(square,this);
          this.value = (int[])tmp.value.clone();
        }
        by /= 2;
      } while (by != 0);
    }

    /** Writes the polynomial as a string to stdout
    */
    void writeString() {
      for (int i = maxLen-1; i > 0; i--) {
        if (value[i] != 0) {
          if (value[i] != 1) System.out.print(value[i]);
          System.out.print("x^" + i + " + ");
        }
      }
      System.out.print(value[0]);
      System.out.println(" (" + polyInt() + ")");
    }

    /** Returns the polynomial as an integer.
       Bits 0/1 are the coefficient for x^0, Bits 2/3 are the coefficients for x^1, ...
    */
    int polyInt() {
      int power = 1;
      int res = 0;
      for (int i = 0; i < maxLen; i++) {
        res += value[i]*power;
        power *= 4;
      }
      return res;
    }

    /** Tests the functionality for the whole class.
        Reports errors to stdout.
        <i>Works only for maxLen == 15</i>
    */
    static void testIt () {
      System.out.println("start testing Poly4Max ...");

      Random rnd = new Random();
      Poly4Max cur = new Poly4Max(6), cur1 = new Poly4Max(), cur2 = new Poly4Max(), cur3 = new Poly4Max();
      int curRnd, curRnd1, curRnd2, curRnd3, curRes;

      // first test the constructors and the I/O functions
      Poly4Max first = new Poly4Max();
      Poly4Max second = new Poly4Max(3);
      Poly4Max third = new Poly4Max(second);

      if (second.polyInt() != 3) System.out.println("Error 1.1");
      if (third.polyInt() != 3) System.out.println("Error 1.2a");
      if (second.coeff(0) != 3) System.out.println("Error 1.2b");
      if (second.coeff(1) != 0) System.out.println("Error 1.2c");
      //second.writeString();

      cur = new Poly4Max(130);
      if (130 != cur.polyInt()) System.out.println("Error 1.3 with " + cur.polyInt());
      cur = new Poly4Max(1024);
      if (1024 != cur.polyInt()) System.out.println("Error 1.4 with " + cur.polyInt());
      for (int i = 0; i < 100; i++) {
        curRnd = rnd.nextInt(1 << (maxLen*2));  // gives values in 0..2^30-1
        cur = new Poly4Max(curRnd);
        cur2 = new Poly4Max(cur);
        if (curRnd != cur.polyInt()) System.out.println("Error 1.5 with " + curRnd);
        if (curRnd != cur2.polyInt()) System.out.println("Error 1.6 with " + curRnd);
      }
      int [] zwArr = {1,0, 2,1};
      cur = new Poly4Max(new int [] {1,0, 2,1});
      if (9 != cur.polyInt()) System.out.println("Error 1.7 with " + cur.polyInt());
      cur.array2poly(new int [] {1,1, 2,2, 3,3});
      if (228 != cur.polyInt()) System.out.println("Error 1.8 with " + cur.polyInt());


      // test isBigger, isLess, isEqual
      for (int i = 0; i < 100; i++) {
        curRnd1 = rnd.nextInt(1 << (maxLen*2));
        curRnd2 = rnd.nextInt(1 << (maxLen*2));
        cur1 = new Poly4Max(curRnd1);
        cur2 = new Poly4Max(curRnd2);

        // we have now three different cases:
        if (curRnd1 < curRnd2) {
          if (!cur1.isLess(cur2)) System.out.println("Error 2.1a with " + curRnd1 + " " + curRnd2);
          if (cur1.isBigger(cur2)) System.out.println("Error 2.1b with " + curRnd1 + " " + curRnd2);
          if (cur1.isEqual(cur2)) System.out.println("Error 2.1c with " + curRnd1 + " " + curRnd2);
          if (!cur1.isEqual(cur1)) System.out.println("Error 2.1d with " + curRnd1 + " " + curRnd2);
          if (cur2.isZero()) System.out.println("Error 2.1e with " + curRnd1 + " " + curRnd2);
        } else {
          if (curRnd1 > curRnd2) {
            if (cur1.isLess(cur2)) System.out.println("Error 2.2a with " + curRnd1 + " " + curRnd2);
            if (!cur1.isBigger(cur2)) System.out.println("Error 2.2b with " + curRnd1 + " " + curRnd2);
            if (cur1.isEqual(cur2)) System.out.println("Error 2.2c with " + curRnd1 + " " + curRnd2);
            if (!cur2.isEqual(cur2)) System.out.println("Error 2.2d with " + curRnd1 + " " + curRnd2);
            if (cur1.isZero()) System.out.println("Error 2.2e with " + curRnd1 + " " + curRnd2);
          } else {
            if (cur1.isLess(cur2)) System.out.println("Error 2.3a with " + curRnd1 + " " + curRnd2);
            if (cur1.isBigger(cur2)) System.out.println("Error 2.3b with " + curRnd1 + " " + curRnd2);
            if (!cur1.isEqual(cur2)) System.out.println("Error 2.3c with " + curRnd1 + " " + curRnd2);
            if (!cur2.isEqual(cur1)) System.out.println("Error 2.3d with " + curRnd1 + " " + curRnd2);
          }
        }
      }
      // test if the "isNull" parameter works
      cur = new Poly4Max(0);
      if (!cur.isZero()) System.out.println("Error 2.4");
      cur = new Poly4Max(23);
      cur.setZero();
      if (!cur.isZero()) System.out.println("Error 2.5");

      //Test addition
      cur1 = new Poly4Max();
      cur1.add(new Poly4Max((1 << (maxLen-1)) + 1), new Poly4Max(3));
      if (cur1.polyInt() != ((1 << (maxLen-1)) + 2)) System.out.println("Error 3.1 with " + cur1.polyInt());
      for (int i = 0; i < 100; i++) {
        curRnd1 = rnd.nextInt(1 << maxLen);
        curRnd2 = rnd.nextInt(1 << maxLen);
        curRes = curRnd1 ^ curRnd2;
        cur1 = new Poly4Max();
        cur1.add(new Poly4Max(curRnd1), new Poly4Max(curRnd2));
        cur2 = new Poly4Max(curRnd1);
        cur2.add(new Poly4Max(curRnd2));
        if (!cur1.isEqual(cur2)) System.out.println("Error 3.2 with " + curRnd1 + " " + curRnd2);
        if (cur1.polyInt() != curRes) System.out.println("Error 3.3 with " + curRnd1 + " " + curRnd2);
      }

      // Test multiplication
      cur = new Poly4Max();
      cur.mul(new Poly4Max(2), new Poly4Max(5));
      if (cur.polyInt() != 10) System.out.println("Error 4.1 with " + cur.polyInt());
      cur.mul(new Poly4Max(3), new Poly4Max(3));
      if (cur.polyInt() != 2) System.out.println("Error 4.2 with " + cur.polyInt());
      cur.mul(new Poly4Max(10), new Poly4Max(21));
      if (cur.polyInt() != 130) System.out.println("Error 4.3 with " + cur.polyInt());

      // Test both addition and multiplication
      // using the distributive law a(b+c)=ab+ac
      for (int i = 0; i < 100; i++) {
        curRnd1 = rnd.nextInt(1 << (maxLen/2));
        curRnd2 = rnd.nextInt(1 << (maxLen/2));
        curRnd3 = rnd.nextInt(1 << (maxLen/2));
        // left hand side
        cur1 = new Poly4Max(curRnd2);
        cur1.add(new Poly4Max(curRnd3));
        cur2.mul(cur1,new Poly4Max(curRnd1));
        curRes = cur2.polyInt();
        // right hand side
        cur1.mul(new Poly4Max(curRnd1), new Poly4Max(curRnd2));
        cur2.mul(new Poly4Max(curRnd1), new Poly4Max(curRnd3));
        cur3.add(cur1,cur2);

        if (cur3.polyInt() != curRes) System.out.println("Error 4.3 with " + curRnd1 + " " + curRnd2 + " " + curRnd3);
      }

      // test the shift constructor
      cur1 = new Poly4Max(new Poly4Max(3),2);
      if (cur1.polyInt() != (3 << 4)) System.out.println("Error 4.4");
      cur1 = new Poly4Max(new Poly4Max(1 << 28),0);
      if (cur1.polyInt() != (1 << 28)) System.out.println("Error 4.5");
      cur1 = new Poly4Max(new Poly4Max(3),13);
      if (cur1.polyInt() != (3 << 26)) System.out.println("Error 4.6");

      // Test multiplication together with the shift constructor
      for (int i = 0; i < 100; i++) {
        curRnd1 = rnd.nextInt(maxLen-1)+1;
        curRnd2 = rnd.nextInt(1 << ((maxLen-1)/curRnd1));
        cur1 = new Poly4Max(curRnd2);
        cur2.mul(cur1,new Poly4Max(1 << (curRnd1*2)));
        cur3 = new Poly4Max(cur1,curRnd1);

        if (!cur3.isEqual(cur2)) System.out.println("Error 4.7 with " + curRnd1 + " " + curRnd2);
      }

      // Test power function
      cur1.pow(new Poly4Max(5),2);  // x + 1  ->  x^2 + 1
      if (cur1.polyInt() != 17) System.out.println("Error 4.8a with " + cur1.polyInt());
      cur1.pow(new Poly4Max(9),3);  // 2x + 1  ->  x^3 + 3x^2 + 2x + 1
      if (cur1.polyInt() != 121) System.out.println("Error 4.8b with " + cur1.polyInt());
      cur1.pow(new Poly4Max(3),6);  // 3^6  ->  1
      if (cur1.polyInt() != 1) System.out.println("Error 4.8c with " + cur1.polyInt());
      cur1.pow(new Poly4Max(5),6);  // x + 1  -> x^6 + x^4 + x^2 + 1
      if (cur1.polyInt() != 4369) System.out.println("Error 4.8d with " + cur1.polyInt());
      cur1.pow(new Poly4Max(3),1);
      if (cur1.polyInt() != 3) System.out.println("Error 4.8e with " + cur1.polyInt());
      cur1.pow(new Poly4Max(3),0);
      if (cur1.polyInt() != 1) System.out.println("Error 4.8f with " + cur1.polyInt());
      // test mulFactor
      cur1 = new Poly4Max(225);  // 3x^3 + 2x^2 + 1  -> 0
      cur1.mulFactor(0);
      if (cur1.polyInt() != 0) System.out.println("Error 4.9a with " + cur1.polyInt());
      cur1 = new Poly4Max(225);  // 3x^3 + 2x^2 + 1, the same
      cur1.mulFactor(1);
      if (cur1.polyInt() != 225) System.out.println("Error 4.9b with " + cur1.polyInt());
      cur1 = new Poly4Max(225);  // 3x^3 + 2x^2 + 1  -> x^3 + 3x^2 + 2
      cur1.mulFactor(2);
      if (cur1.polyInt() != 114) System.out.println("Error 4.9c with " + cur1.polyInt());
      cur1 = new Poly4Max(225);  // 3x^3 + 2x^2 + 1  ->  2x^3 + x^2 + 3
      cur1.mulFactor(3);
      if (cur1.polyInt() != 147) System.out.println("Error 4.9d with " + cur1.polyInt());

      // test mod function
      cur1.mod(new Poly4Max(1), new Poly4Max(1)); // 1  %  1  -> 0
      if (cur1.polyInt() != 0) System.out.println("Error 5.1 with " + cur1.polyInt());
      cur1.mod(new Poly4Max(128), new Poly4Max(11)); // 2x^3  %  2x + 3  ->  2
      if (cur1.polyInt() != 2) System.out.println("Error 5.2 with " + cur1.polyInt());
      cur1.mod(new Poly4Max(129), new Poly4Max(5)); // 2x^3 + 1  %  x + 1  ->  3
      if (cur1.polyInt() != 3) System.out.println("Error 5.3 with " + cur1.polyInt());
      cur1.mod(new Poly4Max(180), new Poly4Max(49)); // 2x^3 + 3x^2 + x  %  3x^2 + 1  ->  2x + 1
      if (cur1.polyInt() != 9) System.out.println("Error 5.4 with " + cur1.polyInt());
      cur1.mod(new Poly4Max(75), new Poly4Max(17)); // x^3 + 2x + 3  %  x^2 + 1  -> 3x + 3
      if (cur1.polyInt() != 15) System.out.println("Error 5.5 with " + cur1.polyInt());
      cur1.mod(new Poly4Max(29), new Poly4Max(17)); // x^2 + 3x + 1  %  x^2 + 1  ->  3x
      if (cur1.polyInt() != 12) System.out.println("Error 5.6 with " + cur1.polyInt());
      cur1.mod(new Poly4Max(10), new Poly4Max(6)); // 2x + 2  %  x + 2  ->  1
      if (cur1.polyInt() != 1) System.out.println("Error 5.7 with " + cur1.polyInt());
      cur1.mod(new Poly4Max(32), new Poly4Max(1)); // 2x^2  %  1  -> 0
      if (cur1.polyInt() != 0) System.out.println("Error 5.8 with " + cur1.polyInt());
      cur1.mod(new Poly4Max(31), new Poly4Max(5)); // x^2 + 3x + 3  %  x + 1  ->  1
      if (cur1.polyInt() != 1) System.out.println("Error 5.9 with " + cur1.polyInt());
      cur1.mod(new Poly4Max(15), new Poly4Max(3)); // 3x + 3  %  3  ->  0
      if (cur1.polyInt() != 0) System.out.println("Error 5.10 with " + cur1.polyInt());


      // test addMod functions using the add and the mod function
      for (int i = 0; i < 100; i++) {
        curRnd1 = rnd.nextInt(1 << (2*maxLen));
        curRnd2 = rnd.nextInt(1 << (2*maxLen));
        curRnd3 = rnd.nextInt((1 << maxLen) - 1) + 1;
        setModulus(new Poly4Max(curRnd3));
        cur1.mod(new Poly4Max(curRnd1), new Poly4Max(curRnd3));
        cur2.mod(new Poly4Max(curRnd2), new Poly4Max(curRnd3));
        cur3.add(cur1,cur2);
        cur1.addMod(new Poly4Max(curRnd1), new Poly4Max(curRnd2));
        cur2 = new Poly4Max(curRnd1);
        cur2.addMod(new Poly4Max(curRnd2));
        if (!cur1.isEqual(cur3)) System.out.println("Error 6.1 with " + curRnd1 + " " + curRnd2 + " " + curRnd3);
        if (!cur1.isEqual(cur2)) System.out.println("Error 6.2 with " + curRnd1 + " " + curRnd2 + " " + curRnd3);
      }

      // test mulMod functions using the mul and the mod function
      for (int i = 0; i < 100; i++) {
        curRnd1 = rnd.nextInt(1 << maxLen);
        curRnd2 = rnd.nextInt(1 << maxLen);
        curRnd3 = rnd.nextInt((1 << maxLen) - 1) + 1;
        setModulus(new Poly4Max(curRnd3));
        cur1.mod(new Poly4Max(curRnd1), new Poly4Max(curRnd3));
        cur2.mod(new Poly4Max(curRnd2), new Poly4Max(curRnd3));
        cur3.mul(cur1,cur2);
        cur3.mod(new Poly4Max(curRnd3));
        cur1.mulMod(new Poly4Max(curRnd1), new Poly4Max(curRnd2));
        if (!cur1.isEqual(cur3)) System.out.println("Error 6.3 with " + curRnd1 + " " + curRnd2 + " " + curRnd3);
      }

      // test powMod by different values
      setModulus(new Poly4Max(4101));  // x^6 + x + 1
      cur1.powMod(new Poly4Max(4),1);  // x  ->  x
      if (cur1.polyInt() != 4) System.out.println("Error 6.4a with " + cur1.polyInt());
      cur1.powMod(new Poly4Max(4),4);  // x  ->  x^4
      if (cur1.polyInt() != 256) System.out.println("Error 6.4b with " + cur1.polyInt());
      cur1.powMod(new Poly4Max(4),6);  // x  ->  x + 1
      if (cur1.polyInt() != 5) System.out.println("Error 6.4c with " + cur1.polyInt());
      cur1.powMod(new Poly4Max(4),12);  // x  ->  x^2 + 1
      if (cur1.polyInt() != 17) System.out.println("Error 6.4d with " + cur1.polyInt());
      cur1.powMod(new Poly4Max(4),24);  // x  ->  x^4 + 1
      if (cur1.polyInt() != 257) System.out.println("Error 6.4e with " + cur1.polyInt());
      cur1.powMod(new Poly4Max(4),36);  // x  ->  x^4 + x^2 + x
      if (cur1.polyInt() != 276) System.out.println("Error 6.4f with " + cur1.polyInt());
      curRnd1 = rnd.nextInt(1 << (maxLen-1));
      cur1.powMod(new Poly4Max(curRnd1),0);
      if (cur1.polyInt() != 1) System.out.println("Error 6.5 with " + cur1.polyInt() + " " + curRnd1);
      cur1.powMod(new Poly4Max(5),3);  // x + 1
      if (cur1.polyInt() != 85) System.out.println("Error 6.6a with " + cur1.polyInt());

      // go on with a new modulus
      setModulus(new Poly4Max(Poly64Max.genPoly));
      cur1.powMod(new Poly4Max(new int[] {1,0}),4);
      if (!cur1.isEqual(new Poly4Max(new int[] {1,0}))) { System.out.println("Error 6.7a with "); cur1.writeString(); }
      cur1.powMod(new Poly4Max(new int[] {1,0}),3);
      if (!cur1.isEqual(new Poly4Max(new int[] {1,0}))) { System.out.println("Error 6.7b with "); cur1.writeString(); }
      cur1.powMod(new Poly4Max(new int[] {1,0}),2);
      if (!cur1.isEqual(new Poly4Max(new int[] {1,0}))) { System.out.println("Error 6.7c with "); cur1.writeString(); }
      cur1.powMod(new Poly4Max(new int[] {1,0}),1);
      if (!cur1.isEqual(new Poly4Max(new int[] {1,0}))) { System.out.println("Error 6.7d with "); cur1.writeString(); }

      cur1.pow(new Poly4Max(new int[] {1,0}),4);
      if (!cur1.isEqual(new Poly4Max(new int[] {1,0}))) { System.out.println("Error 6.8a with "); cur1.writeString(); }
      cur1.pow(new Poly4Max(new int[] {1,0}),3);
      if (!cur1.isEqual(new Poly4Max(new int[] {1,0}))) { System.out.println("Error 6.8b with "); cur1.writeString(); }
      cur1.pow(new Poly4Max(new int[] {1,0}),2);
      if (!cur1.isEqual(new Poly4Max(new int[] {1,0}))) { System.out.println("Error 6.8c with "); cur1.writeString(); }
      cur1.pow(new Poly4Max(new int[] {1,0}),1);
      if (!cur1.isEqual(new Poly4Max(new int[] {1,0}))) { System.out.println("Error 6.8d with "); cur1.writeString(); }
      cur1.pow(new Poly4Max(new int[] {1,1}),4);
      if (!cur1.isEqual(new Poly4Max(new int[] {1,4}))) { System.out.println("Error 6.8e with "); cur1.writeString(); }
      cur1.pow(new Poly4Max(new int[] {1,1}),3);
      if (!cur1.isEqual(new Poly4Max(new int[] {1,3}))) { System.out.println("Error 6.8f with "); cur1.writeString(); }
      cur1.pow(new Poly4Max(new int[] {1,1}),2);
      if (!cur1.isEqual(new Poly4Max(new int[] {1,2}))) { System.out.println("Error 6.8g with "); cur1.writeString(); }
      cur1.pow(new Poly4Max(new int[] {1,1}),1);
      if (!cur1.isEqual(new Poly4Max(new int[] {1,1}))) { System.out.println("Error 6.8h with "); cur1.writeString(); }

      System.out.println("...finished testing.");
    }

    static void plotFunction() {
      setModulus(new Poly4Max(94));  // x^6 + x + 1, 4101
                                     // x^3 + x^2 + 2x + 3, 91
                                     // x^3 + x^2 + 3x + 2, 94
      Poly4Max base = new Poly4Max(4);
      Poly4Max res = new Poly4Max();
      for (int i = 0; i < 64; i++) {
        res.powMod(base,i);
        System.out.print(i + ". "); res.writeString();
      }
    }

    static void rootFunction() {
      int counter = 0;
      setModulus(new Poly4Max(4101));    // x^6 + x + 1
      for (int i = 0; i < 4096; i++) {
        Poly4Max curRoot = new Poly4Max(i);
        Poly4Max power6 = new Poly4Max();
        power6.powMod(curRoot,6);
        power6.addMod(curRoot);
        power6.addMod(new Poly4Max(1));
        if (power6.polyInt() == 0) {
          curRoot.writeString();
          System.out.print("Square: ");
          power6.powMod(curRoot,2);
          power6.writeString();
          System.out.println();
          counter++;
        }
      }
      System.out.println("Roots: " + counter);
    }

    static void rootBeta() {
      int counter = 0;
      System.out.println("Equation b^2 + b + 1 =  b^4 + b^2 + 1 = 0");
      for (int i = 0; i < 64; i++) {
        setModulus(new Poly4Max(21));  // x^2 + x + 1, 21
        Poly4Max curRoot = new Poly4Max(i);
        Poly4Max power2 = new Poly4Max();
        power2.powMod(curRoot,2);
        power2.addMod(curRoot);
        power2.addMod(new Poly4Max(1));
        if (power2.isZero()) {
          setModulus(new Poly4Max(273));  // x^4 + x^2 + 1, 273
          Poly4Max power4 = new Poly4Max();
          power4.powMod(curRoot,4);
          power2.powMod(curRoot,2);
          power4.addMod(power2);
          power4.addMod(new Poly4Max(1));
          if (power4.isZero()) {
            curRoot.writeString();
          }
        }
      }
    }

    /** Tests if a given polynomial (specified as a number, see {@link #int2poly}) is irreducible.
        If so, prints it to stdout.
    */
    static void isIrr(int poly) {
      Poly4Max testPoly = new Poly4Max(poly);
      //System.out.print("Try to find a polynomial which reduces ");
      //testPoly.writeString();
      for (int i = 4; i < (1 << (testPoly.degree()*2)); i++) {
        Poly4Max modPoly = new Poly4Max();
        Poly4Max divBy = new Poly4Max(i);
        modPoly.mod(testPoly,divBy);
        if (modPoly.isZero()) {
          //System.out.print("Polynomial reducible with ");
          //modPoly.writeString();
          return;
        }
      }
      System.out.print("Polynomial irreducible "); testPoly.writeString();
    }

    /** Find all irreducible polynomials.
    */
    static void findIrr() {
      for (int i = 4*16; i < 255; i++) {
        isIrr(i);
      }
    }

    /** Mainly invoces the {@link #testIt} method.
    */
    public static void main(String[] args) {
        testIt();
        //findIrr();
        //rootFunction();
        //rootBeta();
        //isIrr(91);
        //isIrr(94);
        //plotFunction();
        //Poly4Max mod = new Poly4Max();
        //mod.mod(new Poly4Max(4101), new Poly4Max(91));
        //mod.writeString();
    }
}
