package HFE;

import HFE.*;
import java.util.*;

//-----------------------------------------------------------------------------
// Project:     Implementing HFE at UCC as Project CS4095
// Supervisors: Simon Foley (Computer Science) and Patrick Fitzpatrick (Maths)
//-----------------------------------------------------------------------------

/**
<p>
This class encapsulates the operations for polynomials
over the finite field F_64 which is an extension over
F_4. The generation polynomial is f(x) = x^3 + x^2 + 2x + 3
<br>
The polynomials are bounded by {@link #maxLen}.
<p>
@author Christopher Wolf (hfe@Christopher-Wolf.de)
@version 0.1
*/

//Copyright (C) 2002    Christopher Wolf
//
//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License (file LICENCE) for more details.


class Poly64Max {

    /** Maximal length of the polynomials
      @serial constant
    */
    final static int maxLen = 21;
    //final static int maxLen = 15;

    /** Generating Polynomial over {@link Field4}.
    */
    final static Poly4Max genPoly = new Poly4Max(new int [] {1,3, 1,2, 2,1, 3,0}); // generating polynomial x^3 + x^2 + 2x + 3 -> 91

    /** Basicly the degree of the generation polynomial
        de facto the dimension of the vector space which can
        be assigned to F_64 in terms of F_4.
    */
    final static int extension = 3;

    /** Modulus for the mod version of the operations
      @serial constant
    */
    static Poly64Max modulus;

    /** Current value of the polynomial.
        The polynomial has at most degree (maxLen-1). Its coefficients are
        taken from F_64. The element value[i] is the coefficient for x^i so
        value[0] is the constant term for this polynomial.
    */
    int value[];

    /** sets all the coefficients to zero
    */
    Poly64Max() {
      setZero();
    }

    /** Creates a new polynomial. {@link #int2poly}
    */
    Poly64Max(int initValue) {
      setZero();
      int2poly(initValue);
    }

    /** Creates a copy from the given polynomial
    */
    Poly64Max(final Poly64Max father) {
      value = (int[])father.value.clone();
    }

    /** Takes one polynomial and shifts it by the given shift.
        It basicly leaves out shift number of coefficients
        and starts copying at value[shift].
        Makes a copy of this polynomial.
    */
    // !! Exception
    Poly64Max(final Poly64Max father, int shift) {
      setZero();
      Check.assert((father.degree()+shift) < maxLen);
      for (int i = 0; i <= father.degree(); i++) {
        value[shift+i] = father.value[i];
      }
    }


    /** {@link #array2poly}
    */
    // !! Exception
    Poly64Max(final int[] inArray) {
      array2poly(inArray);
    }

    /** Generates a polynomial with random coefficients.
        The degree of this polynomial can be expected
        to be (maxLen-1).
    */
    Poly64Max(Random rnd) {
      setZero();
      for (int i = 0; i < maxLen; i++) {
        value[i] = rnd.nextInt(64);
      }
    }


    /** Sets the private field modulus.
        Does not clone the moduls but storing it. Checks if the degree
        of the moduls is less than (maxLen/2).
    */
    // !! Exception
    static void setModulus(final Poly64Max newModulus) {
      // security check: is the degree of modulus small enough?
      if ((newModulus.degree()*2) > maxLen) {
        System.out.println("Poly64Max.setModulus: modulus too big");
      }
      modulus = newModulus;
    }


    /** Returns the degree of the object.
        The degree of a constant polynomial is 0, the degree of the zero
        polynomial is -1.
    */
    int degree() {
      for (int i = maxLen-1; i >= 0; i--)
        if (value[i] != 0) return i;
      return -1;
    }

    /** Compares the current polynomial with the given one.
      @return true iff (this < whom)
    */
    boolean isLess(final Poly64Max whom) {
      for (int i = maxLen-1; i >= 0; i--) {
        if (value[i] < whom.value[i]) return true;
        if (value[i] > whom.value[i]) return false;
      }
      return false;
    }

    /** Compares the current polynomial with the given one.
      @return true iff (this > whom)
    */
    boolean isBigger(final Poly64Max whom) {
      for (int i = maxLen-1; i >= 0; i--) {
        if (value[i] < whom.value[i]) return false;
        if (value[i] > whom.value[i]) return true;
      }
      return false;
    }

    /** Compares the current polynomial with the given one.
      @return true iff (this == whom) for all coefficients
    */
    boolean isEqual(final Poly64Max whom) {
      for (int i = maxLen-1; i >= 0; i--) {
        if (value[i] != whom.value[i]) return false;
      }
      return true;
    }

    /** Checks if a polynomial is the zero polynomial.
      @return true iff all coefficients are zero
    */
    boolean isZero() {
      for (int i = 0; i < maxLen; i++)
        if (value[i] != 0) return false;
      return true;
    }

    /** Sets all the coefficients to zero by allocating new memory
    */
    // !! Version without new memory
    void setZero() {
      value = new int[maxLen];
    }

    /** Takes one polynomial and shifts it by the given shift.
        It basicly leaves out shift number of coefficients
        and starts copying at value[shift].
        Makes a copy of this polynomial.
    */
    void copyShift(final Poly64Max father, int shift) {
      Check.assert((father.degree()+shift) < maxLen);
      Check.assert(shift >= 0, "shift is too small, namely " + shift);
      // first fill everything with 0
      for (int i = 0; i < maxLen; i++)
        value[i] = 0;
      // now copy everything into the value array
      for (int i = 0; i <= father.degree(); i++)
        value[shift+i] = father.value[i];
    }

    /** Sets the coefficients according to intValue.
        The bits 0..3 are the coefficient for
        x^0, bits 4..7 are the coefficient for x^1, ...
        So this function can only set the least significant
        7 coeficients.
    */
    void int2poly(int intValue) {
      for (int i = 0; i < maxLen; i++) {
        value[i] =  (intValue & 63);
        intValue >>= 6;
        if (intValue < 0) intValue = 0;
      }
    }

    /** Sets the coefficients using the given coefficient / power array.
        The elements in this array have the form {c_0,p_0, c_1,p_1, ...}
        where c_i is the coefficient for x^(p_i). If there are doublicate
        values for p_i, the last one is taken.
    */
    // !! Exception for wrong array length, power to big
    void array2poly(final int[] inArray) {
      setZero();  // first overwrite the current values
      if (inArray.length % 2 == 1) {
        System.out.println("Poly64Max: wrong array length.");
        return;
      }
      for (int i = 0; i < inArray.length; i+=2)
        value[inArray[i+1]] = inArray[i];
    }

    /** adds two numbers using Poly4Max
        as representation of the numbers.
    */
    static int addNumByPoly(int ter1, int ter2) {
      Poly4Max poly;
      poly = new Poly4Max(ter1);
      poly.add(new Poly4Max(ter2));
      return poly.polyInt();
    }

    /** Adds two numbers using xor
    */
    static int addNum(int ter1, int ter2) {
      return ter1 ^ ter2;
    }


    /** Multiplies two numbers using Poly4Max
        as representation of the numbers.
    */
    static int mulNumByPoly(int fac1, int fac2) {
      Poly4Max res = new Poly4Max();
      res.mul(new Poly4Max(fac1), new Poly4Max(fac2));
      res.mod(genPoly);
      return res.polyInt();
    }

    /** Produce the multiplication table
        by using mulNumByPoly. The table is writen to
        stdout as valid java statement and can be directed to a file.
    */
    static void writeMulTab() {
      System.out.println("static final int[][] mulTab = {");
      for (int i = 0; i < 64; i++) {
        System.out.print("{");
        int cur = 1;
        for (int j = 0; j < 64; j++) {
          if ((cur % 20) == 0) System.out.println();
          System.out.print(mulNumByPoly(i,j));
          if (j != 63) System.out.print(", ");
          cur++;
        }
        System.out.println("},");
      }
      System.out.println("};");
    }

    /** multiplies two numbers
       by looking up the result
       in the multiplication table which is provided
       by the class {@link Field64_Tab}    
    */
    static int mulNum(int fac1, int fac2) {
      return Field64_Tab.mulTab[fac1][fac2];
    }

    /** Produce the table with the multiplicative inverses
        of every element by using mulNumByPoly. The table is writen to
        stdout as valid java statement and can be directed to a file.
        zero is set to zero.
    */
    static void writeInverseTab() {
      System.out.println("static final int[] inverseTab = {");
      System.out.print("0, ");
      for (int i = 1; i < 64; i++) {
        // find out the multiplicative inverse by brute force
        int j = 1;
        while (mulNum(i,j) != 1) {
          j++;
          if (j > 63) {
            System.out.println("No inverse for " + i);
            return;
          }
        }
        System.out.print(j);
        if (i != 63) System.out.print(", ");
        if ((i % 20) == 0) System.out.println();
      }
      System.out.println("};");
    }


    /** Produce the division table
        by using mulNumByPoly and {@link Field64_Tab#inverseTab}.
        The table is writen to
        stdout as valid java statement and can be directed to a file.
        zero is set to zero.
    */
    static void writeDivTab() {
      System.out.println("static final int[][] divTab = {");
      for (int i = 0; i < 64; i++) {
        System.out.print("{");
        int cur = 1;
        for (int j = 0; j < 64; j++) {
          if ((cur % 20) == 0) System.out.println();
          System.out.print(mulNum(i,Field64_Tab.inverseTab[j]));
          if (j != 63) System.out.print(", ");
          cur++;
        }
        System.out.println("},");
      }
      System.out.println("};");
    }

    /** divides whom by the number by.
        uUses the division table provided
        by {@link Field64_Tab}.
      @return (whom/by) in F_64
    */
    static int divNum(int whom, int by) {
      return Field64_Tab.divTab[whom][by];
    }

    /** multiplies all coefficients by the given factor in F_64.
    */
    void mulFactor(int factor) {
      for (int i = 0; i < maxLen; i++)
        value[i] = mulNum(factor,value[i]);
    }

    /** calculates this = wen1 + wen2
        as polynomial addition
    */
    void add(final Poly64Max a, final Poly64Max b) {
      for (int i = 0; i < maxLen; i++)
        value[i] = addNum(a.value[i],b.value[i]);
    }

    /** calculate this += whom
        as polynomial addition
    */
    void add(final Poly64Max whom) {
      for (int i = 0; i < maxLen; i++)
        value[i] = addNum(whom.value[i],value[i]);
    }

    /** calculates this = a * b
        as polynomial addition.
        Does <b>not</b> check if the result is of degree < maxLen
    */
    void mul(final Poly64Max a, final Poly64Max b) {
      // first reset this
      setZero();
      for (int i = 0; i < maxLen; i++) {
        // only do something if it is necessary
        if (a.value[i] != 0) {
          for (int j = 0; j+i < maxLen; j++) {
            if (b.value[j] != 0) {
              value[i+j] = addNum(value[i+j],mulNum(a.value[i],b.value[j]));  // add at the correct place
            }
          }
        }
      }
    }

    /** calculates this = base^by.
        Does <b>not</br> check if the result is of degree < maxLen
    */
    void pow(final Poly64Max base, int by) {
      if (by == 0) {
        int2poly(1);
        return;
      }
      if ((by % 2) == 1) this.value = (int[])base.value.clone(); else int2poly(1);
      by /= 2;
      Poly64Max square = new Poly64Max(base);
      Poly64Max tmp = new Poly64Max();
      do {
        tmp.mul(square,square);
        square.value = (int[])tmp.value.clone();
        if ((by % 2) == 1) {
          tmp.mul(square,this);
          this.value = (int[])tmp.value.clone();
        }
        by /= 2;
      } while (by != 0);
    }

    /** Calculates this %= moduls.
        Uses the static field modulus.
    */
    void mod () {
      mod(modulus);
    }


    /** Calculates this = a % b
    */
    void mod (final Poly64Max a, final Poly64Max b) {
      // copy before doing anything
      value = (int[])a.value.clone();
      mod(b);
    }

    // static variable for our mod method
    static Poly64Max mod_additive = new Poly64Max();
    /** Calculates this %= b
    */
    void mod (final Poly64Max b) {
      // do not divide by zero
      Check.assert(!b.isZero());

      // work so long as there is something to do
      while (b.degree() <= this.degree()) {
        mod_additive.copyShift(b,this.degree()-b.degree());
        mod_additive.mulFactor(divNum(value[degree()], b.value[b.degree()]));
        add(mod_additive);
      }
    }

    /** Computes this = (a+b)%modul.
        Assumes a<modul, b<modul. Does <b>not</b> check this condition.
    */
    void addMod(final Poly64Max a, final Poly64Max b) {
      add(a,b);
      mod();
    }

    // computes this = (this+b)%m
    void addMod(final Poly64Max b) {
      add(b);
      mod();
    }


    /** Computes this = (this+b)%modul.
        Assumes this<modul, b<modul. Does <b>not</b> check this condition.
    */
    void mulMod(final Poly64Max a, final Poly64Max b) {
      mul(a,b);
      mod();
    }

    /** Computes this = (base^by)%modul.
        Assumes base<modul. Does <b>not</b> check this condition.
    */
    void powMod(final Poly64Max base, int by) {
      if (by == 0) {
        int2poly(1);
        return;
      }
      if ((by % 2) == 1) this.value = (int[])base.value.clone(); else int2poly(1);
      by /= 2;
      Poly64Max square = new Poly64Max(base);
      Poly64Max tmp = new Poly64Max();
      do {
        tmp.mulMod(square,square);
        square.value = (int[])tmp.value.clone();
        if ((by % 2) == 1) {
          tmp.mulMod(square,this);
          this.value = (int[])tmp.value.clone();
        }
        by /= 2;
      } while (by != 0);
    }

    /** Computes (num^by) in F_64.
    */
    static int powNum(int num, int by) {
      if (by == 0) {
        return 1;
      }
      int res; // result of this function
      if ((by % 2) == 1) res = num; else res = 1;
      by /= 2;
      int square = num;
      do {
        mulNum(square,square);
        square = mulNum(square,square);
        if ((by % 2) == 1)
          res = mulNum(square,res);
        by /= 2;
      } while (by != 0);
      return res;
    }

    /** Uses the given polynomial as a function.
        Applies the value num (from F_64) to
        the polynomial.
       @return f(num)
    */
    int applyPoly(int num) {
      int res = value[0];  // result
      int cur = 1;
      for (int i = 1; i < maxLen; i++) {
        cur = mulNum(cur,num);
        res = addNum(res, mulNum(value[i],cur));
      }
      return res;
    }

    /** Writes the polynomial as a string to stdout
    */
    void writeString() {
      for (int i = maxLen-1; i > 0; i--) {
        if (value[i] != 0)
          System.out.print(value[i] + "x^" + i + " + ");
      }
      System.out.print(value[0]);
      System.out.println(" (" + polyInt() + ")");
    }


    /** Returns the polynomial as an integer.
       Bits 0..3 are the coefficient for x^0,
       Bits 4..7 are the coefficients for x^1, ...
       The result is only valid if maxLen < 7
    */
    int polyInt() {
      int power = 1;
      int res = 0;
      for (int i = 0; i < maxLen; i++) {
        res += value[i]*power;
        power *= 64;
      }
      return res;
    }

    /** Tests the functionality for the whole class.
        Reports errors to stdout.
        <i>Works only for maxLen == 15</i>
    */
    static void testIt () {
      System.out.println("start testing...");
      Random rnd = new Random();
      Poly64Max cur = new Poly64Max(6), cur1 = new Poly64Max(), cur2 = new Poly64Max(),
                cur3 = new Poly64Max(), cur4 = new Poly64Max(), cur5 = new Poly64Max(), cur6 = new Poly64Max();
      int curRnd, curRnd1, curRnd2, curRnd3, curRes;

      // first test the constructors and the I/O functions
      Poly64Max first = new Poly64Max();
      Poly64Max second = new Poly64Max(3);
      Poly64Max third = new Poly64Max(second);

      if (second.polyInt() != 3) System.out.println("Error 1.1");
      if (third.polyInt() != 3) System.out.println("Error 1.2");
      second = new Poly64Max(rnd);
      //second.writeString();

      cur = new Poly64Max(130);
      if (130 != cur.polyInt()) System.out.println("Error 1.3 with " + cur.polyInt());
      cur = new Poly64Max(1024);
      if (1024 != cur.polyInt()) System.out.println("Error 1.4 with " + cur.polyInt());
      for (int i = 0; i < 100; i++) {
        curRnd = rnd.nextInt(1 << 30);  // gives values in 0..2^30-1
        cur = new Poly64Max(curRnd);
        cur2 = new Poly64Max(cur);
        if (curRnd != cur.polyInt()) System.out.println("Error 1.5 with " + curRnd);
        if (curRnd != cur2.polyInt()) System.out.println("Error 1.6 with " + curRnd);
      }
      int [] zwArr = {1,0, 2,1};
      cur = new Poly64Max(new int [] {1,0, 2,1});
      if (129 != cur.polyInt()) System.out.println("Error 1.7 with " + cur.polyInt());
      cur.array2poly(new int [] {1,1, 2,2, 3,3});
      if (794688 != cur.polyInt()) System.out.println("Error 1.8 with " + cur.polyInt());
      // test isBigger, isLess, isEqual
      for (int i = 0; i < 100; i++) {
        curRnd1 = rnd.nextInt(1 << 30);
        curRnd2 = rnd.nextInt(1 << 30);
        cur1 = new Poly64Max(curRnd1);
        cur2 = new Poly64Max(curRnd2);

        // we have now three different cases:
        if (curRnd1 < curRnd2) {
          if (!cur1.isLess(cur2)) System.out.println("Error 2.1a with " + curRnd1 + " " + curRnd2);
          if (cur1.isBigger(cur2)) System.out.println("Error 2.1b with " + curRnd1 + " " + curRnd2);
          if (cur1.isEqual(cur2)) System.out.println("Error 2.1c with " + curRnd1 + " " + curRnd2);
          if (!cur1.isEqual(cur1)) System.out.println("Error 2.1d with " + curRnd1 + " " + curRnd2);
          if (cur2.isZero()) System.out.println("Error 2.1e with " + curRnd1 + " " + curRnd2);
        } else {
          if (curRnd1 > curRnd2) {
            if (cur1.isLess(cur2)) System.out.println("Error 2.2a with " + curRnd1 + " " + curRnd2);
            if (!cur1.isBigger(cur2)) System.out.println("Error 2.2b with " + curRnd1 + " " + curRnd2);
            if (cur1.isEqual(cur2)) System.out.println("Error 2.2c with " + curRnd1 + " " + curRnd2);
            if (!cur2.isEqual(cur2)) System.out.println("Error 2.2d with " + curRnd1 + " " + curRnd2);
            if (cur1.isZero()) System.out.println("Error 2.2e with " + curRnd1 + " " + curRnd2);
          } else {
            if (cur1.isLess(cur2)) System.out.println("Error 2.3a with " + curRnd1 + " " + curRnd2);
            if (cur1.isBigger(cur2)) System.out.println("Error 2.3b with " + curRnd1 + " " + curRnd2);
            if (!cur1.isEqual(cur2)) System.out.println("Error 2.3c with " + curRnd1 + " " + curRnd2);
            if (!cur2.isEqual(cur1)) System.out.println("Error 2.3d with " + curRnd1 + " " + curRnd2);
          }
        }
      }

      // test if the "isZero" parameter works
      cur = new Poly64Max(0);
      if (!cur.isZero()) System.out.println("Error 2.4");
      cur = new Poly64Max(23);
      cur.setZero();
      if (!cur.isZero()) System.out.println("Error 2.5");

      //Test addition
      cur1 = new Poly64Max();
      cur1.add(new Poly64Max(1 << 30), new Poly64Max(3));
      if (cur1.polyInt() != ((1 << 30)+3)) System.out.println("Error 3.1 with " + cur1.polyInt());
      for (int i = 0; i < 100; i++) {
        curRnd1 = rnd.nextInt(1 << 30);
        curRnd2 = rnd.nextInt(1 << 30);
        curRes = curRnd1 ^ curRnd2;
        cur1 = new Poly64Max();
        cur1.add(new Poly64Max(curRnd1), new Poly64Max(curRnd2));
        cur2 = new Poly64Max(curRnd1);
        cur2.add(new Poly64Max(curRnd2));
        if (!cur1.isEqual(cur2)) System.out.println("Error 3.2 with " + curRnd1 + " " + curRnd2);
        if (cur1.polyInt() != curRes) System.out.println("Error 3.3 with " + curRnd1 + " " + curRnd2);
      }
      // test the both addition methods
      for (int i = 0; i < 64; i++)
        for (int j = 0; j < 64; j++) {
          if (addNum(i,j) != addNumByPoly(i,j)) System.out.println("Error 3.4 with " + i + " " + j);
        }

      // Test multiplication
      cur = new Poly64Max();
      cur.mul(new Poly64Max(2), new Poly64Max(5));
      if (cur.polyInt() != 10) System.out.println("Error 4.1 with " + cur.polyInt());
      cur.mul(new Poly64Max(3), new Poly64Max(3));
      if (cur.polyInt() != 2) System.out.println("Error 4.2 with " + cur.polyInt());
      cur.mul(new Poly64Max(10), new Poly64Max(21));
      if (cur.polyInt() != 47) System.out.println("Error 4.3 with " + cur.polyInt());
      //Test both multiplication methods
      for (int i = 0; i < 64; i++) {
        for (int j = 0; j < 64; j++) {
          if (mulNum(i,j) != mulNumByPoly(i,j)) System.out.println("Error 4.4 with " + i + " " + j);
        }
      }

      // Test both addition and multiplication
      // using the distributive law a(b+c)=ab+ac
      for (int i = 0; i < 100; i++) {
        cur1 = new Poly64Max(rnd);
        cur2 = new Poly64Max(rnd);
        cur3 = new Poly64Max(rnd);

        // left hand side
        cur4.add(cur2,cur3);
        cur6.mul(cur1,cur4);
        // right hand side
        cur4.mul(cur1,cur2);
        cur5.mul(cur1,cur3);
        cur5.add(cur4);

        if (!cur6.isEqual(cur5)) System.out.println("Error 4.5");
      }
      // test the inverse table
      for (int i = 1; i < 64; i++)
        if (mulNum(i,Field64_Tab.inverseTab[i]) != 1) System.out.println("Error 4.6 with " + i);
      // test the division table
      for (int i = 1; i < 64; i++)
        for (int j = 1; j < 64; j++) {
          if (divNum(mulNum(i,j),j) != i) System.out.println("Error 4.7a with " + i + " " + j);
          if (divNum(mulNum(i,j),i) != j) System.out.println("Error 4.7b with " + i + " " + j);
        }

      // test the shift constructor
      cur1 = new Poly64Max(new Poly64Max(3),2);
      if (cur1.polyInt() != (3 << 12)) System.out.println("Error 4.8a");
      cur1 = new Poly64Max(new Poly64Max(1 << 28),0);
      if (cur1.polyInt() != (1 << 28)) System.out.println("Error 4.8b");
      cur1 = new Poly64Max(new Poly64Max(3),4);
      if (cur1.polyInt() != (3 << 24)) System.out.println("Error 4.8c");

      // Test multiplication together with the shift constructor
      for (int i = 0; i < 100; i++) {
        curRnd1 = rnd.nextInt(5)+1;
        curRnd2 = rnd.nextInt(63 << (5/curRnd1));
        cur1 = new Poly64Max(curRnd2);
        cur2.mul(cur1,new Poly64Max(1 << (curRnd1*6)));
        cur3 = new Poly64Max(cur1,curRnd1);

        if (!cur3.isEqual(cur2)) System.out.println("Error 4.9 with " + curRnd1 + " " + curRnd2);
      }

      // Test power function
      cur1.pow(new Poly64Max(new int[] {1,1, 1,0}),2);  // x + 1  ->  x^2 + 1
      if (!cur1.isEqual(new Poly64Max(new int[] {1,2, 1,0}))) System.out.println("Error 4.9a with " + cur1.polyInt());
      cur1.pow(new Poly64Max(new int[] {1,1, 1,0}),3);  // x + 1  ->  x^2 + 1
      if (!cur1.isEqual(new Poly64Max(new int[] {1,3, 1,2, 1,1, 1,0}))) System.out.println("Error 4.9b with " + cur1.polyInt());

      // test mod function
      cur1.mod(new Poly64Max(1), new Poly64Max(1)); // 1  %  1  -> 0
      if (cur1.polyInt() != 0) System.out.println("Error 5.1 with " + cur1.polyInt());
      cur1.mod(new Poly64Max(new int[] {2,3}), new Poly64Max(new int[] {2,3, 3,0})); // 2x^3  %  2x^3 + 3  ->  3
      if (cur1.polyInt() != 3) System.out.println("Error 5.2 with " + cur1.polyInt());
      cur1.mod(new Poly64Max(new int[] {3,0}), new Poly64Max(new int[] {7,0})); // 3  %  7  ->  0
      if (cur1.polyInt() != 0) System.out.println("Error 5.3 with " + cur1.polyInt());
      cur1.mod(new Poly64Max(new int[] {7,2}), new Poly64Max(new int[] {3,1})); // 7x^2  %  3x  ->  0
      if (cur1.polyInt() != 0) System.out.println("Error 5.4 with " + cur1.polyInt());

      // test addMod functions using the add and the mod function
      for (int i = 0; i < 100; i++) {
        curRnd1 = rnd.nextInt(1 << 30);
        curRnd2 = rnd.nextInt(1 << 30);
        curRnd3 = rnd.nextInt(1 << 30) + 1;
        setModulus(new Poly64Max(curRnd3));
        cur1.mod(new Poly64Max(curRnd1), new Poly64Max(curRnd3));
        cur2.mod(new Poly64Max(curRnd2), new Poly64Max(curRnd3));
        cur3.add(cur1,cur2);
        cur1.addMod(new Poly64Max(curRnd1), new Poly64Max(curRnd2));
        cur2 = new Poly64Max(curRnd1);
        cur2.addMod(new Poly64Max(curRnd2));
        if (!cur1.isEqual(cur3)) System.out.println("Error 6.1 with " + curRnd1 + " " + curRnd2 + " " + curRnd3);
        if (!cur1.isEqual(cur2)) System.out.println("Error 6.2 with " + curRnd1 + " " + curRnd2 + " " + curRnd3);
      }

      // test mulMod functions using the mul and the mod function
      for (int i = 0; i < 100; i++) {
        curRnd1 = rnd.nextInt(1 << maxLen);
        curRnd2 = rnd.nextInt(1 << maxLen);
        curRnd3 = rnd.nextInt((1 << maxLen) - 1) + 1;
        setModulus(new Poly64Max(curRnd3));
        cur1.mod(new Poly64Max(curRnd1), new Poly64Max(curRnd3));
        cur2.mod(new Poly64Max(curRnd2), new Poly64Max(curRnd3));
        cur3.mul(cur1,cur2);
        cur3.mod(new Poly64Max(curRnd3));
        cur1.mulMod(new Poly64Max(curRnd1), new Poly64Max(curRnd2));
        if (!cur1.isEqual(cur3)) System.out.println("Error 6.3 with " + curRnd1 + " " + curRnd2 + " " + curRnd3);
      }

      // test powMod by different values
      setModulus(new Poly64Max(new int[] {1,6, 1,1, 1,0}));  // x^6 + x + 1
      cur1.powMod(new Poly64Max(new int[] {1,1}),1);  // x  ->  x
      if (!cur1.isEqual(new Poly64Max(new int[] {1,1}))) System.out.println("Error 6.4a with " + cur1.polyInt());
      cur1.powMod(new Poly64Max(new int[] {1,1}),4);  // x  ->  x^4
      if (!cur1.isEqual(new Poly64Max(new int[] {1,4}))) System.out.println("Error 6.4b with " + cur1.polyInt());
      cur1.powMod(new Poly64Max(new int[] {1,1}),6);  //  x  ->  x + 1
      if (!cur1.isEqual(new Poly64Max(new int[] {1,1, 1,0}))) System.out.println("Error 6.4c with " + cur1.polyInt());
      cur1.powMod(new Poly64Max(new int[] {1,1}),12);  //  x  -> x^2 + 1
      if (!cur1.isEqual(new Poly64Max(new int[] {1,2, 1,0}))) System.out.println("Error 6.4d with " + cur1.polyInt());
      cur1.powMod(new Poly64Max(new int[] {1,1}),24);  //  x  -> x^4 + 1
      if (!cur1.isEqual(new Poly64Max(new int[] {1,4, 1,0}))) System.out.println("Error 6.4e with " + cur1.polyInt());
      cur1.powMod(new Poly64Max(new int[] {1,1}),36);  //  x  -> x^4 + x^2 + x
      if (!cur1.isEqual(new Poly64Max(new int[] {1,4, 1,2, 1,1}))) System.out.println("Error 6.4f with " + cur1.polyInt());
      cur1.powMod(new Poly64Max(new int[] {4,1}),0);
      if (!cur1.isEqual(new Poly64Max(new int[] {1,0}))) System.out.println("Error 6.4g with " + cur1.polyInt());

      // test powNum function
      if (powNum(4,0) != 1) System.out.println("Error 7.1a");
      if (powNum(4,2) != mulNum(4,4)) System.out.println("Error 7.1b");
      if (powNum(4,3) != mulNum(4,mulNum(4,4))) System.out.println("Error 7.1c");
      // ... using random values
      for (int i = 0; i < 100; i++) {
        curRnd1 = rnd.nextInt(65536);
        curRnd2 = rnd.nextInt(65536);
        curRnd3 = rnd.nextInt(63)+1;

        if (mulNum(powNum(curRnd3,curRnd1),powNum(curRnd3,curRnd2)) != powNum(curRnd3,curRnd1+curRnd2)) System.out.println("Error 7.2a");
        if (powNum(curRnd3,curRnd1%63) != powNum(curRnd3,curRnd1)) System.out.println("Error 7.2b");
      }

      // test applyPoly
      cur = new Poly64Max(new int[] {5,0});
      if (cur.applyPoly(2) != 5) System.out.println("Error 7.3a");
      cur = new Poly64Max(new int[] {1,1});
      if (cur.applyPoly(2) != 2) System.out.println("Error 7.3b");
      cur = new Poly64Max(new int[] {1,1, 2,0});
      if (cur.applyPoly(2) != 0) System.out.println("Error 7.3c");
      cur = new Poly64Max(new int[] {1,2, 3,0});
      if (cur.applyPoly(2) != 0) System.out.println("Error 7.3d");

      // test copyShift
      cur = new Poly64Max(new int[] {5,0});
      cur1 = new Poly64Max();
      cur2 = new Poly64Max(cur,2);  // use shift constructor
      cur1.copyShift(cur,2);
      if (!cur1.isEqual(cur2)) System.out.println("Error 8.1");
      cur = new Poly64Max(new int[] {0,0});
      cur1 = new Poly64Max();
      cur2 = new Poly64Max(cur,0);  // use shift constructor
      cur1.copyShift(cur,0);
      if (!cur1.isEqual(cur2)) System.out.println("Error 8.2");
      cur = new Poly64Max(new int[] {5,9});
      cur1 = new Poly64Max();
      cur2 = new Poly64Max(cur,5);  // use shift constructor
      cur1.copyShift(cur,5);
      if (!cur1.isEqual(cur2)) System.out.println("Error 8.3");
      cur = new Poly64Max(new int[] {1,0});
      cur1 = new Poly64Max();
      cur2 = new Poly64Max(cur,0);  // use shift constructor
      cur1.copyShift(cur,0);
      if (!cur1.isEqual(cur2)) System.out.println("Error 8.4");
      cur = new Poly64Max(new int[] {3,1, 4,0});
      cur1 = new Poly64Max();
      cur2 = new Poly64Max(cur,0);  // use shift constructor
      cur1.copyShift(cur,0);
      if (!cur1.isEqual(cur2)) System.out.println("Error 8.5");

      // test the shift method with the same values as the shift constructor
      cur1.copyShift(new Poly64Max(3),2);
      if (cur1.polyInt() != (3 << 12)) System.out.println("Error 8.6");
      cur1.copyShift(new Poly64Max(1 << 28),0);
      if (cur1.polyInt() != (1 << 28)) System.out.println("Error 8.7");
      cur1.copyShift(new Poly64Max(3),4);
      if (cur1.polyInt() != (3 << 24)) System.out.println("Error 8.8");

      // Test multiplication together with shift method
      for (int i = 0; i < 100; i++) {
        curRnd1 = rnd.nextInt(5)+1;
        curRnd2 = rnd.nextInt(63 << (5/curRnd1));
        cur1 = new Poly64Max(curRnd2);
        cur2.mul(cur1,new Poly64Max(1 << (curRnd1*6)));
        cur3.copyShift(cur1,curRnd1);

        if (!cur3.isEqual(cur2)) System.out.println("Error 8.9 with " + curRnd1 + " " + curRnd2);
      }

      System.out.println("...finished testing.");
    }

    static void plotFunction() {
      setModulus(new Poly64Max(94));  // x^6 + x + 1, 4101
                                     // x^3 + x^2 + 2x + 3, 91
                                     // x^3 + x^2 + 3x + 2, 94
      Poly64Max base = new Poly64Max(4);
      Poly64Max res = new Poly64Max();
      for (int i = 0; i < 64; i++) {
        res.powMod(base,i);
        System.out.print(i + ". "); res.writeString();
      }
    }

    static void rootFunction() {
      int counter = 0;
      setModulus(new Poly64Max(4101));    // x^6 + x + 1
      for (int i = 0; i < 4096; i++) {
        Poly64Max curRoot = new Poly64Max(i);
        Poly64Max power6 = new Poly64Max();
        power6.powMod(curRoot,6);
        power6.addMod(curRoot);
        power6.addMod(new Poly64Max(1));
        if (power6.polyInt() == 0) {
          curRoot.writeString();
          System.out.print("Square: ");
          power6.powMod(curRoot,2);
          power6.writeString();
          System.out.println();
          counter++;
        }
      }
      System.out.println("Roots: " + counter);
    }

    static void rootBeta() {
      int counter = 0;
      System.out.println("Equation b^2 + b + 1 =  b^4 + b^2 + 1 = 0");
      for (int i = 0; i < 64; i++) {
        setModulus(new Poly64Max(21));  // x^2 + x + 1, 21
        Poly64Max curRoot = new Poly64Max(i);
        Poly64Max power2 = new Poly64Max();
        power2.powMod(curRoot,2);
        power2.addMod(curRoot);
        power2.addMod(new Poly64Max(1));
        if (power2.isZero()) {
          setModulus(new Poly64Max(273));  // x^4 + x^2 + 1, 273
          Poly64Max power4 = new Poly64Max();
          power4.powMod(curRoot,4);
          power2.powMod(curRoot,2);
          power4.addMod(power2);
          power4.addMod(new Poly64Max(1));
          if (power4.isZero()) {
            curRoot.writeString();
          }
        }
      }
    }


    /** Tests if a given polynomial (specified as a number, see {@link #int2poly}) is irreducible.
        If so, prints it to stdout.
    */
    static void isIrr(int poly) {
      Poly64Max testPoly = new Poly64Max(poly);
      System.out.print("Try to find a polynomial which reduces ");
      testPoly.writeString();
      for (int i = 4; i < (1 << (testPoly.degree()*2)); i++) {
        Poly64Max modPoly = new Poly64Max();
        Poly64Max divBy = new Poly64Max(i);
        modPoly.mod(testPoly,divBy);
        if (modPoly.isZero()) {
          System.out.print("Polynomial reducible with ");
          modPoly.writeString();
          return;
        }
      }
      System.out.println("Polynomial irreducible");
    }

    /** Mainly invoces the {@link #testIt} method.
    */
    public static void main(String[] args) {
        testIt();
        //writeDivTab();
        //rootFunction();
        //rootBeta();
        //isIrr(91);
        //isIrr(94);
        //plotFunction();
        //Poly64Max mod = new Poly64Max();
        //mod.mod(new Poly64Max(4101), new Poly64Max(91));
        //mod.writeString();
    }
}
