package HFE;

import java.util.*;
import java.lang.Math;
import HFE.*;

//-----------------------------------------------------------------------------
// Project:     Implementing HFE at UCC as Project CS4095
// Supervisors: Simon Foley (Computer Science) and Patrick Fitzpatrick (Maths)
//-----------------------------------------------------------------------------

/**
<p>
Transform4 provides a affine transformation (i.e. S,T in terms of HFE),
i.e. computes y = M*x + m where M and m are the affine transformation
and x/y are input/output.
All elements in the underlaying matrix and vector are taken from
{@link Field4}.
<p>
@author Christopher Wolf (hfe@Christopher-Wolf.de)
@version 0.1
*/

//Copyright (C) 2002    Christopher Wolf
//
//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License (file LICENCE) for more details.


class Transform4 {

  /** holds the bijective matrix M for the affine transformation.
  */
  private Matrix4 M;

  /** holds the LU decomposition of matrix M for the reverse affine transformation.
  */
  private LUDecomposition4 lu;

  /** holds the vector m for the affine transformation.
  */
  private int[] m;

  /** the number of rows and columns, which is taken from the field extension.
  */
  static private int rowCol = Poly64Max.extension;

  /** Generates a affine transformation which is the identity, i.e.
      M = I and m = 0.
  */
  public Transform4() {
    M = Matrix4.identity(rowCol,rowCol);
    lu = M.lu();
    m = new int[rowCol];
  }

  /** Generates a affine transformation using a PRNG
    @param rnd Random number generator for generating the transformation
  */
  public Transform4(Random rnd) {
    Object tmp[] = LUDecomposition4.generateNonsingular(rowCol,rnd);
    M = (Matrix4)tmp[0];
    lu = (LUDecomposition4)tmp[1];
    m = new int[rowCol];
    for (int i = 0; i < rowCol; i++)
      m[i] = Field4.rnd(rnd);
  }

  /** applies the transformation to a given vector, i.e. res = M*v^T + m
    @param vec vector which should be transformed using this
    @param res vector under this transformation
  */
  public void transform(int[] vec, int[] res) {
    Check.assert(Field4.isElement(vec), "vec is not over Field4.");
    Check.assert(vec.length == rowCol, "vec has the wrong length, namely " + vec.length);
    Check.assert(res.length == rowCol, "res has the wrong length, namely " + vec.length);

    M.mul(vec,res);
    // add the two vectors
    for (int i = 0; i < rowCol; i++)
      res[i] = Field4.add(res[i],m[i]);
  }

  /** Applies the transformation to a given vector of polynomials, i.e. res = M*in + m.
    @param in vector which should be transformed using this
    @return A vector of polynomials
  */
  public Poly4MultiSquare[] transform(Poly4MultiSquare[] in) {
    Check.assert(in.length == rowCol, "Wrong number of polynomials.");
    final int terms = Poly4MultiSquare.tau(rowCol);  // how many terms has every polynomial?
    int[][] Ms = M.getArray(); // all the coefficients of our matrix

    int[][] polys = new int[rowCol][]; // save all the coefficients of our polynomials for faster access
    for (int i = 0; i < rowCol; i++)
      polys[i] = in[i].coeffs();

    // in this section we first iterate over the result polynomials
    // second over each individual coefficient and third over
    // the input polynomials
    Poly4MultiSquare[] res = new Poly4MultiSquare[rowCol];
    for (int i = 0; i < rowCol; i++) {
      int[] a = new int[terms];  // create a vector which has the right number of elements
      for (int j = 0; j < terms; j++) {
        int sum = (j == 0) ? m[i] : 0;  // get the correct init value
        for (int k = 0; k < rowCol; k++) {
          sum = Field4.add(sum,Field4.mul(polys[k][j],Ms[i][k]));
        }
        a[j] = sum;  // set the coefficient in the right way
      }
      res[i] = new Poly4MultiSquare(rowCol,a); // this does not use new memory nor copy, therefore quite fast
    }
    return res;
  }

  /** reverts the transformation to a given vector, i.e. computes res in vec = M*res^T + m
    @param vec vector which should be untransformed using this - <b>is changed in this method</b>
    @param res vector under this transformation
  */
  void untransform(int[] vec, int[] res) {
    Check.assert(Field4.isElement(vec), "vec is not over Field4.");
    Check.assert(vec.length == rowCol, "vec has the wrong length, namely " + vec.length);
    Check.assert(res.length == rowCol, "res has the wrong length, namely " + vec.length);

    // first subtract the two vectors
    for (int i = 0; i < rowCol; i++)
      vec[i] = Field4.sub(vec[i],m[i]);
    lu.solve(vec,res);
  }

  /** returns the transformation in terms of polynomials
  */
  public Poly4MultiLinear[] getTransform() {
    int[] coeff = new int[rowCol+1];  // one extra element for the constant term
    int[][] Ms = M.getArray();
    Poly4MultiLinear[] res = new Poly4MultiLinear[rowCol];
    for (int i = 0; i < rowCol; i++) {
      coeff[0] = m[i];
      System.arraycopy(Ms[i],0,coeff,1,rowCol);
      res[i] = new Poly4MultiLinear(coeff);
    }
    return res;
  }

  /** Writes the current transformation to stdout
  */
  public void writeString() {
    int[][] Ms = M.getArray();
    for (int i = 0; i < rowCol; i++) {
      for (int j = 0; j < rowCol; j++) {
        System.out.print(" " + Ms[i][j]);
      }
      System.out.println("  " + m[i]);
    }
  }

  /** Tests the functionality of the class.
      Is called by the main method.
  */
  public static void testIt () {
    System.out.println("start testing Transform4 ...");

    Random rnd = new Random();
    Transform4 T;
    int[] a,b,c,d,v,r;
    final int repeats = 100;   // how often do we run our tests

    // first try the identiy transformation
    T = new Transform4();
    a = new int[] {1,2,3};
    b = new int[] {1,2,3};
    r = new int[rowCol];
    T.transform(a,r);
    for(int i = 0; i < rowCol; i++) {
      if (r[i] != b[i]) System.out.println("Error 1.1 at " + i);
    }
    T.transform(r,a);
    for(int i = 0; i < rowCol; i++) {
      if (a[i] != b[i]) System.out.println("Error 1.2 at " + i);
    }

    // now use random vectors to a random transformation
    for (int i = 0; i < repeats; i++) {
      T = new Transform4(rnd);
      for (int j = 0; j < rowCol; j++) {
        a[j] = Field4.rnd(rnd);
        b[j] = a[j];
      }
      T.transform(a,r);
      T.untransform(r,a);
      for (int j = 0; j < rowCol; j++)
        if (a[j] != b[j]) System.out.println("Error 1.3 at " + i);
    }

    a = new int[] {1,2,3};
    r = new int[rowCol];
    T.M.set(0,0,1); T.M.set(0,1,2); T.M.set(0,2,3);
    T.M.set(1,0,2); T.M.set(1,1,3); T.M.set(1,2,1);
    T.M.set(2,0,3); T.M.set(2,1,0); T.M.set(2,2,2);
    T.m = new int[] {1,2,3};
    T.transform(a,r);
    if ((r[0] != 1) || (r[1] != 2) || (r[2] != 1)) System.out.println("Error 1.4 with r[0,1,2] " + r[0] + " " + r[1] + " "+ r[2]);

    a = new int[] {2,1,3};
    r = new int[rowCol];
    T.M.set(0,0,0); T.M.set(0,1,0); T.M.set(0,2,3);
    T.M.set(1,0,0); T.M.set(1,1,3); T.M.set(1,2,1);
    T.M.set(2,0,3); T.M.set(2,1,0); T.M.set(2,2,2);
    T.m = new int[] {0,2,0};
    T.transform(a,r);
    if ((r[0] != 2) || (r[1] != 2) || (r[2] != 0)) System.out.println("Error 1.5 with r[0,1,2] " + r[0] + " " + r[1] + " "+ r[2]);

    // now test if polynomials are the same as a transformation
    Poly4MultiLinear[] poly;
    // first try the identiy transformation
    T = new Transform4();
    poly = T.getTransform();
    for(int i = 0; i < rowCol; i++) {
      for (int j = 0; j < rowCol+1; j++) {
        if (((poly[i].coeff(j) == 0) && ((i+1) == j)) ||
            ((poly[i].coeff(j) != 0) && ((i+1) != j))) System.out.println("Error 2.1 at (" + i + ", " + j + ")");
      }
    }

    // a little more sophisticated example
    a = new int[]    {1,2,3};
    b = new int[] {0, 1,2,3};
    r = new int[rowCol];
    T.M.set(0,0,1); T.M.set(0,1,2); T.M.set(0,2,3);
    T.M.set(1,0,2); T.M.set(1,1,3); T.M.set(1,2,1);
    T.M.set(2,0,3); T.M.set(2,1,0); T.M.set(2,2,2);
    T.m = new int[] {1,2,3};
    T.transform(a,r);
    if ((r[0] != 1) || (r[1] != 2) || (r[2] != 1)) System.out.println("Error 2.2a with r[0,1,2] " + r[0] + " " + r[1] + " "+ r[2]);
    poly = T.getTransform();
    if (poly[0].apply(b) != 1) System.out.println("Error 2.2b with " + poly[0].apply(b));
    if (poly[1].apply(b) != 2) System.out.println("Error 2.2c with " + poly[1].apply(b));
    if (poly[2].apply(b) != 1) System.out.println("Error 2.2d with " + poly[2].apply(b));

    // example two of this kind
    a = new int[]    {2,1,3};
    b = new int[] {0, 2,1,3};
    r = new int[rowCol];
    T.M.set(0,0,0); T.M.set(0,1,0); T.M.set(0,2,3);
    T.M.set(1,0,0); T.M.set(1,1,3); T.M.set(1,2,1);
    T.M.set(2,0,3); T.M.set(2,1,0); T.M.set(2,2,2);
    T.m = new int[] {0,2,0};
    T.transform(a,r);
    if ((r[0] != 2) || (r[1] != 2) || (r[2] != 0)) System.out.println("Error 2.3a with r[0,1,2] " + r[0] + " " + r[1] + " "+ r[2]);
    poly = T.getTransform();
    if (poly[0].apply(b) != 2) System.out.println("Error 2.3b with " + poly[0].apply(b));
    if (poly[1].apply(b) != 2) System.out.println("Error 2.3c with " + poly[1].apply(b));
    if (poly[2].apply(b) != 0) System.out.println("Error 2.3d with " + poly[2].apply(b));

    // now work with random transformations
    b = new int[rowCol+1];
    b[0] = 0;  // we do not use the first value
    for (int i = 0; i < repeats; i++) {
      T = new Transform4(rnd);
      for (int j = 0; j < rowCol; j++) {
        a[j] = j+1;
        b[1+j] = a[j];
      }
      T.transform(a,r);
      poly = T.getTransform();
      for (int j = 0; j < rowCol; j++) {
        if (r[j] != poly[j].apply(b)) System.out.println("Error 2.4 with " + a[j] + " and " + poly[j].apply(b));
      }
    }

    // now test how polynomials are transformed using this transformation
    Poly4MultiSquare[] in,out;
    in = new Poly4MultiSquare[rowCol];
    // start with an identity transformation
    T = new Transform4();
    for (int i = 0; i < rowCol; i++) {
      a = new int[Poly4MultiSquare.tau(rowCol)];
      for (int j = 0; j < Poly4MultiSquare.tau(rowCol); j++)
        a[j] = Field4.rnd(rnd);
      in[i] = new Poly4MultiSquare(rowCol,a);
    }
    out = T.transform(in);
    // we expect that the two polynomial systems are the same after that tranformation
    for (int i = 0; i < rowCol; i++) {
      a = in[i].coeffs();
      b = out[i].coeffs();
      for (int j = 0; j < a.length; j++)
        if (a[j] != b[j]) System.out.println("Error 3.1 at " + i + " and " + j);
    }

    // now do some tries with random transformations
    for (int k = 0; k < repeats; k++) {
      T = new Transform4(rnd);  // generate a random transformation
      // generate a random polynomial
      for (int i = 0; i < rowCol; i++) {
        a = new int[Poly4MultiSquare.tau(rowCol)];
        for (int j = 0; j < Poly4MultiSquare.tau(rowCol); j++)
          a[j] = Field4.rnd(rnd);
        in[i] = new Poly4MultiSquare(rowCol,a);
      }
      out = T.transform(in);  // apply that transformation
      // now generate a random vector and apply that transformation to that vector
      a = new int[rowCol];   // vector
      b = new int[rowCol+1]; // values again for getValues
      c = new int[Poly4MultiSquare.tau(rowCol)];  // result of getValues
      d = new int[rowCol];   // vector after transformation
      for (int i = 0; i < rowCol; i++)
        a[i] = Field4.rnd(rnd);
      System.arraycopy(a,0,b,1,rowCol);  // get a copy of that in b
      Poly4MultiSquare.getValues(b,c);   // generate the correct values for the polynomials
      for (int i = 0; i < rowCol; i++)   // now apply the polynomials
        a[i] = in[i].apply(c);
      T.transform(a,d);  // get our transformation
      for (int i = 0; i < rowCol; i++) {
        if (d[i] != out[i].apply(c)) System.out.println("Error 3.2 at " + i + " and run " + k);
      }
    }

    System.out.println("...finished testing.");
  }

  /** Calls the {@link #testIt} method
  */
  public static void main(String[] args) {
    testIt();
  }
}
